!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!  21 January 2010                                         !!
!!                      Get Flux                            !!
!!                    Post-Procesing                        !!
!!                                                          !!
!! flux_mod                                                 !!
!! getflux                                                  !!
!! flux_read_init                                           !!
!! flux_compute                                             !!
!! flux_compute2                                            !!
!! flux_compute3                                            !!
!! nlin_vec                                                 !!
!! addtonl                                                  !!
!!                                        version 1.21_0.30 !!
!!                                                          !!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!



!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!                                                          !!
!!                      flux_mod                            !!
!!                                                          !!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
MODULE flux_mod

  INTEGER   :: flux_type
  INTEGER   :: nshell   
  REAL      :: kfirst      
  REAL      :: klast
  NAMELIST /flux_parameters/ flux_type, nshell, kfirst, klast

  REAL, ALLOCATABLE, DIMENSION(:)   :: kshell
  REAL, ALLOCATABLE, DIMENSION(:)   :: nkshell, nkshell_node

END MODULE flux_mod



!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!                                                          !!
!!                        getflux                           !!
!!                                                          !!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
PROGRAM getflux
  USE parameters
  USE mpi_var
  USE flux_mod
  IMPLICIT NONE

  INTEGER                :: rc
  INTEGER                :: iscal, ishell
  INTEGER                :: pid, qqy 
  CHARACTER(LEN=2)       :: csc  
  CHARACTER(LEN=3)       :: cnum
  REAL                   :: ener

  COMPLEX, ALLOCATABLE, DIMENSION(:,:,:,:)  :: vel, mag, scal
  REAL, ALLOCATABLE, DIMENSION(:,:)         :: flux_vel
  REAL, ALLOCATABLE, DIMENSION(:,:)         :: flux_mag
  REAL, ALLOCATABLE, DIMENSION(:,:)         :: flux_scal

  CALL MPI_INIT( ierr )
  CALL MPI_COMM_RANK( MPI_COMM_WORLD, myid, ierr )
  CALL MPI_COMM_SIZE( MPI_COMM_WORLD, numprocs, ierr )
    
  IF (myid.eq.ioid) THEN
    READ (*,*) simname
    WRITE(6,*) 'simulation name is ', simname  
  END IF

  CALL read_para()  
  CALL read_time()
 
  CALL init_fft3d()
  CALL init_wavevectors()

  IF (myid.eq.ioid) THEN
    WRITE(cnum,'(i3.3)') nfields
    WRITE(6,*) 'The file that will be analysed is number  ', nfields  
  END IF 

  CALL flux_read_init()
  !! Pi=[k.Z][Y.X*]
  !! flux_compute(Z,Y,X)


  IF (VELEQ) THEN
    ALLOCATE(vel(0:cx,0:cz,0:cy,3),flux_vel(0:nshell,1))
	  flux_vel=0.0
    CALL read_vec_divfree(vel,simname//'_u')
    CALL correl_vector(vel,vel,0,0.5,ener)
    IF (myid.eq.ioid) WRITE(6,*) 'Kinetic energy   = ',ener
    CALL flux_compute(vel,vel,vel,flux_vel(:,1))  !! u< to u>
 	  IF(.not.MAGEQ) DEALLOCATE(vel)
  END IF

  IF (MAGEQ) THEN
    ALLOCATE(mag(0:cx,0:cz,0:cy,3),flux_mag(0:nshell,5))
	  flux_mag=0.0
    pid = whereis_ky(0,1)
    qqy = whereis_ky(0,2)
    IF (myid.eq.pid) THEN
      IF (MAGEQ) THEN
        mag(0,0,qqy,1)=b0_x*Ns
        mag(0,0,qqy,2)=b0_y*Ns
        mag(0,0,qqy,3)=b0_z*Ns
      END IF
    END IF
    CALL read_vec_divfree(mag,simname//'_b')
    CALL correl_vector(mag,mag,0,0.5,ener)
    IF (myid.eq.ioid) WRITE(6,*) 'Magnetic energy  = ',ener
    CALL flux_compute(vel,mag,mag,flux_mag(:,1))  !! b< to b>
    CALL flux_compute(mag,mag,vel,flux_mag(:,2))  !! u< to b>
    CALL flux_compute(mag,vel,mag,flux_mag(:,3))  !! b< to u>
    CALL flux_compute2(mag,mag,vel,flux_mag(:,4)) !! u< to b<
    CALL flux_compute3(mag,mag,vel,flux_mag(:,5)) !! u> to b>
    DEALLOCATE(vel,mag)
  END IF

  IF (SCAEQ) THEN
    ALLOCATE(scal(0:cx,0:cz,0:cy,NSCAL),flux_scal(0:nshell,NSCAL))
    DO iscal=1,NSCAL
      WRITE(csc,'(i2.2)') iscal
      CALL read_scal(,simname//'_s'//csc//'.')
 	  END DO
    DEALLOCATE(scal)
  END IF

  !! Save flux
  IF ((VELEQ).AND.(myid.eq.ioid)) THEN
    OPEN (UNIT=10, FILE=simname//'_flx.'//cnum, STATUS='UNKNOWN')
    DO ishell=0,nshell
      IF (MAGEQ) THEN
        WRITE (10,'(I5,E12.4,I16,6E18.8)') ishell, kshell(ishell), nkshell(ishell), flux_vel(ishell,1), flux_mag(ishell,1), &
                                    flux_mag(ishell,2), flux_mag(ishell,3), flux_mag(ishell,4), flux_mag(ishell,5)
      ELSE
        WRITE (10,'(I5,E12.4,I16,6E18.8)') ishell, kshell(ishell), nkshell(ishell), flux_vel(ishell,1), 0.0, 0.0, 0.0, 0.0, 0.0
      END IF
    END DO
    CLOSE (10)
  END IF

  IF (myid.eq.ioid) WRITE(6,*) 'Fluxes computed'
  IF (VELEQ) DEALLOCATE(flux_vel)
  IF (MAGEQ) DEALLOCATE(flux_mag)
  CALL MPI_FINALIZE(rc)

END PROGRAM getflux



!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!                                                          !!
!!                     flux_read_init                       !!
!!                                                          !!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
SUBROUTINE flux_read_init()
  USE parameters
  USE mpi_var
  USE flux_mod
  IMPLICIT NONE

  CHARACTER(LEN=namelength+4)  :: filename
  CHARACTER(LEN=2)             :: csc 
  INTEGER                      :: qx, qy, qz, ishell
  REAL                         :: energy


  !! Read parameters
  filename=simname//'.par'
  IF (myid.eq.ioid) THEN 
    OPEN(10,FILE=filename,STATUS='OLD',FORM='FORMATTED')
    READ(unit=10,NML=flux_parameters)
    CLOSE(10)
  END IF
  CALL MPI_BCAST(flux_type,1,MPI_INTEGER,ioid,MPI_COMM_WORLD,ierr)
  CALL MPI_BCAST(nshell,1,MPI_INTEGER,ioid,MPI_COMM_WORLD,ierr)
  CALL MPI_BCAST(kfirst,1,MPI_MYREAL,ioid,MPI_COMM_WORLD,ierr)
  CALL MPI_BCAST(klast,1,MPI_MYREAL,ioid,MPI_COMM_WORLD,ierr)


  MM=2
  MM(0)=1
  MM(nx/2)=1 !! ORIGINALLY 0
!  kshell = [ 0., 1., 2., 4., 8., 9.514, 11.314, 13.454, 16., 19.027, 22.627, 26.909, 32., &
!                 & 38.055, 45.255, 53.817, 64., 76.109, 90.51, 107.63, 128., 256.]

  ALLOCATE(kshell(0:nshell))  
  SELECT CASE(flux_type)
  CASE(0)  
    kshell(0)=kfirst
    DO ishell = 1,nshell
      kshell(ishell)=kshell(ishell-1)+1.0
    END DO
    kshell(nshell)=klast
  CASE(1)
    kshell(0)=0.0
    kshell(1)=2.0
    kshell(2)=4.0
    DO ishell = 4,nshell
      kshell(ishell-1)=2.0**(ishell/4.0+2.0)
    END DO
    kshell(nshell)=klast
  END SELECT
  IF (myid.eq.ioid) WRITE(6,*) kshell 


  !! Determining the number of modes
  ALLOCATE(nkshell(0:nshell),nkshell_node(0:nshell))  
  nkshell_node=0.0
  DO qy=0,cy
    DO qz=0,cz 
      DO qx=0,cx
        DO ishell = 0, nshell
          IF( knorm(qx,qz,qy) <= kshell(ishell) )  nkshell_node(ishell)=nkshell_node(ishell)+MM(qx)
        END DO
      END DO
    END DO
  END DO
  CALL mpi_allreduce(nkshell_node,nkshell,(nshell+1),MPI_MYREAL,MPI_SUM,MPI_COMM_WORLD,ierr)
  DO ishell = 0, nshell
    IF (myid.eq.ioid) WRITE(6,*) ishell, nkshell(ishell)
  END DO
  IF (myid.eq.ioid) THEN
    WRITE(6,*) 'Ns:=nx*ny*nz     =', REAL(Ns)
    WRITE(6,*) 'Last Shell Modes =', REAL(nkshell(nshell))
  END IF

END SUBROUTINE flux_read_init



!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!                                                          !!
!!                      flux_compute                        !!
!!                                                          !!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
SUBROUTINE flux_compute(Z, Y, X, flux)
  USE parameters
  USE mpi_var
  USE flux_mod
  IMPLICIT NONE
    
  COMPLEX, DIMENSION(0:cx,0:cz,0:cy,3), INTENT(IN)   :: X
  COMPLEX, DIMENSION(0:cx,0:cz,0:cy,3), INTENT(IN)   :: Y
  COMPLEX, DIMENSION(0:cx,0:cz,0:cy,3), INTENT(IN)   :: Z
  REAL, DIMENSION(0:nshell), INTENT(OUT)             :: flux

  COMPLEX, ALLOCATABLE, DIMENSION(:,:,:,:)           :: less
  COMPLEX, ALLOCATABLE, DIMENSION(:,:,:,:)           :: great
  COMPLEX, ALLOCATABLE, DIMENSION(:,:,:,:)           :: nl
  INTEGER     :: ishell, qx, qy, qz

  ALLOCATE(less(0:cx,0:cz,0:cy,3),great(0:cx,0:cz,0:cy,3),nl(0:cx,0:cz,0:cy,3))
  flux=0.0

  DO ishell=0,nshell
    less =0.0
    great=0.0
    DO qy=0,cy
      DO qz=0,cz
        DO qx=0,cx
          IF(knorm(qx,qz,qy) <= kshell(ishell)) THEN
            less(qx,qz,qy,1:3) = X(qx,qz,qy,1:3)
          ELSE
            great(qx,qz,qy,1:3)= Y(qx,qz,qy,1:3) 
          END IF
        END DO
      END DO
    END DO
    CALL nlin_vec(less,Z,nl)
    CALL correl_vector(great,nl,0,1.0,flux(ishell))
  END DO 

  DEALLOCATE(less, great, nl)

END SUBROUTINE flux_compute



!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!                                                          !!
!!                     flux_compute2                        !!
!!                                                          !!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
SUBROUTINE flux_compute2(Z, Y, X, flux)
  USE parameters
  USE mpi_var
  USE flux_mod
  IMPLICIT NONE
    
  COMPLEX, DIMENSION(0:cx,0:cz,0:cy,3), INTENT(IN)   :: X
  COMPLEX, DIMENSION(0:cx,0:cz,0:cy,3), INTENT(IN)   :: Y
  COMPLEX, DIMENSION(0:cx,0:cz,0:cy,3), INTENT(IN)   :: Z
  REAL, DIMENSION(0:nshell), INTENT(OUT)             :: flux

  COMPLEX, ALLOCATABLE, DIMENSION(:,:,:,:)           :: less
  COMPLEX, ALLOCATABLE, DIMENSION(:,:,:,:)           :: great
  COMPLEX, ALLOCATABLE, DIMENSION(:,:,:,:)           :: nl
  INTEGER     :: ishell, qx, qy, qz

  ALLOCATE(less(0:cx,0:cz,0:cy,3),great(0:cx,0:cz,0:cy,3),nl(0:cx,0:cz,0:cy,3))
  flux=0.0

  DO ishell=0,nshell
    less =0.0
    great=0.0
    DO qy=0,cy
      DO qz=0,cz
        DO qx=0,cx
          IF(knorm(qx,qz,qy) <= kshell(ishell)) THEN
            less(qx,qz,qy,1:3) = X(qx,qz,qy,1:3)
            great(qx,qz,qy,1:3)= Y(qx,qz,qy,1:3) 
          END IF
        END DO
      END DO
    END DO
    CALL nlin_vec(less,Z,nl)
    CALL correl_vector(great,nl,0,1.0,flux(ishell))
  END DO 

  DEALLOCATE(less, great, nl)

END SUBROUTINE flux_compute2



!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!                                                          !!
!!                     flux_compute3                        !!
!!                                                          !!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
SUBROUTINE flux_compute3(Z, Y, X, flux)
  USE parameters
  USE mpi_var
  USE flux_mod
  IMPLICIT NONE
    
  COMPLEX, DIMENSION(0:cx,0:cz,0:cy,3), INTENT(IN)   :: X
  COMPLEX, DIMENSION(0:cx,0:cz,0:cy,3), INTENT(IN)   :: Y
  COMPLEX, DIMENSION(0:cx,0:cz,0:cy,3), INTENT(IN)   :: Z
  REAL, DIMENSION(0:nshell), INTENT(OUT)             :: flux

  COMPLEX, ALLOCATABLE, DIMENSION(:,:,:,:)           :: less
  COMPLEX, ALLOCATABLE, DIMENSION(:,:,:,:)           :: great
  COMPLEX, ALLOCATABLE, DIMENSION(:,:,:,:)           :: nl
  INTEGER     :: ishell, qx, qy, qz

  ALLOCATE(less(0:cx,0:cz,0:cy,3),great(0:cx,0:cz,0:cy,3),nl(0:cx,0:cz,0:cy,3))
  flux=0.0

  DO ishell=0,nshell
    less =0.0
    great=0.0
    DO qy=0,cy
      DO qz=0,cz
        DO qx=0,cx
          IF(knorm(qx,qz,qy) > kshell(ishell)) THEN
            less(qx,qz,qy,1:3) = X(qx,qz,qy,1:3)
            great(qx,qz,qy,1:3)= Y(qx,qz,qy,1:3) 
          END IF
        END DO
      END DO
    END DO
    CALL nlin_vec(less,Z,nl)
    CALL correl_vector(great,nl,0,1.0,flux(ishell))
  END DO 

  DEALLOCATE(less, great, nl)

END SUBROUTINE flux_compute3



!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!                                                          !!
!!                       nlin_vec                           !!
!!                                                          !!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
SUBROUTINE nlin_vec(va,vb,nl)
  USE parameters
  USE mpi_var
  IMPLICIT NONE
  
  COMPLEX, DIMENSION(0:cx,0:cz,0:cy,3), INTENT(IN)  :: va, vb
  COMPLEX, DIMENSION(0:cx,0:cz,0:cy,3), INTENT(OUT) :: nl
  COMPLEX, ALLOCATABLE, DIMENSION (:,:,:)   :: vc
  REAL,    ALLOCATABLE, DIMENSION (:,:,:)   :: prod
  REAL,    ALLOCATABLE, DIMENSION (:,:,:)   :: va1r, va2r, va3r
  REAL,    ALLOCATABLE, DIMENSION (:,:,:)   :: vb1r, vb2r, vb3r
  
  nl = 0.0
  ALLOCATE(va1r(0:rx,0:ry,0:rz),va2r(0:rx,0:ry,0:rz),va3r(0:rx,0:ry,0:rz))
  ALLOCATE(vb1r(0:rx,0:ry,0:rz),vb2r(0:rx,0:ry,0:rz),vb3r(0:rx,0:ry,0:rz))
  ALLOCATE(vc(0:cx,0:cz,0:cy))
  ALLOCATE(prod(0:rx,0:ry,0:rz))

 
  vc=va(:,:,:,1)
  CALL ifft3ds(vc,va1r)
  vc=va(:,:,:,2)
  CALL ifft3ds(vc,va2r)
  vc=va(:,:,:,3)
  CALL ifft3ds(vc,va3r)

  vc=vb(:,:,:,1)
  CALL ifft3ds(vc,vb1r)
  vc=vb(:,:,:,2)
  CALL ifft3ds(vc,vb2r)
  vc=vb(:,:,:,3)
  CALL ifft3ds(vc,vb3r)
  

  !!!  Term 11 
  prod = va1r*vb1r
  CALL fft3ds(prod,vc)
  CALL addtonl(vc,nl(:,:,:,1),1)
    
  !!!  Term 12 
  prod = va1r*vb2r
  CALL fft3ds(prod,vc)
  CALL addtonl(vc,nl(:,:,:,1),2)
 
  !!!  Term 13 
  prod = va1r*vb3r
  CALL fft3ds(prod,vc)
  CALL addtonl(vc,nl(:,:,:,1),3)
  
  !!!  Term 21 
  prod = va2r*vb1r
  CALL fft3ds(prod,vc)
  CALL addtonl(vc,nl(:,:,:,2),1)
    
  !!!  Term 22 
  prod = va2r*vb2r
  CALL fft3ds(prod,vc)
  CALL addtonl(vc,nl(:,:,:,2),2)
 
  !!!  Term 23 
  prod = va2r*vb3r
  CALL fft3ds(prod,vc)
  CALL addtonl(vc,nl(:,:,:,2),3)
  
  !!!  Term 31 
  prod = va3r*vb1r
  CALL fft3ds(prod,vc)
  CALL addtonl(vc,nl(:,:,:,3),1)
    
  !!!  Term 32 
  prod = va3r*vb2r
  CALL fft3ds(prod,vc)
  CALL addtonl(vc,nl(:,:,:,3),2)
 
  !!!  Term 33 
  prod = va3r*vb3r
  CALL fft3ds(prod,vc)
  CALL addtonl(vc,nl(:,:,:,3),3)
  
  DEALLOCATE(va1r,va2r,va3r)
  DEALLOCATE(vb1r,vb2r,vb3r)
  DEALLOCATE(vc,prod)

END SUBROUTINE nlin_vec



!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!                                                          !!
!!                       addtonl                            !!
!!                                                          !!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
SUBROUTINE addtonl(c1,dv,igrad)
  USE parameters
  IMPLICIT NONE
  
  COMPLEX, DIMENSION(0:cx,0:cz,0:cy), INTENT(IN)       :: c1
  COMPLEX, DIMENSION(0:cx,0:cz,0:cy), INTENT(INOUT)    :: dv
  INTEGER, INTENT(IN)                                  :: igrad
  
  INTEGER :: qx, qy, qz
  
  SELECT CASE (igrad)
  CASE(1)
    DO qy=0, cy
      DO qz=0, cz
        DO qx=0, cx
          dv(qx,qz,qy) = dv(qx,qz,qy)+ikx(qx)*c1(qx,qz,qy)
        END DO
      END DO
    END DO
  CASE(2)
    DO qy=0, cy
      DO qz=0, cz
        DO qx=0, cx
          dv(qx,qz,qy) = dv(qx,qz,qy)+iky(qy)*c1(qx,qz,qy)
        END DO
      END DO
    END DO
  CASE(3)
    DO qy=0, cy
      DO qz=0, cz
        DO qx=0, cx
          dv(qx,qz,qy) = dv(qx,qz,qy)+ikz(qz)*c1(qx,qz,qy)
        END DO
      END DO
    END DO
  CASE DEFAULT
    WRITE(*,*) 'Wrong value of igrad (igrad=', igrad,')'
    STOP
  END SELECT

END SUBROUTINE addtonl



!!!!!
!!!!!
!!!!!
