!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!  12 August 2011                                          !!
!!                     Get Spectrum                         !!
!!                    Post-Procesing                        !!
!!                                                          !!
!! spec_mod                                                 !!
!! getspectrum                                              !!
!! spec_read                                                !!
!! spec_init                                                !!
!! spec_write                                               !!
!! spec_compute                                             !!
!! spec_compute_1D                                          !!
!! spec_compute_2D                                          !!
!! spec_compute_3D                                          !!
!! spec_compute_3DN                                         !!
!!                                        version 2.05_0.90 !!
!!                                                          !!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! 



!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!                                                          !!
!!                      spec_mod                            !!
!!                                                          !!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
MODULE spec_mod

  INTEGER   :: spec_type 
  REAL      :: kfirst      
  REAL      :: klast
  REAL      :: shell_width
  NAMELIST /spec_parameters/ spec_type, kfirst, klast, shell_width
  
  REAL, ALLOCATABLE, DIMENSION(:)   :: kshell, kc
  REAL, ALLOCATABLE, DIMENSION(:)   :: nkshell, nkshell_node
  INTEGER   :: nshell

END MODULE spec_mod



!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!                                                          !!
!!                       getspectrum                        !!
!!                                                          !!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
PROGRAM getspectrum
  USE parameters
  USE mpi_var
  USE spec_mod
  IMPLICIT NONE

  COMPLEX, ALLOCATABLE, DIMENSION(:,:,:,:)  :: vel, mag, scal
  REAL, ALLOCATABLE, DIMENSION(:,:)         :: spec_vel, spec_mag, spec_scal

  INTEGER                :: rc
  INTEGER                :: iscal
  CHARACTER(LEN=2)       :: csc, inum  
  CHARACTER(LEN=3)       :: cnum
  REAL                   :: ener

  CALL mpi_environment_init()    
  IF (my_fpart.eq.io_fpart) THEN
    WRITE(out_unit,*) '>>   G E T S P E C T R U M : '
    WRITE(out_unit,*) '>> simulation name is ', simname  
  END IF

  CALL read_para()
  CALL read_time()
  CALL init_fft3d()
  CALL init_wavevectors()  
 
  IF (my_fpart.eq.io_fpart) THEN
    WRITE(out_unit,*) '>> The file that will be analysed is number  ', nfields  
  END IF
  WRITE(cnum,'(i3.3)') nfields
  WRITE(inum,'(i2.2)') my_inst

  CALL spec_read()
  CALL spec_init()
  
  IF (VELEQ) THEN
    ALLOCATE(vel(0:cx,0:cz,0:cy,3),spec_vel(0:nshell+1,6))
    spec_vel=0.0
    CALL read_vec_divfree(vel,simname//'_u')
    CALL correl_vector(vel,vel,0,0.5,ener)
    IF (my_fpart.eq.io_fpart) WRITE(out_unit,*) '>> Kinetic energy   = ',ener
    CALL spec_compute(vel,spec_vel)
    IF(numinsts.GT.1.0)THEN
      CALL spec_write(spec_vel,inum//'_'//simname//'_spu.'//cnum)
    ELSE
      CALL spec_write(spec_vel,simname//'_spu.'//cnum)
    END IF
    DEALLOCATE(spec_vel)
	IF(.not.MAGEQ) DEALLOCATE(vel)
  END IF

  IF (MAGEQ) THEN
    ALLOCATE(mag(0:cx,0:cz,0:cy,3),spec_mag(0:nshell+1,6))
    spec_mag=0.0
    CALL read_vec_divfree(mag,simname//'_b')
    CALL correl_vector(mag,mag,0,0.5,ener)
    IF (my_fpart.eq.io_fpart) WRITE(out_unit,*) '>> Magnetic energy  = ',ener
    CALL spec_compute(mag,spec_mag)
    IF(numinsts.GT.1.0)THEN
      CALL spec_write(spec_mag,inum//'_'//simname//'_spb.'//cnum)
    ELSE
      CALL spec_write(spec_mag,simname//'_spb.'//cnum)
    END IF
    DEALLOCATE(vel,mag,spec_mag)
  END IF

  IF (SCAEQ) THEN
    ALLOCATE(scal(0:cx,0:cz,0:cy,NSCAL),spec_scal(0:nshell+1,2*NSCAL))
    DO iscal=1,NSCAL
      WRITE(csc,'(i2.2)') iscal
      CALL read_scal(scal(:,:,:,iscal),simname//'_s'//csc//'.')
!	  CALL spec_compute_sca(scal(:,:,:,iscal),spec_scal(:,iscal))
    END DO
!   CALL spec_write(spec_scal)
    DEALLOCATE(scal,spec_scal)
  END IF

  CALL mpi_environment_end()

END PROGRAM getspectrum



!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!                                                          !!
!!                       spec_read                          !!
!!                                                          !!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
SUBROUTINE spec_read()
  USE parameters
  USE mpi_var
  USE spec_mod
  IMPLICIT NONE

  CHARACTER(LEN=namelength+4)  :: filename
  filename=simname//'.par'

  IF (my_fpart.eq.io_fpart) THEN 
    OPEN(10,FILE=filename,STATUS='OLD',FORM='FORMATTED')
    READ(unit=10,NML=spec_parameters)
    CLOSE(10)
  END IF
  CALL MPI_BCAST(spec_type,1,MPI_INTEGER,io_fpart,MPI_COMM_INSTANCE,ierr)
  CALL MPI_BCAST(kfirst,1,MPI_MYREAL,io_fpart,MPI_COMM_INSTANCE,ierr)
  CALL MPI_BCAST(klast,1,MPI_MYREAL,io_fpart,MPI_COMM_INSTANCE,ierr)
  CALL MPI_BCAST(shell_width,1,MPI_MYREAL,io_fpart,MPI_COMM_INSTANCE,ierr)
   
END SUBROUTINE spec_read



!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!                                                          !!
!!                       spec_init                          !!
!!                                                          !!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
SUBROUTINE spec_init()
  USE parameters
  USE mpi_var
  USE spec_mod
  IMPLICIT NONE
  
  INTEGER   :: qx, qy, qz, ishell

  MM=2
  MM(0)=1
  MM(nx/2)=1 !! ORIGINALLY 0

  !! Determining the number of shells (cilinders)
  nshell=CEILING((klast-kfirst)/shell_width)
  IF (my_fpart.eq.io_fpart) WRITE(out_unit,*) '>> Number of spectrum shells', nshell

  !! Determining the shells boundaries
  ALLOCATE(kshell(0:nshell+2))
  kshell(0)=0.0
  kshell(1)=kfirst
  DO ishell = 2, nshell
    kshell(ishell)=kfirst+shell_width*(ishell-1)
  END DO
  kshell(nshell+1)=klast
  kshell(nshell+2)=REAL(MIN(nx,ny,nx))

  !! Determining the shells midpoint wavevector
  ALLOCATE(kc(0:nshell))
  kc(0)=0.0
  DO ishell = 1, nshell
	kc(ishell)=kfirst+shell_width*(REAL(ishell)-0.5)
	IF(kc(ishell) > klast) kc(ishell) = klast
  END DO
  IF (my_fpart.eq.io_fpart) WRITE(out_unit,*) '>> Spectrum shells borders: ', kshell  
  IF (my_fpart.eq.io_fpart) WRITE(out_unit,*) '>> Spectrum shells centers: ', kc

  !! Determining the number of modes in a shell
  ALLOCATE(nkshell(0:nshell+1),nkshell_node(0:nshell+1)) 
  nkshell_node=0.0
  DO qy=0,cy
   DO qz=0,cz 
      DO qx=0,cx
        DO ishell = 0, nshell+1
          SELECT CASE(spec_type)
          CASE(1)   !! 1D
            IF( ( ABS(kx(qx)) > kshell(ishell)).AND.(ABS(kx(qx)) <= kshell(ishell+1)) ) THEN
              nkshell_node(ishell)=nkshell_node(ishell)+MM(qx)
            END IF
          CASE(2,5)   !! 2D
            IF( ( knorm(qx,0,qy) > kshell(ishell)).AND.(knorm(qx,0,qy) <= kshell(ishell+1)) ) THEN
              nkshell_node(ishell)=nkshell_node(ishell)+MM(qx)
            END IF
          CASE(3,4) !! 3D
            IF( ( knorm(qx,qz,qy) > kshell(ishell)).AND.(knorm(qx,qz,qy) <= kshell(ishell+1)) ) THEN
              nkshell_node(ishell)=nkshell_node(ishell)+MM(qx)
            END IF
          CASE DEFAULT
          END SELECT 
        END DO
      END DO
    END DO
  END DO
  CALL mpi_allreduce(nkshell_node,nkshell,(nshell+2),MPI_MYREAL,MPI_SUM,MPI_COMM_INSTANCE,ierr)
  
  SELECT CASE(spec_type)
  CASE(1)   !! 1D
    nkshell(0)= nkshell(0) + nx**2  !! asumes square box
  CASE(2,5) !! 2D
    nkshell(0)= nkshell(0) + nz
  CASE(3,4) !! 3D
    nkshell(0)= nkshell(0) + 1.0
  CASE DEFAULT
  END SELECT 
  DO ishell = 0, nshell
    IF (my_fpart.eq.io_fpart) WRITE(out_unit,*) ishell, nkshell(ishell)
  END DO
  IF (my_fpart.eq.io_fpart) THEN
    WRITE(out_unit,*) '>> Ns:=nx*ny*nz    =', REAL(Ns)
    WRITE(out_unit,*) '>> Total SUM Modes =', SUM(nkshell(:))
    WRITE(out_unit,*) '>> Shell SUM Modes =', SUM(nkshell(0:nshell))
  END IF

END SUBROUTINE spec_init



!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!                                                          !!
!!                       spec_write                         !!
!!                                                          !!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
SUBROUTINE spec_write(spec_a,filename)
  USE parameters
  USE mpi_var
  USE spec_mod
  IMPLICIT NONE

  REAL, DIMENSION(0:nshell+1,6), INTENT(IN)     :: spec_a
  CHARACTER(LEN=namelength+11),  INTENT(IN)     :: filename
  CHARACTER(LEN=namelength+8)                   :: shortname
  INTEGER   :: ishell

  IF (my_fpart.eq.io_fpart) THEN
    IF(numinsts.GT.1)THEN
      OPEN (UNIT=10, FILE=filename, STATUS='UNKNOWN', FORM='FORMATTED')
    ELSE
      shortname=filename(1:namelength+8)
      OPEN (UNIT=10, FILE=shortname, STATUS='UNKNOWN', FORM='FORMATTED')
    END IF
    DO ishell = 0, nshell
      WRITE(10,'(I10,2F15.2,8F21.8)')    ishell, kc(ishell), nkshell(ishell), &
                      spec_a(ishell,1),  spec_a(ishell,2),  spec_a(ishell,3), &
                      spec_a(ishell,1) + spec_a(ishell,2) + spec_a(ishell,3), &
                      spec_a(ishell,4),  spec_a(ishell,5),  spec_a(ishell,6), &
                      spec_a(ishell,4) + spec_a(ishell,5) + spec_a(ishell,6)
    END DO
    CLOSE (UNIT=10)
  END IF

END SUBROUTINE spec_write



!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!                                                          !!
!!                     spec_compute                         !!
!!                                                          !!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
SUBROUTINE spec_compute(a, spec_a)
  USE parameters
  USE mpi_var
  USE spec_mod
  IMPLICIT NONE
  
  COMPLEX, DIMENSION(0:cx,0:cz,0:cy,3), INTENT(IN) :: a
  REAL, DIMENSION(0:nshell+1,6), INTENT(OUT)       :: spec_a

  SELECT CASE(spec_type)
  CASE(1) !! 1D
    CALL spec_compute_1D(a, spec_a)
    IF (my_fpart.eq.io_fpart) THEN
      WRITE(out_unit,*) '>> Total SUM energy = ',SUM(spec_a(:,1:3))
      WRITE(out_unit,*) '>> Shell SUM energy = ',SUM(spec_a(0:nshell,1:3))
    END IF
  CASE(2) !! 2D
    CALL spec_compute_2D(a, spec_a)
    IF (my_fpart.eq.io_fpart) THEN
      WRITE(out_unit,*) '>> Total SUM energy = ',SUM(spec_a(:,1)), SUM(spec_a(:,3))
      WRITE(out_unit,*) '>> Shell SUM energy = ',SUM(spec_a(0:nshell,1)), SUM(spec_a(0:nshell,3))
    END IF
  CASE(3) !! 3D
    CALL spec_compute_3D(a, spec_a)
    IF (my_fpart.eq.io_fpart) THEN
      WRITE(out_unit,*) '>> Total SUM energy = ',SUM(spec_a(:,1:3))
      WRITE(out_unit,*) '>> Shell SUM energy = ',SUM(spec_a(0:nshell,1:3))
    END IF
  CASE(4) !! 3DN
    CALL spec_compute_3DN(a, spec_a)
    IF (my_fpart.eq.io_fpart) THEN
      WRITE(out_unit,*) '>> Total SUM energy = ',SUM(spec_a(:,1:3))
      WRITE(out_unit,*) '>> Shell SUM energy = ',SUM(spec_a(0:nshell,1:3))
    END IF
  CASE DEFAULT
    WRITE(out_unit,*) '>> Wrong value for spec_type parameter: ', spec_type 
  END SELECT 

END SUBROUTINE spec_compute



!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!                                                          !!
!!                     spec_compute_1D                      !!
!!                                                          !!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
SUBROUTINE spec_compute_1D(a, spec)
  USE parameters
  USE mpi_var
  USE spec_mod
  IMPLICIT NONE

  COMPLEX, DIMENSION(0:cx,0:cz,0:cy,3), INTENT(IN)   :: a
  REAL, DIMENSION(0:nshell+1,6), INTENT(OUT)         :: spec
  REAL, DIMENSION(0:nshell+1,6)                      :: spec_node
  REAL, ALLOCATABLE, DIMENSION(:)                    :: Ekx_node, Eky_node, Ekz_node
  REAL, ALLOCATABLE, DIMENSION(:)                    :: Dkx_node, Dky_node, Dkz_node
  INTEGER    :: qx, qy, qz, pid, qqy, ishell
  REAL       :: u2, v2, w2

  ALLOCATE(Ekx_node(0:cx),Eky_node(0:cy),Ekz_node(0:cz))
  ALLOCATE(Dkx_node(0:cx),Dky_node(0:cy),Dkz_node(0:cz))

  Ekx_node=0.0
  Eky_node=0.0
  Ekz_node=0.0
  Dkx_node=0.0
  Dky_node=0.0
  Dkz_node=0.0
  DO qy=0,cy
    DO qz=0,cz
      DO qx=0,cx
        u2=ABS(a(qx,qz,qy,1))**2*MM(qx)
        v2=ABS(a(qx,qz,qy,2))**2*MM(qx)
        w2=ABS(a(qx,qz,qy,3))**2*MM(qx)
        Ekx_node(qx) = Ekx_node(qx) + u2*0.5
        Eky_node(qy) = Eky_node(qy) + v2*0.5
        Ekz_node(qz) = Ekz_node(qz) + w2*0.5
        Dkx_node(qx) = Dkx_node(qx) + u2*ksquare(qx,qz,qy)
        Dky_node(qy) = Dky_node(qy) + v2*ksquare(qx,qz,qy)
        Dkz_node(qz) = Dkz_node(qz) + w2*ksquare(qx,qz,qy)
      END DO
    END DO
  END DO

  spec_node=0.0
  DO ishell=0,nshell
    DO qx=0,cx
      IF( (ABS(kx(qx)) > kshell(ishell)).AND.(ABS(kx(qx)) <= kshell(ishell+1)) ) THEN
        spec_node(ishell,1)=spec_node(ishell,1)+Ekx_node(qx)
        spec_node(ishell,4)=spec_node(ishell,4)+Dkx_node(qx)
      END IF
    END DO
    DO qy=0,cy
      IF( (ABS(ky(qy)) > kshell(ishell)).AND.(ABS(ky(qy)) <= kshell(ishell+1)) ) THEN
        spec_node(ishell,2)=spec_node(ishell,2)+Eky_node(qy)
        spec_node(ishell,5)=spec_node(ishell,5)+Dky_node(qy)
	  END IF
    END DO
    DO qz=0,cz
      IF( (ABS(kz(qz)) > kshell(ishell)).AND.(ABS(kz(qz)) <= kshell(ishell+1)) ) THEN
        spec_node(ishell,3)=spec_node(ishell,3)+Ekz_node(qz)
        spec_node(ishell,6)=spec_node(ishell,6)+Dkz_node(qz)
	  END IF
    END DO
  END DO
  
  pid=whereis_ky(0,1)
  qqy=whereis_ky(0,2)
  IF(myid.eq.pid) THEN
    spec_node(0,2)=spec_node(0,2)+Eky_node(qqy)
    spec_node(0,5)=spec_node(0,5)+Dky_node(qqy)
  END IF
  spec_node(0,1)=spec_node(0,1)+Ekx_node(0)
  spec_node(0,3)=spec_node(0,3)+Ekz_node(0)
  spec_node(0,4)=spec_node(0,4)+Dkx_node(0)
  spec_node(0,6)=spec_node(0,6)+Dkz_node(0)
        
  CALL mpi_allreduce(spec_node,spec,6*(nshell+1),MPI_MYREAL,MPI_SUM,MPI_COMM_INSTANCE,ierr)
  spec=spec/(Ck*Ns)**2

  DEALLOCATE(Ekx_node,Eky_node,Ekz_node)
  DEALLOCATE(Dkx_node,Dky_node,Dkz_node)

END SUBROUTINE spec_compute_1D



!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!                                                          !!
!!                     spec_compute_2D                      !!
!!                                                          !!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
SUBROUTINE spec_compute_2D(a, spec)
  USE parameters
  USE mpi_var
  USE spec_mod
  IMPLICIT NONE
  
  COMPLEX, DIMENSION(0:cx,0:cz,0:cy,3), INTENT(IN)   :: a
  REAL, DIMENSION(0:nshell+1,6), INTENT(OUT)         :: spec
  REAL, DIMENSION(0:nshell+1,6)                      :: spec_node
  REAL, ALLOCATABLE, DIMENSION(:,:)                  :: Ek
  REAL, ALLOCATABLE, DIMENSION(:,:)                  :: Dk
  INTEGER    :: qx, qy, qz, pid, qqy, ishell, jshell
  REAL       :: u2, v2, w2, ko

  ALLOCATE(Ek(0:nshell+1,0:nshell+1),Dk(0:nshell+1,0:nshell+1))
  Ek=0.0
  Dk=0.0
  
  DO qy=0,cy
    DO qx=0,cx
      DO qz=0,cz
  		DO ishell=0,nshell+1
          DO jshell=0,nshell+1	  
            IF( (knorm(qx,0,qy) > kshell(ishell)).AND.(knorm(qx,0,qy) <= kshell(ishell+1)) ) THEN
              IF( (ABS(kz(qz)) > kshell(jshell)).AND.(ABS(kz(qz)) <= kshell(jshell+1)) ) THEN 
                u2=ABS(a(qx,qz,qy,1))**2*MM(qx)
                v2=ABS(a(qx,qz,qy,2))**2*MM(qx)
                w2=ABS(a(qx,qz,qy,3))**2*MM(qx)
                Ek(ishell,jshell)=Ek(ishell,jshell)+(u2+v2+w2)*0.5
                Dk(ishell,jshell)=Dk(ishell,jshell)+(u2+v2+w2)*ksquare(qx,qz,qy) 
              END IF              
            END IF
          END DO
        END DO
      END DO
    END DO
  END DO  
  
  DO qy=0,cy
    DO qx=0,cx
      DO ishell=0,nshell+1
        IF( (knorm(qx,0,qy) > kshell(ishell)).AND.(knorm(qx,0,qy) <= kshell(ishell+1)) ) THEN
          u2=ABS(a(qx,0,qy,1))**2*MM(qx)
          v2=ABS(a(qx,0,qy,2))**2*MM(qx)
          w2=ABS(a(qx,0,qy,3))**2*MM(qx)
          Ek(ishell,0)=Ek(ishell,0)+(u2+v2+w2)*0.5
          Dk(ishell,0)=Dk(ishell,0)+(u2+v2+w2)*ksquare(qx,0,qy) 
        END IF              
      END DO
    END DO    
  END DO  
    
  pid=whereis_ky(0,1)
  qqy=whereis_ky(0,2)
  IF(myid.eq.pid) THEN
    DO qz=0,cz
      DO jshell=0,nshell+1	  
        IF( (ABS(kz(qz)) > kshell(jshell)).AND.(ABS(kz(qz)) <= kshell(jshell+1)) ) THEN 
          u2=ABS(a(0,qz,qqy,1))**2*MM(0)
          v2=ABS(a(0,qz,qqy,2))**2*MM(0)
          w2=ABS(a(0,qz,qqy,3))**2*MM(0)
          Ek(0,jshell)=Ek(0,jshell)+(u2+v2+w2)*0.5
          Dk(0,jshell)=Dk(0,jshell)+(u2+v2+w2)*ksquare(0,qz,qqy) 
        END IF
      END DO
    END DO
  END IF
  
  spec_node=0.0
  DO ishell=0,nshell+1
    DO jshell=0,nshell+1
      spec_node(ishell,1)=spec_node(ishell,1)+Ek(ishell,jshell)
      spec_node(ishell,3)=spec_node(ishell,3)+Ek(jshell,ishell)    
      spec_node(ishell,4)=spec_node(ishell,4)+Dk(ishell,jshell)
      spec_node(ishell,6)=spec_node(ishell,6)+Dk(jshell,ishell)     
    END DO
  END DO

  DO ishell=0,nshell+1
    spec_node(ishell,2)=spec_node(ishell,1)*(nz*pi*(kshell(ishell+1)**2-kshell(ishell)**2))/nkshell(ishell)  
    spec_node(ishell,5)=spec_node(ishell,4)*(nz*pi*(kshell(ishell+1)**2-kshell(ishell)**2))/nkshell(ishell)
  END DO

  CALL mpi_allreduce(spec_node,spec,6*(nshell+1),MPI_MYREAL,MPI_SUM,MPI_COMM_INSTANCE,ierr)
  spec=spec/(Ck*Ns)**2
   
  DEALLOCATE(Ek,Dk)

END SUBROUTINE spec_compute_2D



!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!                                                          !!
!!                     spec_compute_3D                      !!
!!                                                          !!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
SUBROUTINE spec_compute_3D(a, spec)
  USE parameters
  USE mpi_var
  USE spec_mod
  IMPLICIT NONE
  
  COMPLEX, DIMENSION(0:cx,0:cz,0:cy,3), INTENT(IN)   :: a
  REAL, DIMENSION(0:nshell+1,6), INTENT(OUT)         :: spec
  REAL, DIMENSION(0:nshell+1,6)                      :: spec_node
  INTEGER    :: qx, qy, qz, ishell
  REAL       :: u2, v2, w2

  spec_node=0.0
  spec=0.0
  DO qy=0,cy
    DO qx=0,cx
      DO qz=0,cz
  		DO ishell=0,nshell+1
          IF( (knorm(qx,qz,qy) > kshell(ishell)).AND.(knorm(qx,qz,qy) <= kshell(ishell+1)) ) THEN
            u2=ABS(a(qx,qz,qy,1))**2*MM(qx)
            v2=ABS(a(qx,qz,qy,2))**2*MM(qx)
            w2=ABS(a(qx,qz,qy,3))**2*MM(qx)
            spec_node(ishell,1)=spec_node(ishell,1)+u2*0.5
            spec_node(ishell,2)=spec_node(ishell,2)+v2*0.5
            spec_node(ishell,3)=spec_node(ishell,3)+w2*0.5
            spec_node(ishell,4)=spec_node(ishell,4)+u2*ksquare(qx,qz,qy)
            spec_node(ishell,5)=spec_node(ishell,5)+v2*ksquare(qx,qz,qy)
            spec_node(ishell,6)=spec_node(ishell,6)+w2*ksquare(qx,qz,qy)
          END IF
        END DO
      END DO
    END DO
  END DO  
  spec_node=spec_node/(Ck*Ns)**2
  CALL mpi_allreduce(spec_node,spec,6*(nshell+2),MPI_MYREAL,MPI_SUM,MPI_COMM_INSTANCE,ierr)

END SUBROUTINE spec_compute_3D



!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!                                                          !!
!!                   spec_compute_3DN                       !!
!!                                                          !!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
SUBROUTINE spec_compute_3DN(a, spec)
  USE parameters
  USE mpi_var
  USE spec_mod
  IMPLICIT NONE
  
  COMPLEX, DIMENSION(0:cx,0:cz,0:cy,3), INTENT(IN)   :: a
  REAL, DIMENSION(0:nshell+1,6), INTENT(OUT)         :: spec
  REAL, DIMENSION(0:nshell+1,6)                      :: spec_node
  INTEGER    :: qx, qy, qz, ishell, j
  REAL       :: u2, v2, w2

  spec = 0.0
  spec_node = 0.0
  DO qy=0,cy
    DO qx=0,cx
      DO qz=0,cz
  		DO ishell=0,nshell
          IF( (knorm(qx,qz,qy) > kshell(ishell)).AND.(knorm(qx,qz,qy) <= kshell(ishell+1)) ) THEN
            u2=ABS(a(qx,qz,qy,1))**2*MM(qx)
            v2=ABS(a(qx,qz,qy,2))**2*MM(qx)
            w2=ABS(a(qx,qz,qy,3))**2*MM(qx)
            spec_node(ishell,1)=spec_node(ishell,1)+u2*0.5
            spec_node(ishell,2)=spec_node(ishell,2)+v2*0.5
            spec_node(ishell,3)=spec_node(ishell,3)+w2*0.5
            spec_node(ishell,4)=spec_node(ishell,4)+u2*ksquare(qx,qz,qy)
            spec_node(ishell,5)=spec_node(ishell,5)+v2*ksquare(qx,qz,qy)
            spec_node(ishell,6)=spec_node(ishell,6)+w2*ksquare(qx,qz,qy)
          END IF
        END DO
      END DO
    END DO
  END DO  
  spec_node=spec_node/(Ck*Ns)**2
  CALL mpi_allreduce(spec_node,spec,6*(nshell+2),MPI_MYREAL,MPI_SUM,MPI_COMM_INSTANCE,ierr)

  DO j=1,6
    DO ishell=0,nshell
      spec(ishell,j)=spec(ishell,j)*(4*pi*kc(ishell)**2*shell_width)/nkshell(ishell)
    END DO
  END DO

END SUBROUTINE spec_compute_3DN
