!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!  12 August 2011                                          !!
!!           Probability Density Function of C              !!
!!                    Post-Procesing                        !!
!!                                                          !!
!! pdf_c_mod                                                !!
!! pdf_c                                                    !!
!! pdf_c_init                                               !!
!! pdf_c_write                                              !!
!! pdf_c_compute                                            !!
!!                                        version 2.05_0.90 !!
!!                                                          !!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! 



!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!                                                          !!
!!                     pdf_c_mod                            !!
!!                                                          !!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
MODULE pdf_c_mod

  REAL     ::  pdf_half_width
  INTEGER  ::  num_intervals, show_clipping

  NAMELIST /pdf_c_parameters/ pdf_half_width, num_intervals, show_clipping
  
  REAL,     ALLOCATABLE,  DIMENSION(:)   ::  c_values
  INTEGER,  ALLOCATABLE,  DIMENSION(:)   ::  pdf_c_node, pdf_c_global
  INTEGER   ::  total_sum, total_interv
  REAL      ::  interv_width, normalizing

END MODULE pdf_c_mod



!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!                                                          !!
!!                          pdf_c                           !!
!!                                                          !!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
PROGRAM pdf_c
  USE parameters
  USE variables
  USE mpi_var
  USE pdf_c_mod
  IMPLICIT NONE
 
  CALL mpi_environment_init()

  CALL read_para()
  time_to_zero=0
  init_vel=0
  CALL read_time()
  
  CALL init_variables() 

  CALL pdf_c_init() 
  CALL pdf_c_compute()
  CALL pdf_c_write()
  
  DEALLOCATE(c_values)
  DEALLOCATE(pdf_c_node)
  DEALLOCATE(pdf_c_global)
  IF (VELEQ) DEALLOCATE(vel,dvel)
  IF (MAGEQ) DEALLOCATE(mag,dmag)
  IF (SCAEQ) DEALLOCATE(scal,dscal)
  IF (filter_type.GT.0) DEALLOCATE(test_filter)

  CALL mpi_environment_end()

END PROGRAM pdf_c



!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!                                                          !!
!!                      pdf_c_init                          !!
!!                                                          !!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
SUBROUTINE pdf_c_init()
  USE parameters
  USE variables
  USE mpi_var
  USE pdf_c_mod
  IMPLICIT NONE
  
  INTEGER  ::  iCvalue
  CHARACTER(LEN=namelength+4)  ::  filename
  
  filename=simname//'.par'  
  
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!  
! Read and broadcast the needed informations
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  IF (my_fpart.eq.io_fpart) THEN 
    OPEN(10,FILE=filename,STATUS='OLD',FORM='FORMATTED')
    READ(unit=10,NML=pdf_c_parameters)
    CLOSE(10)
  END IF
  
  IF(show_clipping.GT.1) STOP
  
  CALL MPI_BCAST(num_intervals,1,MPI_INTEGER,ioid,MPI_COMM_WORLD,ierr)
  CALL MPI_BCAST(pdf_half_width,1,MPI_MYREAL,ioid,MPI_COMM_WORLD,ierr)
  CALL MPI_BCAST(show_clipping,1,MPI_INTEGER,ioid,MPI_COMM_WORLD,ierr)
  
  ALLOCATE(c_values(-num_intervals:num_intervals+1))
  ALLOCATE(pdf_c_node(-num_intervals:num_intervals+1))
  ALLOCATE(pdf_c_global(-num_intervals:num_intervals+1))
  
  pdf_c_node=0
  pdf_c_global=0
  interv_width=pdf_half_width/num_intervals
  
  DO iCvalue=-num_intervals,num_intervals
    c_values(iCvalue)=iCvalue*interv_width
  END DO
  c_values(num_intervals+1)=999
  
END SUBROUTINE pdf_c_init



!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!                                                          !!
!!                      pdf_c_write                         !!
!!                                                          !!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
SUBROUTINE pdf_c_write()
  USE parameters
  USE variables
  USE mpi_var
  USE pdf_c_mod
  IMPLICIT NONE

  REAL,  DIMENSION(:),  ALLOCATABLE  ::  pdf_c_real
  INTEGER  ::  iCvalue
  
  ALLOCATE(pdf_c_real(-num_intervals:num_intervals+1))

  IF (myid.eq.ioid) THEN
    OPEN (UNIT=10, FILE='pdf_c.txt', STATUS='UNKNOWN', FORM='FORMATTED')
    DO iCvalue = -num_intervals,num_intervals+1
      pdf_c_real(iCvalue)=pdf_c_global(iCvalue)/normalizing
      WRITE(10,'(2F15.9)') c_values(iCvalue), pdf_c_real(iCvalue)
    END DO
    CLOSE (UNIT=10)
  END IF
  
  DEALLOCATE(pdf_c_real)

END SUBROUTINE pdf_c_write



!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!                                                          !!
!!                    pdf_c_compute                         !!
!!                                                          !!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
SUBROUTINE pdf_c_compute()
  USE parameters
  USE variables
  USE mpi_var
  USE pdf_c_mod
  IMPLICIT NONE 
  
  COMPLEX,  DIMENSION(:,:,:,:),  ALLOCATABLE  ::  af
  REAL,     DIMENSION(:,:,:,:),  ALLOCATABLE  ::  Sfij, Sij, Mij, Lij
  REAL,     DIMENSION(:,:,:),    ALLOCATABLE  ::  Snorm, Sfnorm, Sup_node, Inf_node, Sup_fpart, Inf_fpart
  REAL,     DIMENSION(:,:,:),    ALLOCATABLE  ::  C                
  REAL,     DIMENSION(:,:,:),    ALLOCATABLE  ::  r1
  COMPLEX,  DIMENSION(:,:,:),    ALLOCATABLE  ::  c1
  REAL      ::  alph1, nu_tot
  INTEGER   ::  qx, qy, qz, j, iCvalue, flag_found  

  ALLOCATE(Sij(0:rx,0:ry,0:rz,6))
  CALL compute_strain(vel, Sij)  
  
  ALLOCATE(Snorm(0:rx,0:ry,0:rz))                    
  Snorm  = Sij(:,:,:,1)**2 + 2.0*Sij(:,:,:,2)**2 + 2.0 * Sij(:,:,:,3)**2 + Sij(:,:,:,4)**2 + 2.0*Sij(:,:,:,5)**2 + Sij(:,:,:,6)**2
  Snorm  = SQRT(2.0*Snorm)
  
  ALLOCATE(c1(0:cx,0:cz,0:cy))
  ALLOCATE(r1(0:rx,0:ry,0:rz))
   
  ALLOCATE(af(0:cx,0:cz,0:cy,3))
  CALL filter(vel(:,:,:,1), af(:,:,:,1))
  CALL filter(vel(:,:,:,2), af(:,:,:,2))
  CALL filter(vel(:,:,:,3), af(:,:,:,3))
  ALLOCATE(Sfij(0:rx,0:ry,0:rz,6))
  CALL compute_strain(af,Sfij)
  DEALLOCATE(af)
  ALLOCATE(Sfnorm(0:rx,0:ry,0:rz))
  Sfnorm = Sfij(:,:,:,1)**2 + 2.0*Sfij(:,:,:,2)**2 + 2.0*Sfij(:,:,:,3)**2 + Sfij(:,:,:,4)**2 + 2.0*Sfij(:,:,:,5)**2 + Sfij(:,:,:,6)**2
  Sfnorm = SQRT(2.0*Sfnorm)
  
  SELECT CASE (filter_type)
  CASE(1)
    alph1=2.0
  CASE(2)
    alph1=4.0
  END SELECT
  ALLOCATE(Mij(0:rx,0:ry,0:rz,6))
  DO j =1, 6 
    r1 = Snorm*Sij(:,:,:,j) 
    CALL fft3ds(r1,c1)
    CALL filter(c1,c1)
    CALL ifft3ds(c1,r1)
    Mij(:,:,:,j)   = 2.0*(r1 - alph1**2 * Sfnorm*Sfij(:,:,:,j))
  END DO        
  DEALLOCATE(Sfnorm)
  DEALLOCATE(Sfij)
 
  ALLOCATE(Lij(0:rx,0:ry,0:rz,6))
  CALL compute_leonard(vel,Lij)

  ALLOCATE(Sup_node(0:rx,0:ry,0:rz))
  ALLOCATE(Inf_node(0:rx,0:ry,0:rz))
  ALLOCATE(Sup_fpart(0:rx,0:ry,0:rz))
  ALLOCATE(Inf_fpart(0:rx,0:ry,0:rz))
  Sup_node = Lij(:,:,:,1)*Mij(:,:,:,1) + 2.0*Lij(:,:,:,2)*Mij(:,:,:,2) + 2.0*Lij(:,:,:,3)*Mij(:,:,:,3) + Lij(:,:,:,4)*Mij(:,:,:,4) + 2.0*Lij(:,:,:,5)*Mij(:,:,:,5) + Lij(:,:,:,6)*Mij(:,:,:,6)
  Inf_node = Mij(:,:,:,1)**2 + 2.0*Mij(:,:,:,2)**2 + 2.0*Mij(:,:,:,3)**2 + Mij(:,:,:,4)**2 + 2.0*Mij(:,:,:,5)**2 + Mij(:,:,:,6)**2    
  CALL mpi_allreduce(Sup_node, Sup_fpart, (rx+1)*(ry+1)*(rz+1),MPI_MYREAL,MPI_SUM,MPI_COMM_FPART,ierr)
  CALL mpi_allreduce(Inf_node, Inf_fpart, (rx+1)*(ry+1)*(rz+1),MPI_MYREAL,MPI_SUM,MPI_COMM_FPART,ierr)
  DO qz=0, rz
    DO qy=0, ry
      DO qx=0, rx-2
        IF(ABS(Inf_fpart(qx,qy,qz)).LT.EPS) THEN
          IF(Inf_fpart(qx,qy,qz).LT.0.0) THEN
            Inf_fpart(qx,qy,qz)=-EPS
          ELSE
            Inf_fpart(qx,qy,qz)=EPS
          END IF
        END IF
      END DO
    END DO
  END DO
  ALLOCATE(C(0:rx,0:ry,0:rz))
  C=Sup_fpart/Inf_fpart  
  DEALLOCATE(Sup_node)
  DEALLOCATE(Inf_node)
  DEALLOCATE(Inf_fpart)
  DEALLOCATE(Sup_fpart)
  DEALLOCATE(Mij)
  DEALLOCATE(Lij)  
  
  IF(show_clipping.EQ.1) THEN
    DO qz=0, rz      ! ! ! Clipping if nu_tot .LT. 0
      DO qy=0, ry
        DO qx=0, rx
          nu_tot=nu+2*C(qx,qy,qz)*Snorm(qx,qy,qz)
          IF(nu_tot.LT.0.0) C(qx,qy,qz) = - nu / (2*Snorm(qx,qy,qz))
!          IF(C(qx,qy,qz).LT.0) C(qx,qy,qz)=0
        END DO
      END DO
    END DO
  END IF
  
  DO qz=0, rz
    DO qy=0, ry
      DO qx=0, rx-2
        flag_found=0 
        DO iCvalue=-num_intervals,num_intervals
          IF( ( C(qx,qy,qz).GT.(c_values(iCvalue)-0.5*interv_width) ) .AND. ( C(qx,qy,qz).LE.(c_values(iCvalue)+0.5*interv_width) ) ) THEN
            pdf_c_node(iCvalue)=pdf_c_node(iCvalue)+1
            flag_found=1
          END IF
        END DO
        IF(flag_found.EQ.0) THEN 
          pdf_c_node(num_intervals+1)=pdf_c_node(num_intervals+1)+1
          WRITE(6,*) '>> Trashed Value = ',C(qx,qy,qz)
        END IF
      END DO
    END DO
  END DO
  
  CALL mpi_allreduce(pdf_c_node,pdf_c_global,2*(num_intervals+1),MPI_INTEGER,MPI_SUM,MPI_COMM_WORLD,ierr)
  
  normalizing=REAL(SUM(pdf_c_global))
  
  DEALLOCATE(Snorm)
  DEALLOCATE(Sij)
  DEALLOCATE(C)

END SUBROUTINE pdf_c_compute