!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!  12 November 2009                                        !!
!!                     Ring-2-Ring                          !!
!!                    Post-Procesing                        !!
!!                                                          !!
!! r2r_mod                                                  !!
!! r2r                                                      !!
!! r2r_read                                                 !! 
!! r2r_compute                                              !!
!! r2r_compute_sca                                          !!
!! nlin_vec                                                 !!
!! nlin_sca                                                 !!
!! addtonl                                                  !!
!! ring_filter                                              !!
!! ring_filter_sca                                          !!
!! ring_mode_count                                          !!
!! r2r_net_ring_transfer                                    !!
!!                                        version 1.21_0.62 !!
!!                                                          !!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! 



!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!                                                          !!
!!                       r2r_mod                            !!
!!                                                          !!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
MODULE r2r_mod
  
  INTEGER   :: r2r_type
  INTEGER   :: nshell_radial
  INTEGER   :: nshell_angle
  INTEGER   :: filter
  INTEGER   :: kinit  
  REAL      :: gshell
  REAL, DIMENSION(0:100)    :: kgiven   
  NAMELIST /r2r_para/ r2r_type, nshell_radial, nshell_angle, filter, kinit, gshell, kgiven 
  
  COMPLEX, ALLOCATABLE, DIMENSION(:,:,:,:)  :: vel, mag, scal

  REAL , ALLOCATABLE, DIMENSION(:,:,:,:)    :: r2r_uu
  REAL , ALLOCATABLE, DIMENSION(:,:,:,:)    :: r2r_ub
  REAL , ALLOCATABLE, DIMENSION(:,:,:,:)    :: r2r_bu
  REAL , ALLOCATABLE, DIMENSION(:,:,:,:)    :: r2r_bb
  REAL , ALLOCATABLE, DIMENSION(:,:,:,:,:)  :: r2r_ss
 
  REAL , ALLOCATABLE, DIMENSION(:)          :: kshell
  REAL , ALLOCATABLE, DIMENSION(:)          :: theta  

  INTEGER , ALLOCATABLE, DIMENSION(:,:)     :: r2r_Nk
  REAL , ALLOCATABLE, DIMENSION(:,:)        :: r2r_eu
  REAL , ALLOCATABLE, DIMENSION(:,:)        :: r2r_du
  REAL , ALLOCATABLE, DIMENSION(:,:)        :: r2r_eb
  REAL , ALLOCATABLE, DIMENSION(:,:)        :: r2r_db
  REAL , ALLOCATABLE, DIMENSION(:,:)        :: r2r_cr
  REAL , ALLOCATABLE, DIMENSION(:,:)        :: r2r_B0

  REAL , ALLOCATABLE, DIMENSION(:,:)        :: r2r_uuu
  REAL , ALLOCATABLE, DIMENSION(:,:)        :: r2r_ubb
  REAL , ALLOCATABLE, DIMENSION(:,:)        :: r2r_bub
  REAL , ALLOCATABLE, DIMENSION(:,:)        :: r2r_bbu
 
END MODULE r2r_mod



!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!                                                          !!
!!                          r2r                             !!
!!                                                          !!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
PROGRAM r2r
  USE parameters
  USE mpi_var
  USE r2r_mod
  IMPLICIT NONE

  INTEGER                      :: rc
  CHARACTER(LEN=namelength+8)  :: filename
  CHARACTER(LEN=3)             :: cnum
  CHARACTER(LEN=2)             :: csc 
  INTEGER                      :: iscal , pid, qqy 
  INTEGER                      :: ir1, ir2, ia1, ia2

  CALL MPI_INIT(ierr)
  CALL MPI_COMM_RANK( MPI_COMM_WORLD, myid, ierr)
  CALL MPI_COMM_SIZE( MPI_COMM_WORLD, numprocs, ierr)
  
  IF (myid.eq.ioid) THEN
    READ (*,*) simname
    WRITE(6,*) 'simulation name is ', simname  
  END IF

  CALL read_para()  
  CALL read_time()
  CALL init_fft3d()
  CALL init_wavevectors()

  WRITE(cnum,'(i3.3)') nfields
  IF (myid.eq.ioid)  WRITE(6,*) 'R2R Statistics performed on fields:', nfields 

  CALL r2r_read_init()

  ALLOCATE(r2r_Nk(nshell_radial,nshell_angle))
  r2r_Nk=0
  CALL ring_mode_count(r2r_Nk)


  SELECT CASE (r2r_type)
  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  CASE(0)      !! R2R MODE COUNTING
    IF (myid.eq.ioid) THEN 
      OPEN (UNIT=10, FILE=simname//'_r2r_modes.'//cnum, STATUS='UNKNOWN')
      DO ir1=1,nshell_radial
        DO ia1=1, nshell_angle
          WRITE (10,'(2I10,F15.6,I10)') ir1, ia1, kshell(ir1), r2r_Nk(ir1, ia1)
        END DO
      END DO
      CLOSE (10)
    END IF
    IF (myid.eq.ioid) WRITE(6,*) 'Modes counted'

  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  CASE(1)      !! R2R ENERGY TRANSFER
    !!!!! Impose zero modes
    pid = whereis_ky(0,1)
    qqy = whereis_ky(0,2)
    IF (myid.eq.pid) THEN
      IF ((VELEQ).AND.(.NOT.ROTEQ)) THEN
        vel(0,0,qqy,1)=0.0
        vel(0,0,qqy,2)=0.0
        vel(0,0,qqy,3)=0.0
      END IF
      IF (MAGEQ) THEN
        mag(0,0,qqy,1)=0.0
        mag(0,0,qqy,2)=0.0
        mag(0,0,qqy,3)=0.0
      END IF
    END IF

    !! Compute r2r_uu   
    IF (VELEQ) THEN  
      ALLOCATE(r2r_uu(nshell_radial,nshell_angle,nshell_radial,nshell_angle))
      r2r_uu=0.0
      CALL r2r_compute(vel,vel,vel,r2r_uu)
      r2r_uu = -r2r_uu
    END IF  

    !! If MHD run compute r2rub, r2rbu and r2rbb
    IF (MAGEQ) THEN 
      ALLOCATE(r2r_ub(nshell_radial,nshell_angle,nshell_radial,nshell_angle))
      ALLOCATE(r2r_bu(nshell_radial,nshell_angle,nshell_radial,nshell_angle))
      ALLOCATE(r2r_bb(nshell_radial,nshell_angle,nshell_radial,nshell_angle))
      r2r_ub=0.0
      r2r_bu=0.0
      r2r_bb=0.0
      CALL r2r_compute(vel,mag,mag,r2r_ub)
	  CALL r2r_compute(mag,vel,mag,r2r_bu)
	  CALL r2r_compute(mag,mag,vel,r2r_bb)
      r2r_bb = -r2r_bb
    END IF
     
    !! If scalar present compute r2rss
    IF (SCAEQ) THEN 
      ALLOCATE(r2r_ss(nshell_radial,nshell_angle,nshell_radial,nshell_angle,NSCAL))
      r2r_ss=0.0
      DO iscal =1, NSCAL
	    CALL r2r_compute_sca(scal(:,:,:,iscal),vel,r2r_ss(:,:,:,:,iscal))
      END DO 
      r2r_ss = -r2r_ss
    END IF

    !! SAVE RING_TO_RING transfers
    IF ((VELEQ).AND.(myid.eq.ioid)) THEN
      filename=simname//'_r2r.'//cnum
      OPEN (UNIT=10, FILE=filename, STATUS='UNKNOWN')
      DO ir1=1,nshell_radial
        DO ia1=1, nshell_angle
          DO ir2=1, nshell_radial
            DO ia2=1, nshell_angle
              IF (MAGEQ) THEN
                WRITE (10,'(4I5,2I10,4E18.8)') ir1, ia1, ir2, ia2, r2r_Nk(ir1,ia1), r2r_Nk(ir2,ia2), & 
                         r2r_uu(ir1,ia1,ir2,ia2), r2r_ub(ir1,ia1,ir2,ia2), r2r_bu(ir1,ia1,ir2,ia2), r2r_bb(ir1,ia1,ir2,ia2)
              ELSE
                WRITE (10,'(4I5,2I10,E18.8)') ir1, ia1, ir2, ia2,r2r_Nk(ir1,ia1), r2r_Nk(ir2,ia2), r2r_uu(ir1,ia1,ir2,ia2)
              END IF
            END DO
          END DO
        END DO
      END DO
      CLOSE (10)
    END IF
    IF ((SCAEQ).AND.(myid.eq.ioid)) THEN
      filename=simname//'_rss.'//cnum
      OPEN (UNIT=10, FILE=filename, STATUS='UNKNOWN')
      DO iscal=1, NSCAL
        DO ir1=1,nshell_radial
          DO ia1=1, nshell_angle
            DO ir2=1, nshell_radial
              DO ia2=1, nshell_angle
                WRITE (10,'(5I5,2I10,E18.8)') ir1, ia1, ir2, ia2, iscal,r2r_Nk(ir1,ia1), r2r_Nk(ir2,ia2), r2r_ss(ir1,ia1,ir2,ia2,iscal)
              END DO
            END DO
          END DO
        END DO
      END DO
      CLOSE (10)
    END IF
    IF (myid.eq.ioid) WRITE(6,*) 'Energy transfer computed'

  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  CASE(2)      !! R2R MEAN MAGNETIC FIELD TRANSFER
    IF (MAGEQ) THEN  
      ALLOCATE(r2r_B0(nshell_radial,nshell_angle))
	  r2r_B0=0.0
      CALL r2r_B0_transfer(vel,mag,r2r_B0)
      IF (myid.eq.ioid) THEN 
        OPEN (UNIT=10, FILE=simname//'_r2r_B0.'//cnum, STATUS='UNKNOWN')
        DO ir1=1,nshell_radial
          DO ia1=1, nshell_angle
            WRITE (10,'(3I10, E18.8)') ir1, ia1, r2r_Nk(ir1, ia1), r2r_B0(ir1, ia1)
          END DO
        END DO
        CLOSE (10)
      END IF
      DEALLOCATE(r2r_B0)
    END IF 
    IF (myid.eq.ioid) WRITE(6,*) 'B0 linear transfer computed'

  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  CASE(3)      !! R2R ENERGY DENSITY
    ALLOCATE(r2r_eu(nshell_radial,nshell_angle))
    ALLOCATE(r2r_du(nshell_radial,nshell_angle))
    ALLOCATE(r2r_eb(nshell_radial,nshell_angle))
    ALLOCATE(r2r_db(nshell_radial,nshell_angle))
    ALLOCATE(r2r_cr(nshell_radial,nshell_angle))
    r2r_eu=0.0
    r2r_du=0.0
    r2r_eb=0.0
    r2r_db=0.0
    r2r_cr=0.0
    IF (VELEQ) THEN  
      CALL r2r_correl_vector(vel,vel,0,0.5,r2r_eu)
      CALL r2r_correl_vector(vel,vel,1,nu,r2r_du)
    END IF 
    IF (MAGEQ) THEN  
      CALL r2r_correl_vector(mag,mag,0,0.5,r2r_eb)
      CALL r2r_correl_vector(mag,mag,1,eta,r2r_db)
      CALL r2r_correl_vector(vel,mag,0,1.0,r2r_cr)
    END IF 
    IF (myid.eq.ioid) THEN 
      OPEN (UNIT=10, FILE=simname//'_r2r_stat.'//cnum, STATUS='UNKNOWN')
      DO ir1=1,nshell_radial
        DO ia1=1, nshell_angle
          WRITE (10,'(3I10, 5E18.8)') ir1, ia1, r2r_Nk(ir1, ia1), r2r_eu(ir1, ia1), r2r_eb(ir1, ia1), r2r_du(ir1, ia1), r2r_db(ir1, ia1), r2r_cr(ir1, ia1)
        END DO
      END DO
      CLOSE (10)
    END IF
    DEALLOCATE(r2r_eu)
    DEALLOCATE(r2r_du)
    DEALLOCATE(r2r_eb)
    DEALLOCATE(r2r_db)
    DEALLOCATE(r2r_cr)
    IF (myid.eq.ioid) WRITE(6,*) 'R2R stats coumputed'

  CASE(4)      !! R2R NET TRANSFER
    ALLOCATE(r2r_uuu(nshell_radial,nshell_angle))
    ALLOCATE(r2r_ubb(nshell_radial,nshell_angle))
    ALLOCATE(r2r_bub(nshell_radial,nshell_angle))
    ALLOCATE(r2r_bbu(nshell_radial,nshell_angle))
    r2r_uuu=0.0
    r2r_ubb=0.0
    r2r_bub=0.0
    r2r_bbu=0.0	   
    IF (VELEQ) THEN  
      CALL r2r_net_ring_transfer(vel,vel,vel, r2r_uuu)
	  r2r_uuu = -r2r_uuu
    END IF  
    IF (MAGEQ) THEN 
      CALL r2r_net_ring_transfer(vel,mag,mag,r2r_ubb)
	  CALL r2r_net_ring_transfer(mag,vel,mag,r2r_bub)
	  CALL r2r_net_ring_transfer(mag,mag,vel,r2r_bbu)
      r2r_bbu = -r2r_bbu
    END IF
    IF (myid.eq.ioid) THEN 
      OPEN (UNIT=10, FILE=simname//'_r2r_net.'//cnum, STATUS='UNKNOWN')
        DO ir1=1,nshell_radial
          DO ia1=1, nshell_angle
            WRITE (10,'(2I10, 2F16.4, I10, 2F22.4,4E18.8)') ir1, ia1, kshell(ir1), theta(ia1), r2r_Nk(ir1, ia1), 0.5*(kshell(ir1)+kshell(ir1+1))*kshell(ir1)**2, COS(theta(ia1+1))-COS(theta(ia1)),   &
			                                         r2r_uuu(ir1, ia1),r2r_ubb(ir1, ia1),r2r_bub(ir1, ia1),r2r_bbu(ir1, ia1)
          END DO
        END DO
      CLOSE (10)
      WRITE(6,*) 'Net r2r transfer computed',SUM(r2r_uuu)+SUM(r2r_ubb)+SUM(r2r_bub)+SUM(r2r_bbu)
    END IF 
    DEALLOCATE(r2r_uuu)
    DEALLOCATE(r2r_ubb)
    DEALLOCATE(r2r_bub)
    DEALLOCATE(r2r_bbu)
  CASE DEFAULT
    WRITE (6,*) 'Wrong value of r2r_type', r2r_type
    STOP
  END SELECT

  IF (VELEQ) DEALLOCATE(vel)
  IF (MAGEQ) DEALLOCATE(mag)
  IF (SCAEQ) DEALLOCATE(scal)
   
  CALL MPI_FINALIZE(rc)

END PROGRAM r2r



!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!                                                          !!
!!                    r2r_read_init                         !!
!!                                                          !!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
SUBROUTINE r2r_read_init()
  USE parameters
  USE mpi_var
  USE r2r_mod
  IMPLICIT NONE

  CHARACTER(LEN=namelength+4)  :: filename
  CHARACTER(LEN=2)             :: csc 
  INTEGER                      :: iscal, pid, qqy 
  INTEGER                      :: ir1, ir2, ia1, ia2 
  REAL                         :: energy

  filename=simname//'.par'
  kgiven=0.0
!! READ PARAMETERS
  IF (myid.eq.ioid) THEN 
    OPEN(10,FILE=filename,STATUS='OLD',FORM='FORMATTED')
    READ(unit=10,NML=r2r_para)
    CLOSE(10)
  END IF
  CALL MPI_BCAST(r2r_type,1,MPI_INTEGER,ioid,MPI_COMM_WORLD,ierr)
  CALL MPI_BCAST(nshell_radial,1,MPI_INTEGER,ioid,MPI_COMM_WORLD,ierr)
  CALL MPI_BCAST(nshell_angle,1,MPI_INTEGER,ioid,MPI_COMM_WORLD,ierr)
  CALL MPI_BCAST(filter,1,MPI_INTEGER,ioid,MPI_COMM_WORLD,ierr)
  CALL MPI_BCAST(kinit,1,MPI_INTEGER,ioid,MPI_COMM_WORLD,ierr)  
  CALL MPI_BCAST(gshell,1,MPI_MYREAL,ioid,MPI_COMM_WORLD,ierr)
  CALL MPI_BCAST(kgiven,101,MPI_MYREAL,ioid,MPI_COMM_WORLD,ierr)

  MM=2
  MM(0)=1
  MM(nx/2)=1 !! ORIGINALLY 0

!  kshell = [ 0., 1., 2., 4., 8., 9.514, 11.314, 13.454, 16., 19.027, 22.627, 26.909, 32., &
!                 & 38.055, 45.255, 53.817, 64., 76.109, 90.51, 107.63, 128., 256.]

  ALLOCATE(kshell(nshell_radial+1))  
  ALLOCATE(theta(nshell_angle+1))
  
  SELECT CASE(kinit)
  CASE(0)  
    kshell(1:nshell_radial-1)=kgiven(0:nshell_radial-2)
    kshell(nshell_radial)=kgiven(99)   
    kshell(nshell_radial+1)=kgiven(100) 
  CASE(1)
    IF(nshell_radial > 1) kshell(1)=0.0
    IF(nshell_radial > 2) kshell(2)=kgiven(1)
    IF(nshell_radial > 3) kshell(3)=kgiven(2)
    IF(nshell_radial > 4) THEN
      DO ir1 = 4,nshell_radial
        kshell(ir1)=gshell**(ir1/4.0+2.0) !(ir1-2)
      END DO
    END IF
    kshell(nshell_radial)=kgiven(99)
    kshell(nshell_radial+1) = 10000000.0
  CASE(2)
    IF(nshell_radial > 1) kshell(1)=0.0
    IF(nshell_radial > 2) kshell(2)=kgiven(1)
    IF(nshell_radial > 3) THEN
      DO ir1 = 3,nshell_radial
        kshell(ir1)=kgiven(1)*gshell**(ir1/2.0-1)
      END DO
    END IF
    kshell(nshell_radial)=kgiven(99)
    kshell(nshell_radial+1) = 10000000.0   
  CASE(3)
    IF(nshell_radial > 1) kshell(1)=0.0
    IF(nshell_radial > 2) kshell(2)=kgiven(1)
    IF(nshell_radial > 3) THEN
      DO ir1 = 2,nshell_radial
        kshell(ir1+1)=kgiven(1)*gshell**(ir1-2.0) !(ir1-2)
      END DO
    END IF
    IF(nshell_radial > 3) kshell(3)=kgiven(2)    
    kshell(nshell_radial)=kgiven(99)
    kshell(nshell_radial+1) = 10000000.0   
  END SELECT
  IF (myid.eq.ioid) WRITE(6,*) kshell 


!! DEFINE RING BORDERS
  DO ia1 = 1,nshell_angle+1
    theta(ia1)=pi/nshell_angle*(ia1-1)
  END DO 
  IF (myid.eq.ioid) WRITE(6,*) theta
  

!! READ VARIABLES
  IF (VELEQ) THEN
    ALLOCATE(vel(0:cx,0:cz,0:cy,3))
    CALL read_vec_divfree(vel,simname//'_u')
    CALL correl_vector(vel,vel,0.0,0.5,energy)
    IF (myid.eq.ioid) WRITE (*,*) 'energy of the velocity field: ', energy
  END IF
  IF (MAGEQ) THEN
    ALLOCATE(mag(0:cx,0:cz,0:cy,3))
    CALL read_vec_divfree(mag,simname//'_b')
    CALL correl_vector(mag,mag,0.0,0.5,energy)
    IF (myid.eq.ioid) WRITE (*,*) 'energy of the magnetic field: ', energy
  END IF
  IF (SCAEQ) THEN
    ALLOCATE(scal(0:cx,0:cz,0:cy,NSCAL))
    DO iscal=1,NSCAL
      WRITE(csc,'(i2.2)') iscal
      CALL read_scal(scal(:,:,:,iscal),simname//'_s'//csc//'.')
    END DO
  END IF
  WRITE(*,*) 'Variables are read'

  !!!!! Impose zero modes
  pid = whereis_ky(0,1)
  qqy = whereis_ky(0,2)
  IF (myid.eq.pid) THEN
    IF ((VELEQ).AND.(.NOT.ROTEQ)) THEN
      vel(0,0,qqy,1)=u0_x*Ns
      vel(0,0,qqy,2)=u0_y*Ns
      vel(0,0,qqy,3)=u0_z*Ns
    END IF
    IF (MAGEQ) THEN
      mag(0,0,qqy,1)=b0_x*Ns
      mag(0,0,qqy,2)=b0_y*Ns
      mag(0,0,qqy,3)=b0_z*Ns
    END IF
  END IF

END SUBROUTINE r2r_read_init



!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!                                                          !!
!!                       r2r_compute                        !!
!!                                                          !!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
SUBROUTINE r2r_compute(X,Y,Z,r2r)
  USE parameters
  USE mpi_var
  USE r2r_mod
  IMPLICIT NONE

  COMPLEX, DIMENSION(0:cx,0:cz,0:cy,3), INTENT(IN)  :: X  !! reciver
  COMPLEX, DIMENSION(0:cx,0:cz,0:cy,3), INTENT(IN)  :: Y  !! giver 
  COMPLEX, DIMENSION(0:cx,0:cz,0:cy,3), INTENT(IN)  :: Z  !! helper
  REAL, DIMENSION(nshell_radial,nshell_angle,nshell_radial,nshell_angle), INTENT(OUT) :: r2r

  COMPLEX, ALLOCATABLE, DIMENSION(:,:,:,:)  :: giver, receiver, nl
  INTEGER     :: ir1, ir2, ia1, ia2

  ALLOCATE(giver(0:cx,0:cz,0:cy,3), receiver(0:cx,0:cz,0:cy,3), nl(0:cx,0:cz,0:cy,3))

  DO ir1=1,nshell_radial
    IF (myid.eq.ioid) WRITE(*,*) 'main loop: radial index is ', ir1
    DO ia1=1,nshell_angle
      IF (myid.eq.ioid) WRITE(*,*) 'main loop: angular index is ', ia1
      CALL ring_filter(Y,kshell(ir1),kshell(ir1+1),theta(ia1),theta(ia1+1),giver)
      CALL nlin_vec(giver,Z,nl)
      WRITE(*,*) 'nlin_vec done'
      DO ir2=1,nshell_radial
        IF (myid.eq.ioid) WRITE(*,*) 'secondary loop: radial index is ', ir2
        DO ia2=1,nshell_angle
          IF (myid.eq.ioid) WRITE(*,*) 'secondary loop: angular index is ', ia2
          CALL ring_filter(X,kshell(ir2),kshell(ir2+1),theta(ia2),theta(ia2+1),receiver)
          CALL correl_vector(receiver,nl,0,1.0,r2r(ir1,ia1,ir2,ia2))
        END DO
      END DO
    END DO
  END DO 
  
  DEALLOCATE(giver, receiver, nl)

END SUBROUTINE r2r_compute



!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!                                                          !!
!!                     r2r_compute_sca                      !!
!!                                                          !!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
SUBROUTINE r2r_compute_sca(S,Z,r2r)
  USE parameters
  USE mpi_var
  USE r2r_mod
  IMPLICIT NONE

  COMPLEX, DIMENSION(0:cx,0:cz,0:cy), INTENT(IN)    :: S  !! reciver and giver 
  COMPLEX, DIMENSION(0:cx,0:cz,0:cy,3), INTENT(IN)  :: Z  !! helper
  REAL, DIMENSION(nshell_radial,nshell_angle,nshell_radial,nshell_angle), INTENT(OUT) :: r2r

  COMPLEX, ALLOCATABLE, DIMENSION(:,:,:)    :: giv_sca, rec_sca, nl_sca
  INTEGER     :: ir1, ir2, ia1, ia2

  ALLOCATE(giv_sca(0:cx,0:cz,0:cy), rec_sca(0:cx,0:cz,0:cy), nl_sca(0:cx,0:cz,0:cy))
        
  DO ir1=1,nshell_radial
    IF (myid.eq.ioid) WRITE(*,*) 'main loop: radial index is ', ir1
    DO ia1=1,nshell_angle
      IF (myid.eq.ioid) WRITE(*,*) 'main loop: angular index is ', ir1
      CALL ring_filter_sca(S,kshell(ir1),kshell(ir1+1),theta(ia1),theta(ia1+1),giv_sca)
      CALL nlin_sca(giv_sca,Z,nl_sca)
      WRITE(*,*) 'nlin_vec done'
      DO ir2=1,nshell_radial
        IF (myid.eq.ioid) WRITE(*,*) 'secondary loop: radial index is ', ir2
        DO ia2=1,nshell_angle
          IF (myid.eq.ioid) WRITE(*,*) 'secondary loop: angular index is ', ir2
          CALL ring_filter_sca(S,kshell(ir2),kshell(ir2+1),theta(ia2),theta(ia2+1),rec_sca)
          CALL correl_scalar(rec_sca,nl_sca,0,1.0,r2r(ir1,ia1,ir2,ia2))
        END DO
      END DO
    END DO
  END DO 
  
  DEALLOCATE(giv_sca,rec_sca,nl_sca)

END SUBROUTINE r2r_compute_sca



!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!                                                          !!
!!                       nlin_vec                           !!
!!                                                          !!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
SUBROUTINE nlin_vec(va,vb,nl)
  USE parameters
  USE mpi_var
  IMPLICIT NONE
  
  COMPLEX, DIMENSION(0:cx,0:cz,0:cy,3), INTENT(IN)  :: va, vb
  COMPLEX, DIMENSION(0:cx,0:cz,0:cy,3), INTENT(OUT) :: nl
  COMPLEX, ALLOCATABLE, DIMENSION (:,:,:)   :: vc
  REAL,    ALLOCATABLE, DIMENSION (:,:,:)   :: prod
  REAL,    ALLOCATABLE, DIMENSION (:,:,:)   :: va1r, va2r, va3r
  REAL,    ALLOCATABLE, DIMENSION (:,:,:)   :: vb1r, vb2r, vb3r
  
  nl = 0.0
  ALLOCATE(va1r(0:rx,0:ry,0:rz),va2r(0:rx,0:ry,0:rz),va3r(0:rx,0:ry,0:rz))
  ALLOCATE(vb1r(0:rx,0:ry,0:rz),vb2r(0:rx,0:ry,0:rz),vb3r(0:rx,0:ry,0:rz))
  ALLOCATE(vc(0:cx,0:cz,0:cy))
  ALLOCATE(prod(0:rx,0:ry,0:rz))

 
  vc=va(:,:,:,1)
  CALL ifft3ds(vc,va1r)
  vc=va(:,:,:,2)
  CALL ifft3ds(vc,va2r)
  vc=va(:,:,:,3)
  CALL ifft3ds(vc,va3r)

  vc=vb(:,:,:,1)
  CALL ifft3ds(vc,vb1r)
  vc=vb(:,:,:,2)
  CALL ifft3ds(vc,vb2r)
  vc=vb(:,:,:,3)
  CALL ifft3ds(vc,vb3r)
  

  !!!  Term 11 
  prod = va1r*vb1r
  CALL fft3ds(prod,vc)
  CALL addtonl(vc,nl(:,:,:,1),1)
    
  !!!  Term 12 
  prod = va1r*vb2r
  CALL fft3ds(prod,vc)
  CALL addtonl(vc,nl(:,:,:,1),2)
 
  !!!  Term 13 
  prod = va1r*vb3r
  CALL fft3ds(prod,vc)
  CALL addtonl(vc,nl(:,:,:,1),3)
  
  !!!  Term 21 
  prod = va2r*vb1r
  CALL fft3ds(prod,vc)
  CALL addtonl(vc,nl(:,:,:,2),1)
    
  !!!  Term 22 
  prod = va2r*vb2r
  CALL fft3ds(prod,vc)
  CALL addtonl(vc,nl(:,:,:,2),2)
 
  !!!  Term 23 
  prod = va2r*vb3r
  CALL fft3ds(prod,vc)
  CALL addtonl(vc,nl(:,:,:,2),3)
  
  !!!  Term 31 
  prod = va3r*vb1r
  CALL fft3ds(prod,vc)
  CALL addtonl(vc,nl(:,:,:,3),1)
    
  !!!  Term 32 
  prod = va3r*vb2r
  CALL fft3ds(prod,vc)
  CALL addtonl(vc,nl(:,:,:,3),2)
 
  !!!  Term 33 
  prod = va3r*vb3r
  CALL fft3ds(prod,vc)
  CALL addtonl(vc,nl(:,:,:,3),3)
  
  DEALLOCATE(va1r,va2r,va3r)
  DEALLOCATE(vb1r,vb2r,vb3r)
  DEALLOCATE(vc,prod)

END SUBROUTINE nlin_vec



!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!                                                          !!
!!                       nlin_sca                           !!
!!                                                          !!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
SUBROUTINE nlin_sca(phi,u,nl)
  USE parameters
  USE mpi_var
  IMPLICIT NONE
  
  COMPLEX, DIMENSION(0:cx,0:cz,0:cy), INTENT(IN)    :: phi
  COMPLEX, DIMENSION(0:cx,0:cz,0:cy,3), INTENT(IN)  :: u
  COMPLEX, DIMENSION(0:cx,0:cz,0:cy), INTENT(OUT)   :: nl
  REAL, ALLOCATABLE, DIMENSION (:,:,:)      :: ur, phir, prod
  COMPLEX, ALLOCATABLE, DIMENSION (:,:,:)   :: uc
  
  nl = 0.0
  ALLOCATE(ur(0:rx,0:ry,0:rz),phir(0:rx,0:ry,0:rz),uc(0:cx,0:cz,0:cy),prod(0:rx,0:ry,0:rz))

  uc=phi
  CALL ifft3ds(uc,phir)
  
  !!!  Term 11 
  uc=u(:,:,:,1)
  CALL ifft3ds(uc,ur)
  prod = ur*phir
  CALL fft3ds(prod,uc)
  CALL addtonl(uc,nl,1)
    
  !!!  Term 12 
  uc=u(:,:,:,2)
  CALL ifft3ds(uc,ur)
  prod = ur*phir
  CALL fft3ds(prod,uc)
  CALL addtonl(uc,nl,2)
 
  !!!  Term 13 
  uc=u(:,:,:,3)
  CALL ifft3ds(uc,ur)
  prod = ur*phir
  CALL fft3ds(prod,uc)
  CALL addtonl(uc,nl,3)
  
  DEALLOCATE(prod,ur,phir,uc)
    
END SUBROUTINE nlin_sca



!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!                                                          !!
!!                       addtonl                            !!
!!                                                          !!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
SUBROUTINE addtonl(c1,dv,igrad)
  USE parameters
  IMPLICIT NONE
  
  COMPLEX, DIMENSION(0:cx,0:cz,0:cy), INTENT(IN)       :: c1
  COMPLEX, DIMENSION(0:cx,0:cz,0:cy), INTENT(INOUT)    :: dv
  INTEGER, INTENT(IN)                                  :: igrad
  
  INTEGER :: qx, qy, qz
  
  SELECT CASE (igrad)
  CASE(1)
    DO qy=0, cy
      DO qz=0, cz
        DO qx=0, cx
          dv(qx,qz,qy) = dv(qx,qz,qy)+ikx(qx)*c1(qx,qz,qy)
        END DO
      END DO
    END DO
  CASE(2)
    DO qy=0, cy
      DO qz=0, cz
        DO qx=0, cx
          dv(qx,qz,qy) = dv(qx,qz,qy)+iky(qy)*c1(qx,qz,qy)
        END DO
      END DO
    END DO
  CASE(3)
    DO qy=0, cy
      DO qz=0, cz
        DO qx=0, cx
          dv(qx,qz,qy) = dv(qx,qz,qy)+ikz(qz)*c1(qx,qz,qy)
        END DO
      END DO
    END DO
  CASE DEFAULT
    WRITE(*,*) 'Wrong value of igrad (igrad=', igrad,')'
    STOP
  END SELECT

END SUBROUTINE addtonl



!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!                                                          !!
!!                       ring_filter                        !!
!!                                                          !!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
SUBROUTINE ring_filter(u, is_inf, is_sup, ia_inf, ia_sup, uf)
  USE parameters
  IMPLICIT NONE

  COMPLEX, DIMENSION(0:cx,0:cz,0:cy,3), INTENT(IN)  :: u   !! unfilterd field
  COMPLEX, DIMENSION(0:cx,0:cz,0:cy,3), INTENT(OUT) :: uf  !! filterd field
  REAL, INTENT(IN)  :: is_inf, is_sup, ia_inf, ia_sup      !! ring borders 

  CALL ring_filter_sca(u(:,:,:,1), is_inf, is_sup, ia_inf, ia_sup, uf(:,:,:,1))
  CALL ring_filter_sca(u(:,:,:,2), is_inf, is_sup, ia_inf, ia_sup, uf(:,:,:,2))
  CALL ring_filter_sca(u(:,:,:,3), is_inf, is_sup, ia_inf, ia_sup, uf(:,:,:,3))
  
END SUBROUTINE ring_filter



!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!                                                          !!
!!                     ring_filter_sca                      !!
!!                                                          !!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
SUBROUTINE ring_filter_sca(phi, is_inf, is_sup, ia_inf, ia_sup, phi_f)
  USE parameters
  USE r2r_mod
  IMPLICIT NONE

  COMPLEX, DIMENSION(0:cx,0:cz,0:cy), INTENT(IN)  :: phi   !! unfiltered field
  COMPLEX, DIMENSION(0:cx,0:cz,0:cy), INTENT(OUT) :: phi_f !! filtered field
  REAL, INTENT(IN)  :: is_inf, is_sup, ia_inf, ia_sup      !! ring borders 
  REAL, ALLOCATABLE, DIMENSION(:,:,:)             :: Gf    !! smooth filter
  
  REAL     :: kn, theta2, kcut1, kcut2, smoothpar, fac1, fac2
  INTEGER  :: qx, qy, qz
  INTEGER  :: border    
 
  !!NOTE: the 'border' value determins the correct angular selection of modes 
  !!!!!!! so that the rings are simmetrical to the xOy plane   
  IF(ia_sup < pi/2)  border = 1
  IF((ia_inf < pi/2).AND.(is_sup > pi/2))  border = 2
  IF(ia_inf > pi/2)  border = 3

  phi_f=0.0  !! the filterd field will only select the values from the ring of the unfilterd field  

!! Define the radial filter
  ALLOCATE(Gf(0:cx,0:cz,0:cy))
  Gf=0.0

  SELECT CASE(filter)
  CASE (1) !! Sharp cutoff
    DO qy=0,cy
      DO qz=0,cz
        DO qx=0,cx
          kn=knorm(qx,qz,qy)
          IF ((is_inf < kn).AND.(kn <= is_sup)) Gf(qx,qz,qy)=1.0
        END DO
      END DO
    END DO
  CASE(2) !! Smooth Fourier Cutoff
    kcut2=is_inf
    kcut1=is_sup
    smoothpar=0.2
    fac1=Pi/(2.0*smoothpar*kcut1)
    fac2=Pi/(2.0*smoothpar*kcut2)
    IF (kcut2.eq.0.0) THEN
      DO qy=0,cy
        DO qz=0,cz
          DO qx=0,cx
            kn=knorm(qx,qz,qy) 
            Gf(qx,qz,qy)=0.5*(tanh(fac1*(kn+kcut1))-tanh(fac1*(kn-kcut1)))       
          END DO
        END DO
      END DO
    ELSE 
      DO qy=0,cy
        DO qz=0,cz
          DO qx=0,cx
            kn=knorm(qx,qz,qy) 
            Gf(qx,qz,qy)=0.5*(tanh(fac1*(kn+kcut1))-tanh(fac1*(kn-kcut1)))       
            Gf(qx,qz,qy)=Gf(qx,qz,qy)-0.5*(tanh(fac2*(kn+kcut2))-tanh(fac2*(kn-kcut2)))       
          END DO
        END DO
      END DO
    END IF
  CASE DEFAULT
    WRITE(6,*) 'Wrong value for the filter'
  END SELECT


!! Define ring or band
  DO qy=0,cy
    DO qz=0,cz
      DO qx=0,cx
        kn=knorm(qx,qz,qy)
        theta2=ACOS(kz(qz)/kn)
        SELECT CASE (border)
        CASE(1)
          IF((ia_inf <= theta2).AND.(theta2 <  ia_sup))  phi_f(qx,qz,qy) = phi(qx,qz,qy)*Gf(qx,qz,qy)
        CASE(2)
          IF((ia_inf <= theta2).AND.(theta2 <= ia_sup))  phi_f(qx,qz,qy) = phi(qx,qz,qy)*Gf(qx,qz,qy)        
        CASE(3)
          IF((ia_inf <  theta2).AND.(theta2 <= ia_sup))  phi_f(qx,qz,qy) = phi(qx,qz,qy)*Gf(qx,qz,qy)
        END SELECT
      END DO
    END DO
  END DO
  
  DEALLOCATE(Gf)
  WRITE (*,*) 'ring_filter_sca done'

END SUBROUTINE ring_filter_sca



!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!                                                          !!
!!                     ring_mode_count                      !!
!!                                                          !!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
SUBROUTINE ring_mode_count(nr)
  USE parameters
  USE mpi_var
  USE r2r_mod
  IMPLICIT NONE

  INTEGER, DIMENSION(nshell_radial,nshell_angle), INTENT(OUT) :: nr
  INTEGER, DIMENSION(nshell_radial,nshell_angle)              :: nr_node
  INTEGER  :: ir1, ir2, ia1, ia2
  INTEGER  :: border, tot
  INTEGER  :: qx, qy, qz
  REAL     :: kn, theta2 

  nr=0
  nr_node=0
  DO qy=0,cy
    DO qz=0,cz
      DO qx=0,cx
        kn=knorm(qx,qz,qy)
        theta2=ACOS(kz(qz)/MAX(eps,kn))
        DO ir1=1,nshell_radial
          DO ia1=1,nshell_angle
            IF(theta(ia1+1) < pi/2.0)  border = 1
            IF((theta(ia1) < pi/2.0).AND.(theta(ia1+1) > pi/2.0))  border = 2
            IF(theta(ia1) > pi/2.0)  border = 3
            IF ((kshell(ir1) < kn).AND.(kn <= kshell(ir1+1))) THEN
              SELECT CASE (border)
              CASE(1)
                IF((theta(ia1) <= theta2).AND.(theta2 <  theta(ia1+1)))  nr_node(ir1,ia1)=nr_node(ir1,ia1)+MM(qx)
              CASE(2)
                IF((theta(ia1) <= theta2).AND.(theta2 <= theta(ia1+1)))  nr_node(ir1,ia1)=nr_node(ir1,ia1)+MM(qx)        
              CASE(3)
                IF((theta(ia1) <  theta2).AND.(theta2 <= theta(ia1+1)))  nr_node(ir1,ia1)=nr_node(ir1,ia1)+MM(qx)
              END SELECT
            END IF
          END DO
        END DO
      END DO
    END DO
  END DO
  CALL mpi_allreduce(nr_node,nr,nshell_radial*nshell_angle,MPI_INTEGER,MPI_SUM,MPI_COMM_WORLD,ierr)

  IF (myid.eq.ioid) THEN
    tot=0
    DO ir1=1,nshell_radial
      DO ia1=1,nshell_angle
        tot=tot+nr(ir1,ia1)
      END DO
    END DO
    WRITE(6,*) tot+1,Ns
  END IF

END SUBROUTINE ring_mode_count



!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!                                                          !!
!!                    r2r_correl_vector                     !!
!!                                                          !!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
SUBROUTINE r2r_correl_vector(X,Y,j,factor, en)
  USE parameters
  USE mpi_var
  USE r2r_mod
  IMPLICIT NONE

  COMPLEX, DIMENSION(0:cx,0:cz,0:cy,3), INTENT(IN)  :: X  
  COMPLEX, DIMENSION(0:cx,0:cz,0:cy,3), INTENT(IN)  :: Y 
  INTEGER, INTENT(IN)     :: j
  REAL, INTENT(IN)        :: factor
  REAL, DIMENSION(nshell_radial,nshell_angle), INTENT(OUT) :: en
  REAL, ALLOCATABLE, DIMENSION(:,:)  :: en_node
  COMPLEX     :: prx, pry, prz
  REAL        :: kn, theta2
  INTEGER     :: qx, qy, qz 
  INTEGER     :: ir1, ia1, border 
  
  ALLOCATE(en_node(nshell_radial,nshell_angle))
  en=0.0
  en_node=0
  DO qy=0,cy
    DO qz=0,cz
      DO qx=0,cx
        kn=knorm(qx,qz,qy)
        theta2=ACOS(kz(qz)/MAX(eps,kn))
        DO ir1=1,nshell_radial
          DO ia1=1,nshell_angle
            IF(theta(ia1+1) < pi/2.0)  border = 1
            IF((theta(ia1) < pi/2.0).AND.(theta(ia1+1) > pi/2.0))  border = 2
            IF(theta(ia1) > pi/2.0)  border = 3
            IF ((kshell(ir1) < kn).AND.(kn <= kshell(ir1+1))) THEN
              prx=X(qx,qz,qy,1)*CONJG(Y(qx,qz,qy,1))
              pry=X(qx,qz,qy,2)*CONJG(Y(qx,qz,qy,2))
              prz=X(qx,qz,qy,3)*CONJG(Y(qx,qz,qy,3))
              SELECT CASE (border)
              CASE(1)
                IF((theta(ia1) <= theta2).AND.(theta2 <  theta(ia1+1))) THEN
                  en_node(ir1,ia1)=en_node(ir1,ia1)+REAL(prx+pry+prz)*MM(qx)*MAX(ksquare(qx,qz,qy),EPS)**j
                END IF
              CASE(2)
                IF((theta(ia1) <= theta2).AND.(theta2 <= theta(ia1+1))) THEN  
                  en_node(ir1,ia1)=en_node(ir1,ia1)+REAL(prx+pry+prz)*MM(qx)*MAX(ksquare(qx,qz,qy),EPS)**j
                END IF        
              CASE(3)
                IF((theta(ia1) <  theta2).AND.(theta2 <= theta(ia1+1))) THEN
                  en_node(ir1,ia1)=en_node(ir1,ia1)+REAL(prx+pry+prz)*MM(qx)*MAX(ksquare(qx,qz,qy),EPS)**j
                END IF
              END SELECT
            END IF
          END DO
        END DO
      END DO
    END DO
  END DO
  en_node=en_node*factor*Cr**2
  CALL mpi_allreduce(en_node,en,nshell_radial*nshell_angle,MPI_MYREAL,MPI_SUM,MPI_COMM_WORLD,ierr)


END SUBROUTINE r2r_correl_vector



!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!                                                          !!
!!                     r2r_B0_transfer                      !!
!!                                                          !!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
SUBROUTINE r2r_B0_transfer(X,Y, tra)
  USE parameters
  USE mpi_var
  USE r2r_mod
  IMPLICIT NONE

  COMPLEX, DIMENSION(0:cx,0:cz,0:cy,3), INTENT(IN)  :: X  
  COMPLEX, DIMENSION(0:cx,0:cz,0:cy,3), INTENT(IN)  :: Y 
  REAL, DIMENSION(nshell_radial,nshell_angle), INTENT(OUT) :: tra
  REAL, ALLOCATABLE, DIMENSION(:,:)  :: tra_node
  COMPLEX     :: prx, pry, prz
  REAL        :: kn, theta2
  INTEGER     :: qx, qy, qz 
  INTEGER     :: ir1, ia1, border 
  
  ALLOCATE(tra_node(nshell_radial,nshell_angle))
  tra=0.0
  tra_node=0
  DO qy=0,cy
    DO qz=0,cz
      DO qx=0,cx
        kn=knorm(qx,qz,qy)
        theta2=ACOS(kz(qz)/MAX(eps,kn))
        DO ir1=1,nshell_radial
          DO ia1=1,nshell_angle
            IF(theta(ia1+1) < pi/2.0)  border = 1
            IF((theta(ia1) < pi/2.0).AND.(theta(ia1+1) > pi/2.0))  border = 2
            IF(theta(ia1) > pi/2.0)  border = 3
            IF ((kshell(ir1) < kn).AND.(kn <= kshell(ir1+1))) THEN
              prx=X(qx,qz,qy,1)*CONJG(Y(qx,qz,qy,1))
              pry=X(qx,qz,qy,2)*CONJG(Y(qx,qz,qy,2))
              prz=X(qx,qz,qy,3)*CONJG(Y(qx,qz,qy,3))
              SELECT CASE (border)
              CASE(1)
                IF((theta(ia1) <= theta2).AND.(theta2 <  theta(ia1+1))) THEN
                  tra_node(ir1,ia1)=tra_node(ir1,ia1)+AIMAG(prx+pry+prz)*MM(qx)*(b0_x*kx(qx)+b0_y*ky(qy)+b0_z*kz(qz))
                END IF
              CASE(2)
                IF((theta(ia1) <= theta2).AND.(theta2 <= theta(ia1+1))) THEN  
                  tra_node(ir1,ia1)=tra_node(ir1,ia1)+AIMAG(prx+pry+prz)*MM(qx)*(b0_x*kx(qx)+b0_y*ky(qy)+b0_z*kz(qz))
                END IF        
              CASE(3)
                IF((theta(ia1) <  theta2).AND.(theta2 <= theta(ia1+1))) THEN
                  tra_node(ir1,ia1)=tra_node(ir1,ia1)+AIMAG(prx+pry+prz)*MM(qx)*(b0_x*kx(qx)+b0_y*ky(qy)+b0_z*kz(qz))
                END IF
              END SELECT
            END IF
          END DO
        END DO
      END DO
    END DO
  END DO

  tra_node=tra_node*Cr**2
  CALL mpi_allreduce(tra_node,tra,nshell_radial*nshell_angle,MPI_MYREAL,MPI_SUM,MPI_COMM_WORLD,ierr)

END SUBROUTINE r2r_B0_transfer



!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!                                                          !!
!!                    r2r_net_ring_transfer                 !!
!!                                                          !!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
SUBROUTINE r2r_net_ring_transfer(X,Y,Z, en)
  USE parameters
  USE mpi_var
  USE r2r_mod
  IMPLICIT NONE

  COMPLEX, DIMENSION(0:cx,0:cz,0:cy,3), INTENT(IN)  :: X  
  COMPLEX, DIMENSION(0:cx,0:cz,0:cy,3), INTENT(IN)  :: Y 
  COMPLEX, DIMENSION(0:cx,0:cz,0:cy,3), INTENT(IN)  :: Z 
  REAL, DIMENSION(nshell_radial,nshell_angle), INTENT(OUT) :: en
  COMPLEX, DIMENSION(0:cx,0:cz,0:cy,3)  :: nl 
  REAL, ALLOCATABLE, DIMENSION(:,:)     :: en_node
  COMPLEX     :: prx, pry, prz
  REAL        :: kn, theta2
  INTEGER     :: qx, qy, qz 
  INTEGER     :: ir1, ia1, border 
  

  CALL nlin_vec(Y,Z,nl)
  WRITE(*,*) 'nlin_vec done'
  
  ALLOCATE(en_node(nshell_radial,nshell_angle))
  en=0.0
  en_node=0
  DO qy=0,cy
    DO qz=0,cz
      DO qx=0,cx
        kn=knorm(qx,qz,qy)
        theta2=ACOS(kz(qz)/MAX(eps,kn))
        DO ir1=1,nshell_radial
          DO ia1=1,nshell_angle
            IF(theta(ia1+1) < pi/2.0)  border = 1
            IF((theta(ia1) < pi/2.0).AND.(theta(ia1+1) > pi/2.0))  border = 2
            IF(theta(ia1) > pi/2.0)  border = 3
            IF ((kshell(ir1) < kn).AND.(kn <= kshell(ir1+1))) THEN
              prx=nl(qx,qz,qy,1)*CONJG(X(qx,qz,qy,1))
              pry=nl(qx,qz,qy,2)*CONJG(X(qx,qz,qy,2))
              prz=nl(qx,qz,qy,3)*CONJG(X(qx,qz,qy,3))
              SELECT CASE (border)
              CASE(1)
                IF((theta(ia1) <= theta2).AND.(theta2 <  theta(ia1+1))) THEN
                  en_node(ir1,ia1)=en_node(ir1,ia1)+REAL(prx+pry+prz)*MM(qx)
                END IF
              CASE(2)
                IF((theta(ia1) <= theta2).AND.(theta2 <= theta(ia1+1))) THEN  
                  en_node(ir1,ia1)=en_node(ir1,ia1)+REAL(prx+pry+prz)*MM(qx)
                END IF        
              CASE(3)
                IF((theta(ia1) <  theta2).AND.(theta2 <= theta(ia1+1))) THEN
                  en_node(ir1,ia1)=en_node(ir1,ia1)+REAL(prx+pry+prz)*MM(qx)
                END IF
              END SELECT
            END IF
          END DO
        END DO
      END DO
    END DO
  END DO
  en_node=en_node*Cr**2
  CALL mpi_allreduce(en_node,en,nshell_radial*nshell_angle,MPI_MYREAL,MPI_SUM,MPI_COMM_WORLD,ierr)

END SUBROUTINE r2r_net_ring_transfer