!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!  12 August 2011                                          !!
!!                        resize                            !!
!!                    Post-Procesing                        !!
!!                                                          !!
!!                                        version 2.005_0.10 !!
!!                                                          !!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
MODULE ch_size
  INTEGER :: nx1, nx2, ny1, ny2, nz1, nz2, RCV_ident
  INTEGER :: cxl, cyl, czl, nxl, nyl, nzl, ii
  REAL    :: Ns_ratio
END MODULE ch_size

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!                                                          !!
!!                           resize                         !!
!!                                                          !!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

PROGRAM resize
  USE parameters
  USE variables
  USE mpi_var
  USE ch_size
  IMPLICIT NONE
  
  INTEGER :: qx, qy, qz
  
  
  CALL mpi_environment_init()  
  
  CALL read_para()
  CALL read_time()
  
  CALL init_fft3d()
  CALL init_wavevectors()
  

  RCV_ident=my_inst
  
  
  nx1=nx
  ny1=ny
  nz1=nz
  CALL MPI_BCAST(nx1, 1, MPI_INTEGER, ioid,MPI_COMM_WORLD,ierr)
  CALL MPI_BCAST(ny1, 1, MPI_INTEGER, ioid,MPI_COMM_WORLD,ierr)
  CALL MPI_BCAST(nz1, 1, MPI_INTEGER, ioid,MPI_COMM_WORLD,ierr)
  nx2=nx
  ny2=ny
  nz2=nz
  CALL MPI_BCAST(nx2, 1, MPI_INTEGER, numprocs-1,MPI_COMM_WORLD,ierr)
  CALL MPI_BCAST(ny2, 1, MPI_INTEGER, numprocs-1,MPI_COMM_WORLD,ierr)
  CALL MPI_BCAST(nz2, 1, MPI_INTEGER, numprocs-1,MPI_COMM_WORLD,ierr)
  
  Ns_ratio=(nx2*ny2*nz2)/REAL(nx1*ny1*nz1)
        
  nxl=MIN(nx1,nx2)
  nyl=MIN(ny1,ny2)
  nzl=MIN(nz1,nz2)
  cxl=nxl/2
  cyl=nyl-1
  czl=nzl-1
  
  IF (VELEQ) THEN
    ALLOCATE(vel(0:cx,0:cz,0:cy,3))
    IF (RCV_ident.EQ.0) THEN
      CALL read_vec_divfree(vel,simname//'_u')  
    ELSE
      vel=0.0
    END IF
    DO ii=1,3
      CALL send_and_receive(vel(:,:,:,ii))
    END DO
    IF(RCV_ident.EQ.1) CALL write_vec_divfree(vel,simname//'_u')
  END IF

  IF (MAGEQ) THEN
    ALLOCATE(mag(0:cx,0:cz,0:cy,3))
    IF (RCV_ident.EQ.0) THEN
      CALL read_vec_divfree(mag,simname//'_b')
    ELSE
      mag=0.0
    END IF
    DO ii=1,3
      CALL send_and_receive(mag(:,:,:,ii))
    END DO
    IF(RCV_ident.EQ.1) CALL write_vec_divfree(mag,simname//'_b')
  END IF

  CALL external_stats_compute(0)
  
  IF (VELEQ) DEALLOCATE(vel)
  IF (MAGEQ) DEALLOCATE(mag)
  
  CALL mpi_environment_end()

END PROGRAM resize



!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!                                                          !!
!!                   send_and_receive                       !!
!!                                                          !!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
SUBROUTINE send_and_receive(a)
  USE parameters
  USE mpi_var
  USE ch_size
  IMPLICIT NONE
    
  COMPLEX,  DIMENSION(0:cx,0:cz,0:cy),  INTENT(INOUT)  ::  a
  COMPLEX,  DIMENSION(0:cxl,0:czl)      ::  source_slice, destination_slice
  
  INTEGER  ::  qy, qqy, pid, qj
    
  SELECT CASE(RCV_ident)
  
  CASE(0)
    DO qy=0,nyl/2
      pid=whereis_ky(qy,1)
      qqy=whereis_ky(qy,2)
      source_slice(0:cxl,0:nzl/2)    =a(0:cxl,0:nzl/2      ,qqy)
      source_slice(0:cxl,nzl/2+1:czl)=a(0:cxl,nz-nzl/2+1:cz,qqy)
      IF (io_fpart.ne.pid) THEN
        IF (my_fpart.eq.pid)      CALL MPI_SEND(source_slice     ,(cxl+1)*(czl+1),MPI_MYCOMPLEX,io_fpart,10,MPI_COMM_INSTANCE,ierr)
        IF (my_fpart.eq.io_fpart) CALL MPI_RECV(destination_slice,(cxl+1)*(czl+1),MPI_MYCOMPLEX,pid,10,MPI_COMM_INSTANCE,status_mpi,ierr)
      ELSE
        IF (my_fpart.eq.io_fpart) destination_slice=source_slice
      END IF
      IF(my_fpart.eq.io_fpart) THEN
        CALL MPI_SEND(destination_slice,(cxl+1)*(czl+1),MPI_MYCOMPLEX,numfparts,10,MPI_COMM_WORLD,ierr)
      END IF
    END DO
    DO qy=1,nyl/2-1
      pid=whereis_ky(ny-qy,1)
      qqy=whereis_ky(ny-qy,2)
      source_slice(:,0:nzl/2)    =a(0:cxl,0:nzl/2      ,qqy)
      source_slice(:,nzl/2+1:czl)=a(0:cxl,nz-nzl/2+1:cz,qqy)
      IF (io_fpart.ne.pid) THEN
        IF (my_fpart.eq.pid)      CALL MPI_SEND(source_slice     ,(cxl+1)*(czl+1),MPI_MYCOMPLEX,io_fpart,10,MPI_COMM_INSTANCE,ierr)
        IF (my_fpart.eq.io_fpart) CALL MPI_RECV(destination_slice,(cxl+1)*(czl+1),MPI_MYCOMPLEX,pid,10,MPI_COMM_INSTANCE,status_mpi,ierr)
      ELSE
        IF (my_fpart.eq.io_fpart) destination_slice=source_slice
      END IF
      IF(my_fpart.eq.io_fpart) THEN
        CALL MPI_SEND(destination_slice,(cxl+1)*(czl+1),MPI_MYCOMPLEX,numfparts,10,MPI_COMM_WORLD,ierr)
      END IF
    END DO

  CASE(1)
    DO qy=0,nyl/2  
      IF(my_fpart.eq.io_fpart)THEN 
        CALL MPI_RECV(source_slice,(cxl+1)*(czl+1),MPI_MYCOMPLEX,ioid,10,MPI_COMM_WORLD,status_mpi,ierr)
      END IF
      pid=whereis_ky(qy,1)
      qqy=whereis_ky(qy,2)
      IF (io_fpart.ne.pid) THEN
        IF (my_fpart.eq.pid)      CALL MPI_RECV(destination_slice,(cxl+1)*(czl+1),MPI_MYCOMPLEX,io_fpart,10,MPI_COMM_INSTANCE,status_mpi,ierr)
        IF (my_fpart.eq.io_fpart) CALL MPI_SEND(source_slice     ,(cxl+1)*(czl+1),MPI_MYCOMPLEX,pid,10,MPI_COMM_INSTANCE,ierr)
      ELSE
        IF (my_fpart.eq.io_fpart) destination_slice=source_slice
      END IF
      IF (my_fpart.eq.pid) THEN
        a(0:cxl,0:nzl/2,qqy)=destination_slice(:,0:nzl/2)
        a(0:cxl,nz-nzl/2+1:cz,qqy)=destination_slice(:,nzl/2+1:czl)
      END IF
    END DO
    DO qy=1,nyl/2-1
      IF(my_fpart.eq.io_fpart)THEN
        CALL MPI_RECV(source_slice,(cxl+1)*(czl+1),MPI_MYCOMPLEX,ioid,10,MPI_COMM_WORLD,status_mpi,ierr)
      END IF
      pid=whereis_ky(ny-qy,1)
      qqy=whereis_ky(ny-qy,2)
      IF (io_fpart.ne.pid) THEN
        IF (my_fpart.eq.pid)      CALL MPI_RECV(destination_slice,(cxl+1)*(czl+1),MPI_MYCOMPLEX,io_fpart,10,MPI_COMM_INSTANCE,status_mpi,ierr)
        IF (my_fpart.eq.io_fpart) CALL MPI_SEND(source_slice     ,(cxl+1)*(czl+1),MPI_MYCOMPLEX,pid,10,MPI_COMM_INSTANCE,ierr)
      ELSE
        IF (my_fpart.eq.io_fpart) destination_slice=source_slice
      END IF
      IF (my_fpart.eq.pid) THEN
        a(0:cxl,0:nzl/2,qqy)=destination_slice(:,0:nzl/2)
        a(0:cxl,nz-nzl/2+1:cz,qqy)=destination_slice(:,nzl/2+1:czl)
      END IF
    END DO
    a=Ns_ratio*a
  END SELECT
  
END SUBROUTINE send_and_receive