!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!  12 August 2011                                          !!
!!                       s2d                                !!
!!         Single to Double fields conversion               !!
!!                                                          !!
!! s2d                                                      !!
!! single2double                                            !!
!! write_vec_divfree_sgl                                    !!
!! write_vec_divfree_dbl                                    !!
!!                                          version 2.05_01 !!
!!                                                          !!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!



!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!                                                          !!
!!                            s2d                           !!
!!                                                          !!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
PROGRAM s2d
  USE parameters
  USE variables
  USE mpi_var
  IMPLICIT NONE  

  INTEGER                 :: rc
  CHARACTER(LEN=3)        :: cnum
!!! Initializations

  CALL mpi_environment_init()    

  IF (my_fpart.eq.io_fpart) THEN
    WRITE(out_unit,*) '>> + + + Single to Double + + +'  
    WRITE(out_unit,*) '>> Simulation name is: ', simname  
  END IF

  CALL read_para()  
  CALL read_time()
  CALL init_fft3d()
  CALL init_wavevectors()
  CALL single2double()

  IF (my_fpart.eq.io_fpart) THEN
    WRITE(cnum,'(i3.3)') nfields
    WRITE(out_unit,*) '>> The transformation has been made for fields number  ', nfields  
  END IF
  
  CALL mpi_environment_end()

END PROGRAM s2d



!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!                                                          !!
!!                     single2double                        !!
!!                                                          !!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
SUBROUTINE single2double()
  USE parameters
  USE mpi_var  

  COMPLEX, ALLOCATABLE, DIMENSION(:,:,:,:)      ::  field
  COMPLEX(4), ALLOCATABLE, DIMENSION(:,:,:,:)   ::  field_sgl
  COMPLEX(8), ALLOCATABLE, DIMENSION(:,:,:,:)   ::  field_dbl
  
  ALLOCATE(field(0:cx,0:cz,0:cy,3))
  ALLOCATE(field_sgl(0:cx,0:cz,0:cy,3))
  ALLOCATE(field_dbl(0:cx,0:cz,0:cy,3))
  
  field=0.0
  IF (VELEQ) THEN
    CALL read_vec_divfree(field,simname//'_u')
    field_sgl=field
    CALL write_vec_divfree_sgl(field_sgl,simname//'_u') 
    field_dbl=field    
    CALL write_vec_divfree_dbl(field_dbl,simname//'_u')
    WRITE(out_unit,*) '>> Velocity field done!'
  END IF
  
  field=0.0  
  IF (MAGEQ) THEN
    CALL read_vec_divfree(field,simname//'_b')
    field_sgl=field
    CALL write_vec_divfree_sgl(field_sgl,simname//'_b')  
    field_dbl=field
    CALL write_vec_divfree_dbl(field_dbl,simname//'_b')
    WRITE(out_unit,*) '>> Magnetic field done!'    
  END IF
  
  DEALLOCATE(field)
  DEALLOCATE(field_sgl)
  DEALLOCATE(field_dbl)
  
END SUBROUTINE single2double



!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!                                                          !!
!!                  write_vec_divfree_sgl                   !!
!!                                                          !!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
SUBROUTINE write_vec_divfree_sgl(u,beginname)
  USE parameters
  USE mpi_var
  IMPLICIT NONE
 
  COMPLEX(4), DIMENSION(0:cx,0:cz,0:cy,3),INTENT(IN) :: u
  CHARACTER(LEN=namelength+2), INTENT(IN)            :: beginname
  CHARACTER(LEN=namelength+11)                       :: filename  
  INTEGER                                            :: ipart, reccount, reclength
  CHARACTER(LEN=3)                                   :: cnum
  CHARACTER(LEN=2)                                   :: partnum, inum
  COMPLEX(4), ALLOCATABLE, DIMENSION(:,:)            :: utempx, utempz
  INTEGER                                            :: qy, pid, qqy

  WRITE(cnum,'(i3.3)') nfields
  WRITE(inum,'(i2.2)') my_inst
  reclength=2*(cx+1)*(cz+1)*4
  ALLOCATE(utempx(0:cx,0:cz),utempz(0:cx,0:cz))

  DO ipart=0,totpart-1
    IF (my_fpart.eq.io_fpart) THEN
      WRITE(partnum,'(i2.2)') ipart
      IF (numinsts.gt.1) THEN
        filename=inum//'_'//beginname//partnum//'.'//cnum
      ELSE
        filename=beginname//partnum//'.'//cnum
      END IF
      OPEN (UNIT=10, FILE=filename//'_Sgl', STATUS='UNKNOWN', ACTION='WRITE',ACCESS='DIRECT',&
          FORM='UNFORMATTED', RecL=reclength)
      reccount=1
    END IF
    
    IF (ipart.eq.0) THEN
      pid=whereis_ky(0,1)
      qqy=whereis_ky(0,2)
      CALL transfer_slice_sgl(u(:,:,qqy,2),pid,utempx,ioid)
      IF (my_fpart.eq.io_fpart) THEN
        WRITE(10,rec=reccount) utempx
        reccount=2
      END IF
    END IF

    DO qy=ipart*ny/totpart,(ipart+1)*ny/totpart-1
      pid=whereis_ky(qy,1)
      qqy=whereis_ky(qy,2)
      CALL transfer_slice_sgl(u(:,:,qqy,1),pid,utempx,ioid)
      CALL transfer_slice_sgl(u(:,:,qqy,3),pid,utempz,ioid)
      IF (my_fpart.eq.io_fpart) THEN
        WRITE(10,rec=reccount) utempx
        WRITE(10,rec=reccount+1) utempz
        reccount=reccount+2
      END IF
    END DO  

    IF (my_fpart.eq.io_fpart) CLOSE(10)
  END DO ! ipart loop

  DEALLOCATE(utempx,utempz)

END SUBROUTINE write_vec_divfree_sgl



!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!                                                          !!
!!                 write_vec_divfree_dbl                    !!
!!                                                          !!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
SUBROUTINE write_vec_divfree_dbl(u,beginname)
  USE parameters
  USE mpi_var
  IMPLICIT NONE

  COMPLEX(8), DIMENSION(0:cx,0:cz,0:cy,3),INTENT(IN) :: u
  CHARACTER(LEN=namelength+2), INTENT(IN)            :: beginname
  CHARACTER(LEN=namelength+11)                       :: filename  
  INTEGER                                            :: ipart, reccount, reclength
  CHARACTER(LEN=3)                                   :: cnum
  CHARACTER(LEN=2)                                   :: partnum, inum
  COMPLEX(8), ALLOCATABLE, DIMENSION(:,:)            :: utempx, utempz
  INTEGER                                            :: qy, pid, qqy

  WRITE(cnum,'(i3.3)') nfields
  WRITE(inum,'(i2.2)') my_inst
  reclength=2*(cx+1)*(cz+1)*8
  ALLOCATE(utempx(0:cx,0:cz),utempz(0:cx,0:cz))

  DO ipart=0,totpart-1
    IF (my_fpart.eq.io_fpart) THEN
      WRITE(partnum,'(i2.2)') ipart
      IF (numinsts.gt.1) THEN
        filename=inum//'_'//beginname//partnum//'.'//cnum
      ELSE
        filename=beginname//partnum//'.'//cnum
      END IF
      OPEN (UNIT=10, FILE=filename//'_Dbl', STATUS='UNKNOWN', ACTION='WRITE',ACCESS='DIRECT',&
          FORM='UNFORMATTED', RecL=reclength)
      reccount=1
    END IF
    
    IF (ipart.eq.0) THEN
      pid=whereis_ky(0,1)
      qqy=whereis_ky(0,2)
      CALL transfer_slice_dbl(u(:,:,qqy,2),pid,utempx,ioid)
      IF (my_fpart.eq.io_fpart) THEN
        WRITE(10,rec=reccount) utempx
        reccount=2
      END IF
    END IF

    DO qy=ipart*ny/totpart,(ipart+1)*ny/totpart-1
      pid=whereis_ky(qy,1)
      qqy=whereis_ky(qy,2)
      CALL transfer_slice_dbl(u(:,:,qqy,1),pid,utempx,ioid)
      CALL transfer_slice_dbl(u(:,:,qqy,3),pid,utempz,ioid)
      IF (my_fpart.eq.io_fpart) THEN
        WRITE(10,rec=reccount) utempx
        WRITE(10,rec=reccount+1) utempz
        reccount=reccount+2
      END IF
    END DO  

    IF (my_fpart.eq.io_fpart) CLOSE(10)
  END DO ! ipart loop

  DEALLOCATE(utempx,utempz)

END SUBROUTINE write_vec_divfree_dbl



!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!                                                          !!
!!                     transfer_slice_sgl                   !!
!!                                                          !!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
SUBROUTINE transfer_slice_sgl(slice,fromid,destination,destid)
  USE parameters
  USE mpi_var
  IMPLICIT NONE

  COMPLEX(4), DIMENSION(0:cx,0:cz), INTENT(IN)    :: slice
  COMPLEX(4), DIMENSION(0:cx,0:cz), INTENT(OUT)   :: destination
  INTEGER, INTENT(IN)                          :: fromid, destid

  IF (destid.ne.fromid) THEN
    IF (my_fpart.eq.fromid) CALL MPI_SEND(slice,(cx+1)*(cz+1),MPI_COMPLEX,destid,10,MPI_COMM_INSTANCE,ierr)
    IF (my_fpart.eq.destid) CALL MPI_RECV(destination,(cx+1)*(cz+1),MPI_COMPLEX,fromid,10,MPI_COMM_INSTANCE,status_mpi,ierr)
  ELSE
    IF (my_fpart.eq.destid) destination=slice
  END IF

END SUBROUTINE transfer_slice_sgl



!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!                                                          !!
!!                     transfer_slice_dbl                   !!
!!                                                          !!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
SUBROUTINE transfer_slice_dbl(slice,fromid,destination,destid)
  USE parameters
  USE mpi_var
  IMPLICIT NONE

  COMPLEX(8), DIMENSION(0:cx,0:cz), INTENT(IN)    :: slice
  COMPLEX(8), DIMENSION(0:cx,0:cz), INTENT(OUT)   :: destination
  INTEGER, INTENT(IN)                             :: fromid, destid

  IF (destid.ne.fromid) THEN
    IF (my_fpart.eq.fromid) CALL MPI_SEND(slice,(cx+1)*(cz+1),MPI_DOUBLE_COMPLEX,destid,10,MPI_COMM_INSTANCE,ierr)
    IF (my_fpart.eq.destid) CALL MPI_RECV(destination,(cx+1)*(cz+1),MPI_DOUBLE_COMPLEX,fromid,10,MPI_COMM_INSTANCE,status_mpi,ierr)
  ELSE
    IF (my_fpart.eq.destid) destination=slice
  END IF

END SUBROUTINE transfer_slice_dbl  

