!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!  12 August 2011                                          !!
!!                      tt_fft.90                           !!
!!                                                          !!
!!  fft3ds                                                  !!
!!  ifft3ds                                                 !!
!!  init_fft3d                                              !!
!!                                             version 2.05 !!
!!                                                          !!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!



!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!                                                          !!
!!                         fft3ds                           !!
!!                                                          !!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
SUBROUTINE fft3ds(freal,fwave)
  USE parameters
  USE mpi_var
  USE fftwmodule
  IMPLICIT NONE

  REAL, INTENT(IN)       :: freal(0:rx,0:ry,0:rz)
  REAL, INTENT(OUT)      :: fwave(0:rx,0:ry,0:rz)
  REAL, ALLOCATABLE, DIMENSION(:)      :: work
  
  ALLOCATE(work(0:local_nr-1))
  fwave=freal*Ck
  CALL rfftwnd_f77_mpi(plan_RtoC,1,fwave,work,1,FFTW_TRANSPOSED_ORDER)
        
  DEALLOCATE(work)
  
END SUBROUTINE fft3ds      



!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!                                                          !!
!!                         ifft3ds                          !!
!!                                                          !!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
SUBROUTINE ifft3ds(fwave,freal)
  USE parameters
  USE mpi_var
  USE fftwmodule
  IMPLICIT NONE

  REAL, INTENT(IN)        :: fwave(0:rx,0:ry,0:rz)
  REAL, INTENT(OUT)       :: freal(0:rx,0:ry,0:rz)    
  REAL, ALLOCATABLE, DIMENSION(:)      :: work
  
  ALLOCATE(work(0:local_nr-1))  
  freal=fwave*Cr
  CALL rfftwnd_f77_mpi(plan_CtoR,1,freal,work,1,FFTW_TRANSPOSED_ORDER)
  freal(nx:nx+1,:,:)=0.0
  
  DEALLOCATE(work)

END SUBROUTINE ifft3ds
      


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!                                                          !!
!!                       init_fft3d                         !!
!!                                                          !!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
SUBROUTINE init_fft3d()
  USE parameters
  USE mpi_var
  USE fftwmodule
  IMPLICIT NONE

  INTEGER   :: n_dim(3)

  Ns=nx*ny*nz
  Ck=1.0
  Cr=1.0/Ns  !! The value of Cr in the fftw library is 1. We use this value in order to have fft(ifft(u))=u (Cr*Ck*Ns=1)
  
  n_dim(1)=nx
  n_dim(2)=ny
  n_dim(3)=nz

  CALL rfftwnd_f77_mpi_create_plan(plan_RtoC,MPI_COMM_INSTANCE, 3, n_dim,FFTW_REAL_TO_COMPLEX,FFTW_ESTIMATE)
  CALL rfftwnd_f77_mpi_create_plan(plan_CtoR,MPI_COMM_INSTANCE, 3, n_dim,FFTW_COMPLEX_TO_REAL,FFTW_ESTIMATE)
  CALL rfftwnd_f77_mpi_local_sizes(plan_RtoC,local_nz,local_z_start, local_ny_after_transpose,local_y_start_after_transpose,local_nr)

  WRITE (6,*) '>> FFTW plans are computed.', myid, my_inst, my_fpart, local_nr 
  
  cx=nx/2
  cy=local_ny_after_transpose-1
  cz=nz-1
  rx=nx+1
  ry=ny-1
  rz=local_nz-1
  
  WRITE (6,*) '>> Processor number: ', myid, ' instance: ', my_inst
  WRITE (6,*) '>>         local_nz: ', rz+1,' local_z_start: ', local_z_start
  
  
  IF ((cy+1)*(cz+1).ne.(ry+1)*(rz+1)) THEN
    WRITE(6,*) 'This code assumes that on each node the product of the last'
    WRITE(6,*) 'two dimensions of real and complex fields are the same. This'
    WRITE(6,*) 'is not achievable with the choice ny=', ny,', nz=', nz,' and ', numprocs,' processors.'
    WRITE(6,*) 'Please retry with another number of processors.'
    STOP    
  END IF
  
END SUBROUTINE init_fft3d

