!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!  12 August 2011                                          !!
!!                      tt_force.f90                        !!
!!                                                          !!
!!  forcing                                                 !!
!!  forcing_ub                                              !!
!!  modify_vel                                              !!
!!  helical_forcing                                         !!
!!                                             version 2.05 !!
!!                                                          !!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!



!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!                                                          !!
!!                          forcing                         !!
!!                                                          !!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
SUBROUTINE forcing(force)
  USE parameters
  USE variables
  USE mpi_var
  IMPLICIT NONE

  COMPLEX, DIMENSION(0:cx,0:cz,0:cy,3), INTENT(OUT)   :: force
  INTEGER :: qx, qy, qz, qqy
  COMPLEX :: ox, oy, oz, ux, uy, uz
  REAL    :: kn, ener, enst, heli, alph, beta, alph_max, beta_max, sk

  force=0.0  
  alph_max=0.0
  beta_max=0.0
  
  SELECT CASE(FORCEPARA)
  CASE(1)      !!!!! Constant energy and helicity injection in a shell  
    DO qy=0, cy
      DO qz=0, cz
        DO qx=0, cx
          kn=knorm(qx,qz,qy)
          IF ((kn.lt.ksup).AND.(kn.ge.kinf)) THEN 
            ux=vel(qx,qz,qy,1)
            uy=vel(qx,qz,qy,2)
            uz=vel(qx,qz,qy,3)
            ox=iky(qy)*uz-ikz(qz)*uy
            oy=ikz(qz)*ux-ikx(qx)*uz
            oz=ikx(qx)*uy-iky(qy)*ux
            ener=REAL((ux*CONJG(ux)+uy*CONJG(uy)+uz*CONJG(uz)))/2.0
            heli=REAL((ux*CONJG(ox)+uy*CONJG(oy)+uz*CONJG(oz)))       
            sk=heli/(2.0*kn*ener)
            alph=(energy_forcing*(1.0-sk*helicity_para))/(2.0*ener*(1.0-sk**2)) 
            beta=(energy_forcing*(helicity_para-sk))/(2.0*kn*ener*(1.0-sk**2))
            alph_max = MAX(alph_max,ABS(alph))
            beta_max = MAX(beta_max,ABS(beta))             
            force(qx,qz,qy,1)=alph*ux+beta*ox
            force(qx,qz,qy,2)=alph*uy+beta*oy
            force(qx,qz,qy,3)=alph*uz+beta*oz
          END IF
        END DO
      END DO
    END DO 
    force=force*(Ck*Ns)**2/modes_layer
    alph_max=alph_max*(Ck*Ns)**2/modes_layer
    beta_max=beta_max*(Ck*Ns)**2/modes_layer
    dt_f=MIN(1.0/MAX(alph_max,EPS), 1.0/MAX(beta_max,EPS)*pi/kM)
  CASE(3)      !!!!! Forcing provided by the user
    CALL external_forcing(force)
  CASE(4)      !!!!! Kolmogorov with constant force
    DO qy=0,cy
      qqy=qy+local_y_start_after_transpose
      IF (qqy.eq.Fmode)    force(0,0,qy,1)=-Famp*I*Ns/2.0
      IF (qqy.eq.ny-Fmode) force(0,0,qy,1)=+Famp*I*Ns/2.0
    END DO
  CASE(6)      !!!!! Helical force in a shell 
    CALL helical_forcing(force)  
  CASE DEFAULT
  END SELECT

  CALL divfree(force)

END SUBROUTINE forcing



!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!                                                          !!
!!                       forcing_ub                         !!
!!                                                          !!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
SUBROUTINE forcing_ub(force_u, force_b)
  USE parameters
  USE variables
  IMPLICIT NONE

  COMPLEX, DIMENSION(0:cx,0:cz,0:cy,3), INTENT(OUT)   :: force_u
  COMPLEX, DIMENSION(0:cx,0:cz,0:cy,3), INTENT(OUT)   :: force_b
  COMPLEX, DIMENSION(3)                  :: h_plus, h_minus
  COMPLEX                                :: up, um, bp, bm, fup, fum, fbp, fbm
  REAL                                   :: zx, zy, zz, zn
  REAL                                   :: vx, vy, vz, vn, kn
  REAL                                   :: Eup, Eum, Ebp, Ebm, Hcp, Hcm
  REAL                                   :: eeup, eeum, eebp, eebm, ccup, ccum, ccbp, ccbm 
  REAL                                   :: alph_up, alph_um, beta_up, beta_um, factor_p
  REAL                                   :: alph_bp, alph_bm, beta_bp, beta_bm, factor_m 
  INTEGER                                :: qx, qy, qz

  force_u=0.0  
  force_b=0.0
  dt_f = 0.0
   
  SELECT CASE(FORCEPARA)
  CASE(7)      !!!!! Helical force in a shell width cross-helicity, magnetic helicity, kinetic helicity
    eeup = (1.0+kin_hel_forcing)*kin_energy_forcing/2.0
    eeum = (1.0-kin_hel_forcing)*kin_energy_forcing/2.0
    eebp = (1.0+mag_hel_forcing)*mag_energy_forcing/2.0
    eebm = (1.0-mag_hel_forcing)*mag_energy_forcing/2.0
    ccup = eeup*croshel_forcing 
    ccum = eeum*croshel_forcing 
    ccbp = eebp*croshel_forcing 
    ccbm = eebm*croshel_forcing 
   
    !! Define the arbitrary vector Z not ortogonal to any wave-vector k
    zx=1.0
    zy=1.0
    zz=REAL(nz)
 
    DO qy=0, cy
      DO qz=0, cz
        DO qx=0, cx
          kn=knorm(qx,qz,qy)
          IF ((kn.lt.ksup).AND.(kn.ge.kinf)) THEN 
		    !! Compute Z x k
            vx=zy*kz(qz)-zz*ky(qy)  
            vy=zz*kx(qx)-zx*kz(qz)		            
            vz=zx*ky(qy)-zy*kx(qx)
            vn=SQRT(vx**2+vy**2+vz**2)
	   
		    !! Compute the helical basis h
            h_plus(1) =( (vy*kz(qz)-vz*ky(qy))/kn + I*vx )/vn
            h_plus(2) =( (vz*kx(qx)-vx*kz(qz))/kn + I*vy )/vn
            h_plus(3) =( (vx*ky(qy)-vy*kx(qx))/kn + I*vz )/vn
            h_minus(1)=( (vy*kz(qz)-vz*ky(qy))/kn - I*vx )/vn
            h_minus(2)=( (vz*kx(qx)-vx*kz(qz))/kn - I*vy )/vn
            h_minus(3)=( (vx*ky(qy)-vy*kx(qx))/kn - I*vz )/vn

            !! Decompose the velocity and magnetic fields into helical components          
            up=0.5*(vel(qx,qz,qy,1)*CONJG(h_plus(1)) +vel(qx,qz,qy,2)*CONJG(h_plus(2)) +vel(qx,qz,qy,3)*CONJG(h_plus(3)) )
            um=0.5*(vel(qx,qz,qy,1)*CONJG(h_minus(1))+vel(qx,qz,qy,2)*CONJG(h_minus(2))+vel(qx,qz,qy,3)*CONJG(h_minus(3)))
            bp=0.5*(mag(qx,qz,qy,1)*CONJG(h_plus(1)) +mag(qx,qz,qy,2)*CONJG(h_plus(2)) +mag(qx,qz,qy,3)*CONJG(h_plus(3)) )
            bm=0.5*(mag(qx,qz,qy,1)*CONJG(h_minus(1))+mag(qx,qz,qy,2)*CONJG(h_minus(2))+mag(qx,qz,qy,3)*CONJG(h_minus(3)))
		  
            Eup=REAL(up*CONJG(up))/2.0
            Eum=REAL(um*CONJG(um))/2.0
            Ebp=REAL(bp*CONJG(bp))/2.0
            Ebm=REAL(bm*CONJG(bm))/2.0
            Hcp=REAL(up*CONJG(bp)+bp*CONJG(up))/2.0
            Hcm=REAL(um*CONJG(bm)+bm*CONJG(um))/2.0

            factor_p=1.0/(4*Eup*Ebp-Hcp**2) 
            alph_up = factor_p*(2*Ebp*eeup-Hcp*ccup)
            beta_up = factor_p*(2*Eup*ccup-Hcp*eeup)
            alph_bp = factor_p*(2*Ebp*ccbp-Hcp*eebp)  
            beta_bp = factor_p*(2*Eup*eebp-Hcp*ccbp) 

            factor_m=1.0/(4*Eum*Ebm-Hcm**2) 
            alph_um = factor_m*(2*Ebm*eeum-Hcm*ccum)
            beta_um = factor_m*(2*Eum*ccum-Hcm*eeum)
            alph_bm = factor_m*(2*Ebm*ccbm-Hcm*eebm)  
            beta_bm = factor_m*(2*Eum*eebm-Hcm*ccbm) 

            dt_f = MAX(dt_f,ABS(alph_up+alph_um),ABS(alph_bp+alph_bm))
            dt_f = MAX(dt_f,ABS(beta_up+beta_um),ABS(beta_bp+beta_bm))

            fup=0.5*(alph_up*up+beta_up*bp)
            fbp=0.5*(alph_bp*up+beta_bp*bp)
            fum=0.5*(alph_um*um+beta_um*bm)
            fbm=0.5*(alph_bm*um+beta_bm*bm)

		    !! Compute the force 
            force_u(qx,qz,qy,1)=fup*h_plus(1)+fum*h_minus(1)
            force_u(qx,qz,qy,2)=fup*h_plus(2)+fum*h_minus(2)
            force_u(qx,qz,qy,3)=fup*h_plus(3)+fum*h_minus(3)

            force_b(qx,qz,qy,1)=fbp*h_plus(1)+fbm*h_minus(1)
            force_b(qx,qz,qy,2)=fbp*h_plus(2)+fbm*h_minus(2)
            force_b(qx,qz,qy,3)=fbp*h_plus(3)+fbm*h_minus(3)
		  END IF
        END DO
      END DO
    END DO 
  CASE DEFAULT
  END SELECT

  dt_f = dt_f*(Ck*Ns)**2/modes_layer
  dt_f = 1.0/MAX(dt_f,EPS)

  force_u=force_u*(Ck*Ns)**2/modes_layer
  force_b=force_b*(Ck*Ns)**2/modes_layer
  CALL divfree(force_u)
  CALL divfree(force_b)

END SUBROUTINE forcing_ub



!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!                                                          !!
!!                      modify_vel                          !!
!!                                                          !!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
SUBROUTINE modify_vel(u,delta_energy,delta_helicity)
  USE parameters
  IMPLICIT NONE

  COMPLEX, DIMENSION(0:cx,0:cz,0:cy,3), INTENT(INOUT)    :: u
  REAL, INTENT(INOUT)    ::   delta_energy,delta_helicity
  REAL    :: ener_before, heli_before, ener_after, heli_after
  INTEGER :: qx, qy, qz
  COMPLEX :: ox, oy, oz, ux, uy, uz
  REAL    :: kn, ener, heli, sk, alph, beta
 
  CALL correl_vector(u,u,0,0.5,ener_before)
  CALL correl_helicity(u,u,0,1.0,heli_before)

  SELECT CASE(FORCEPARA)
  CASE(2)      !!!!! fixed energy and helicity level
    DO qy=0, cy
      DO qz=0, cz
        DO qx=0, cx
          kn=knorm(qx,qz,qy)
          IF ((kn.lt.ksup).AND.(kn.ge.kinf)) THEN 
            ux=u(qx,qz,qy,1)
            uy=u(qx,qz,qy,2)
            uz=u(qx,qz,qy,3)
            ox=iky(qy)*uz-ikz(qz)*uy
            oy=ikz(qz)*ux-ikx(qx)*uz
            oz=ikx(qx)*uy-iky(qy)*ux
            ener=REAL((ux*CONJG(ux)+uy*CONJG(uy)+uz*CONJG(uz)))/(2.0*(Ck*Ns)**2)
            heli=REAL((ux*CONJG(ox)+uy*CONJG(oy)+uz*CONJG(oz)))/(Ck*Ns)**2
            sk=heli/(2.0*kn*ener) 
            IF(sk.eq.1)  sk=sk-0.0000001
            IF(sk.eq.-1) sk=sk+0.0000001
            alph=SQRT(energy_forcing*((1.0-helicity_para*sk)+SQRT((1.0-helicity_para**2)*(1.0-sk**2)))/(2.0*ener*(1.0-sk**2))) 
            beta=energy_forcing*(helicity_para-sk)/(2.0*kn*ener*(1.0-sk**2)*alph)
            u(qx,qz,qy,1)=alph*ux+beta*ox
            u(qx,qz,qy,2)=alph*uy+beta*oy
            u(qx,qz,qy,3)=alph*uz+beta*oz
          END IF
        END DO
      END DO
    END DO 
  CASE(5)      !!!!! velocity profile provided by the user
    CALL external_modif_vel(u)
  CASE DEFAULT
  END SELECT
  
  CALL correl_vector(u,u,0,0.5,ener_after)
  CALL correl_helicity(u,u,0,1.0,heli_after)
  delta_energy   =  delta_energy   + ener_after-ener_before
  delta_helicity =  delta_helicity + heli_after-heli_before

END SUBROUTINE modify_vel



!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!                                                          !!
!!                     helical_forcing                      !!
!!                                                          !!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
SUBROUTINE helical_forcing(force)
  USE parameters
  USE variables
  IMPLICIT NONE

  COMPLEX, DIMENSION(0:cx,0:cz,0:cy,3), INTENT(OUT)    :: force
  COMPLEX, DIMENSION(3)     :: h_plus, h_minus
  COMPLEX                   :: up, um, fp, fm
  REAL                      :: zx, zy, zz, zn, kn
  REAL                      :: vx, vy, vz, vn
  REAL                      :: eps_p, eps_m, phi_p, phi_m
  INTEGER                   :: qx, qy, qz 

  force=0.0

  eps_p = (1.0+helicity_para)*energy_forcing/2.0
  eps_m = (1.0-helicity_para)*energy_forcing/2.0

  !! Define the arbitrary vector Z not ortogonal to any wave-vector k
  zx=1.0
  zy=1.0
  zz=REAL(nz)
 
  DO qy=0, cy
    DO qz=0, cz
      DO qx=0, cx
        kn=knorm(qx,qz,qy)
        IF ((kn.lt.ksup).AND.(kn.ge.kinf)) THEN 
		  !! Compute Z x k
          vx=zy*kz(qz)-zz*ky(qy)  
          vy=zz*kx(qx)-zx*kz(qz)		            
          vz=zx*ky(qy)-zy*kx(qx)
          vn=SQRT(vx**2+vy**2+vz**2)
	 
		  !! Compute the helical basis h
          h_plus(1) =( (vy*kz(qz)-vz*ky(qy))/kn + I*vx )/vn
          h_plus(2) =( (vz*kx(qx)-vx*kz(qz))/kn + I*vy )/vn
          h_plus(3) =( (vx*ky(qy)-vy*kx(qx))/kn + I*vz )/vn
          h_minus(1)=( (vy*kz(qz)-vz*ky(qy))/kn - I*vx )/vn
          h_minus(2)=( (vz*kx(qx)-vx*kz(qz))/kn - I*vy )/vn
          h_minus(3)=( (vx*ky(qy)-vy*kx(qx))/kn - I*vz )/vn

          !! Decompose the velocity into helical components          
          up=0.5*(vel(qx,qz,qy,1)*CONJG(h_plus(1)) +vel(qx,qz,qy,2)*CONJG(h_plus(2)) +vel(qx,qz,qy,3)*CONJG(h_plus(3)) )
          um=0.5*(vel(qx,qz,qy,1)*CONJG(h_minus(1))+vel(qx,qz,qy,2)*CONJG(h_minus(2))+vel(qx,qz,qy,3)*CONJG(h_minus(3)))
	
          !! Random phases for the force
		  CALL RANDOM_NUMBER(phi_p)
		  CALL RANDOM_NUMBER(phi_m)
          phi_p=0.0
          phi_m=0.0 
		    
          !! Compute the helical components of the force 
		  fp=0.5*eps_p*EXP(I*(phi_p))*up/(COS(phi_p)*CABS(up)**2)
          fm=0.5*eps_m*EXP(I*(phi_m))*um/(COS(phi_m)*CABS(um)**2)		  		 

		  !! Compute the force 
          force(qx,qz,qy,1)=fp*h_plus(1)+fm*h_minus(1)
          force(qx,qz,qy,2)=fp*h_plus(2)+fm*h_minus(2)
          force(qx,qz,qy,3)=fp*h_plus(3)+fm*h_minus(3)
        END IF
      END DO
    END DO
  END DO 
  
  force=force*(Ck*Ns)**2/modes_layer
   
END SUBROUTINE helical_forcing


