!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!  12 August 2011                                          !!
!!                      tt_init.f90                         !!
!!                                                          !!
!!  mpi_environment_init                                    !!
!!  mpi_environment_end                                     !!
!!  init_variables                                          !!
!!  init_random_seed                                        !!
!!  init_wavevectors                                        !!
!!  init_vel_field                                          !!
!!  init_mag_field                                          !!
!!  init_sca_field                                          !!
!!  init_TaylorGreen                                        !!
!!  init_random_vec_divfree                                 !!
!!  init_random_scal                                        !!
!!  buildphases                                             !!
!!  symmetrize_all                                          !!
!!  symmetrize_one                                          !!
!!  divfree                                                 !!
!!  zero_nhalf                                              !!
!!                                             version 2.05 !!
!!                                                          !!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!



!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!                                                          !!
!!                  mpi_environment_init                    !!
!!                                                          !!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
SUBROUTINE mpi_environment_init()
  USE parameters
  USE mpi_var
  IMPLICIT NONE

  INTEGER                            :: rc
  INTEGER                            :: total_processors,numfparts_max 
  INTEGER                            :: inst_counter, fpart_counter, temp_counter
  CHARACTER(LEN=2)                   :: inum


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!! MPI Initiation
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  CALL MPI_INIT( ierr )
  CALL MPI_COMM_RANK( MPI_COMM_WORLD, myid, ierr )
  CALL MPI_COMM_SIZE( MPI_COMM_WORLD, numprocs, ierr )


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!! Read and broadcast the ensemble parameters.
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  instance_name='test00'
  fparts=1
  IF (myid.eq.ioid) THEN 
    OPEN(10,FILE='turbo.par',STATUS='OLD',FORM='FORMATTED')
    READ(unit=10,NML=ensemble)
    CLOSE(10)
  END IF
  CALL MPI_BCAST(numinsts  ,1 ,MPI_INTEGER,ioid,MPI_COMM_WORLD,ierr)
  CALL MPI_BCAST(same_numfp,1 ,MPI_INTEGER,ioid,MPI_COMM_WORLD,ierr) 
  CALL MPI_BCAST(dt_global ,1 ,MPI_INTEGER,ioid,MPI_COMM_WORLD,ierr)
  CALL MPI_BCAST(instance_name,namelength*100,MPI_CHARACTER,ioid,MPI_COMM_WORLD,ierr)
  CALL MPI_BCAST(fparts,100,MPI_INTEGER,ioid,MPI_COMM_WORLD,ierr)


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!! Check the validity of the simulation
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  IF (numinsts.lt.1) THEN
    IF (myid.eq.ioid) WRITE(6,*) 'numinsts has to be larger than 0. Stopping!'
    STOP
  END IF
  IF (numinsts.gt.numprocs) THEN
    IF (myid.eq.ioid) WRITE(6,*) 'numprocs has to be larger than or equal to numinsts. Stopping!'
    STOP
  END IF


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!! Ensemble MPI parameters initiation
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  IF(same_numfp.eq.1) THEN
    fparts(0:numinsts-1)= INT(numprocs/numinsts)
    IF (MOD(numprocs,numinsts).ne.0) THEN
      IF (myid.eq.ioid) WRITE(6,*) 'numprocs is not a multiple of numinsts, required for same_numfp=1 case. Stopping!'
      STOP
    END IF    
  END IF
    
  total_processors = SUM(fparts(0:numinsts-1)) 
  IF (numprocs.ne.total_processors) THEN
    IF (myid.eq.ioid) WRITE(6,*) 'numprocs is not equal to the number of processors requested by mpirun. Stopping!'
    STOP
  END IF
  
  numfparts_max = MAXVAL(fparts(0:numinsts-1))  
  ALLOCATE (which_id(0:numinsts-1,0:numfparts_max-1))
  which_id = undef_id 
  temp_counter = 0
  DO inst_counter=0, numinsts-1
    DO fpart_counter=0, fparts(inst_counter)-1
      IF (myid.eq.temp_counter) THEN
        my_inst = inst_counter
        my_fpart= fpart_counter
        numfparts = fparts(inst_counter)
      END IF
      which_id(inst_counter, fpart_counter) = temp_counter
      temp_counter = temp_counter+1
    END DO
  END DO
          
  
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!! Split MPI_COMM_WORLD into communicators for each instance, and for each fieldpart if necessary
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  CALL MPI_COMM_SPLIT( MPI_COMM_WORLD, my_inst, my_fpart, MPI_COMM_INSTANCE, ierr)
  IF (same_numfp.eq.1) THEN
    CALL MPI_COMM_SPLIT( MPI_COMM_WORLD, my_fpart, my_inst, MPI_COMM_FPART, ierr)
  END IF

 
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!! Open output file for each instance
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! 
  simname = instance_name(my_inst)
  SELECT CASE(numinsts)
  CASE (1)
    out_unit = 6
  CASE DEFAULT
    out_unit = out_unit_first+my_inst
    IF (my_fpart.eq.io_fpart) THEN
      WRITE(inum,'(i2.2)') my_inst
      OPEN(out_unit,FILE=inum//'_'//simname//'_out.txt',STATUS='UNKNOWN',FORM='FORMATTED',ACCESS='APPEND')
    END IF
  END SELECT

END SUBROUTINE mpi_environment_init



!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!                                                          !!
!!                       mpi_environment_end                !!
!!                                                          !!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
SUBROUTINE mpi_environment_end()
  USE mpi_var
  IMPLICIT NONE

  INTEGER :: rc

  IF ((numinsts.gt.1).and.(my_fpart.eq.io_fpart)) THEN
    CLOSE(out_unit)
  END IF

  CALL MPI_COMM_FREE (MPI_COMM_INSTANCE, ierr)
  IF (same_numfp.eq.1) CALL MPI_COMM_FREE (MPI_COMM_FPART, ierr)

  DEALLOCATE (which_id)

  CALL MPI_FINALIZE(rc)

END SUBROUTINE mpi_environment_end



!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!                                                          !!
!!                       init_variables                     !!
!!                                                          !!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
SUBROUTINE init_variables()
  USE parameters
  IMPLICIT NONE

  CALL init_fft3d()
  CALL init_random_seed()
  CALL init_wavevectors()

  IF (VELEQ) CALL init_vel_field()
  IF (MAGEQ) CALL init_mag_field()
  IF (SCAEQ) CALL init_sca_field()
  
  CALL external_stats_init()
  CALL zero_nhalf()
  CALL symmetrize_all()
 
  IF (nbuild.gt.0) CALL buildphases()

END SUBROUTINE init_variables



!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!                                                          !!
!!                     init_random_seed                     !!
!!                                                          !!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
SUBROUTINE init_random_seed()
  USE parameters
  USE mpi_var
  IMPLICIT NONE
  
  INTEGER, ALLOCATABLE, DIMENSION (:)  :: seed 
  INTEGER               :: nseed, ii

    SELECT CASE (userseed)
    CASE(0)
      IF (my_fpart.eq.io_fpart) WRITE(out_unit,*) '>> Processor initializes the seeds randomly from the date and time'
      CALL RANDOM_SEED()
    CASE DEFAULT
      IF (my_fpart.eq.io_fpart) WRITE(out_unit,*) '>> Processor initializes determinist seeds using the parameter userseed'
      CALL RANDOM_SEED(SIZE=nseed)
      ALLOCATE(seed(nseed))
      DO ii=1, nseed
        seed(ii)=ii+userseed+my_fpart
      END DO
      CALL RANDOM_SEED(PUT=seed)
      DEALLOCATE(seed)
    END SELECT

END SUBROUTINE init_random_seed



!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!                                                          !!
!!                     init_wavevectors                     !!
!!                                                          !!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
 SUBROUTINE init_wavevectors()
  USE parameters
  USE mpi_var
  USE spec_rest_mod  
  IMPLICIT NONE 

  INTEGER                 :: qx, qy, qz, qqy, iz, iiz, modes_layer_node
  CHARACTER(LEN=2)        :: inum
  
  INTEGER, ALLOCATABLE, DIMENSION (:)    :: temp
  INTEGER, ALLOCATABLE, DIMENSION (:,:)  :: temp2
  

!!! Provide information on the precision
  IF (my_fpart.eq.io_fpart) THEN   
    SELECT CASE (SIZEOF_REAL)
      CASE(4)
        WRITE(out_unit,*) '>> This is a single precision run'
      CASE(8)
        WRITE(out_unit,*) '>> This is a double precision run'
    END SELECT
  END IF  
  
!!! Allocate and initialize array related to the data structure
  ALLOCATE(ystart(0:numfparts-1), ysize(0:numfparts-1))
  ALLOCATE(zstart(0:numfparts-1), zsize(0:numfparts-1)) 
  ALLOCATE(temp(0:numfparts-1))
  temp=0
  temp(my_fpart)=local_y_start_after_transpose
  CALL mpi_allreduce(temp,ystart,numfparts,MPI_INTEGER,MPI_SUM,MPI_COMM_INSTANCE,ierr) 
  temp=0
  temp(my_fpart)=local_ny_after_transpose
  CALL mpi_allreduce(temp,ysize ,numfparts,MPI_INTEGER,MPI_SUM,MPI_COMM_INSTANCE,ierr) 
  temp=0
  temp(my_fpart)=local_z_start
  CALL mpi_allreduce(temp,zstart,numfparts,MPI_INTEGER,MPI_SUM,MPI_COMM_INSTANCE,ierr) 
  temp=0
  temp(my_fpart)=local_nz
  CALL mpi_allreduce(temp,zsize ,numfparts,MPI_INTEGER,MPI_SUM,MPI_COMM_INSTANCE,ierr) 
  DEALLOCATE(temp)
  
  
  ALLOCATE(whereis_ky(0:ny-1,2),temp2(0:ny-1,2))
  temp2=0
  DO qy=0,cy
    qqy=qy+local_y_start_after_transpose
    temp2(qqy,1)=my_fpart
    temp2(qqy,2)=qy
  END DO
  CALL mpi_allreduce(temp2,whereis_ky,2*ny,MPI_INTEGER,MPI_SUM,MPI_COMM_INSTANCE,ierr) 
  DEALLOCATE(temp2)
  
  ALLOCATE(whereis_z(0:nz-1,2),temp2(0:nz-1,2))
  temp2=0
  DO iz=0,rz
    iiz=iz+local_z_start
    temp2(iiz,1)=my_fpart
    temp2(iiz,2)=iz
  END DO
  CALL mpi_allreduce(temp2,whereis_z,2*nz,MPI_INTEGER,MPI_SUM,MPI_COMM_INSTANCE,ierr) 
  DEALLOCATE(temp2)

!!! Allocate and initialize array related to the wave vectors and to the forcing

  ALLOCATE(kx(0:cx), ky(0:cy), kz(0:cz), ikx(0:cx), iky(0:cy), ikz(0:cz), MM(0:cx))
  ALLOCATE(km2(0:cx,0:cz,0:cy),ksquare(0:cx,0:cz,0:cy),knorm(0:cx,0:cz,0:cy),zeros_twothird(0:cx,0:cz,0:cy))

  DO qx=0,nx/2
    kx(qx)=twopi*qx/lx
  END DO
  DO qz=0,nz/2
    kz(qz)=twopi*qz/lz
  END DO
  DO qz=nz/2+1,nz-1
    kz(qz)=twopi*(qz-nz)/lz
  END DO
  
  DO qy=0,cy
    qqy=qy+local_y_start_after_transpose
    IF (qqy.le.ny/2)  ky(qy)=twopi*(qqy)/ly
    IF (qqy.gt.ny/2)  ky(qy)=twopi*(qqy-ny)/ly
  END DO
  
  ikx=I*kx
  iky=I*ky
  ikz=I*kz
  MM=2
  MM(0)=1
  MM(nx/2)=0
  
  dx=lx/REAL(nx)
  dy=ly/REAL(ny)
  dz=lz/REAL(nz)

  k0x= twopi/lx
  k0y= twopi/ly
  k0z= twopi/lz
  k0 = MAX(k0x, k0y, k0z)
  IF (my_fpart.eq.io_fpart) WRITE(out_unit,*) '>> Kzero= ',k0

  kMx= pi*nx/lx
  kMy= pi*ny/ly
  kMz= pi*nz/lz
  IF(dealias==twothird) THEN
    kMx= 2.0*kMx/3.0
    kMy= 2.0*kMy/3.0
    kMz= 2.0*kMz/3.0
  END IF
  kM = MIN(kMx, kMy, kMz)
  IF (my_fpart.eq.io_fpart) WRITE(out_unit,*) '>> Kmax= ',kM

  modes_layer_node=0
  zeros_twothird=1.0
  DO qy=0, cy
    IF (ABS(ky(qy))>twopi*ny/ly/3.0) zeros_twothird(:,:,qy)=0.0
    DO qz=0, cz
      IF (ABS(kz(qz))>twopi*nz/lz/3.0) zeros_twothird(:,qz,qy)=0.0
      DO qx=0, cx
        IF (ABS(kx(qx))>twopi*nx/lx/3.0) zeros_twothird(qx,qz,qy)=0.0
        ksquare(qx,qz,qy)=kx(qx)*kx(qx)+ky(qy)*ky(qy)+kz(qz)*kz(qz)
        km2(qx,qz,qy)=1/MAX(eps,ksquare(qx,qz,qy))
        knorm(qx,qz,qy)=SQRT(ksquare(qx,qz,qy))
        IF ((knorm(qx,qz,qy).lt.ksup).AND.(knorm(qx,qz,qy).ge.kinf)) modes_layer_node=modes_layer_node+MM(qx)
      END DO
    END DO
  END DO
  CALL mpi_allreduce(modes_layer_node,modes_layer,1,MPI_INTEGER,MPI_SUM,MPI_COMM_INSTANCE,ierr)
  
  IF ((my_fpart.eq.io_fpart).and.VELEQ) THEN 
  IF ((myid.eq.ioid).and.VELEQ) THEN 
    SELECT CASE (FORCEPARA) 
    CASE(1)
      WRITE (out_unit,*) '>> The forcing is acting in the shell defined by : '
      WRITE (out_unit,*) '>>   ', kinf, ' < k < ', ksup
      WRITE (out_unit,*) '>> The number of modes in the forcing layer = ', modes_layer
      WRITE (out_unit,*) '>> The energy injection rate in the shell is imposed : epsilon=  '
      WRITE (out_unit,*) '>>   ', energy_forcing
      WRITE (out_unit,*) '>> The helicity parameter is', helicity_para
    CASE(2)
      WRITE (out_unit,*) '>> The forcing is acting in the shell defined by :  '
      WRITE (out_unit,*) '>>   ', kinf, ' < k < ', ksup
      WRITE (out_unit,*) '>> The number of modes in the forcing layer = ', modes_layer
      WRITE (out_unit,*) '>> The energy level in the shell is imposed : E=  '
      WRITE (out_unit,*) '>>   ', energy_forcing
      WRITE (out_unit,*) '>> The helicity parameter is', helicity_para
    CASE(4)
      WRITE (out_unit,*) '>> Kolmogorov forcing.'
      WRITE (out_unit,'(A24, F10.5)') '>> Forcing amplitude  = ', Famp
      WRITE (out_unit,'(A24, I4)')    '>> Forced mode        = ', Fmode
    CASE(6)
      WRITE (out_unit,*) '>> Helical forcing.'
      WRITE (out_unit,*) '>> The forcing is acting in the shell defined by :  '
      WRITE (out_unit,*) '>>   ', kinf, ' < k < ', ksup
      WRITE (out_unit,*) '>> The number of modes in the forcing layer = ', modes_layer
      WRITE (out_unit,*) '>> The energy injection rate in the shell is imposed : epsilon=  '
      WRITE (out_unit,*) '>>   ', energy_forcing
      WRITE (out_unit,*) '>> The helicity parameter is', helicity_para      
    CASE(7)
      WRITE (out_unit,*) '>> The forcing is acting in the shell defined by :  '
      WRITE (out_unit,*) '>>   ', kinf, ' < k < ', ksup
      WRITE (out_unit,*) '>> The number of modes in the forcing layer = ', modes_layer
      WRITE (out_unit,*) '>> The kinetic energy injection rate in the shell is = ', kin_energy_forcing
      WRITE (out_unit,*) '>> The magnetic energy injection rate in the shell is = ', mag_energy_forcing 
      WRITE (out_unit,*) '>> The kinetic-helicity parameter is = ', kin_hel_forcing 
	  WRITE (out_unit,*) '>> The magnetic-helicity parameter is = ', mag_hel_forcing
	  WRITE (out_unit,*) '>> The cross-helicity parameter is = ', croshel_forcing
    END SELECT
  END IF

  END IF

!!! Allocate and initialize array related to a solid body rotation if needed.
  
  IF (ROTEQ) THEN
    ALLOCATE(rot_omega(0:cx,0:cz,0:cy))
    DO qy=0, cy
      DO qz=0, cz
        DO qx=0, cx
          rot_omega(qx,qz,qy)=2.0*(Omega_x*kx(qx)+Omega_y*ky(qy)+Omega_z*kz(qz))/MAX(eps,knorm(qx,qz,qy))
        END DO
      END DO
    END DO
  END IF
  
!!! Allocate and initialize arrays related to the LES if needed.
  IF (LESTYPE.GT.0) THEN
    SELECT CASE (filter_type)
    CASE(1)
      ALLOCATE(test_filter(0:cx,0:cz,0:cy))
      test_filter=1.0    
      DO qy=0, cy
        IF (ABS(ky(qy)) > 0.5*kMy) test_filter(:,:,qy)=0.0
        DO qz=0, cz
          IF (ABS(kz(qz)) > 0.5*kMz) test_filter(:,qz,qy)=0.0
          DO qx=0, cx
            IF (ABS(kx(qx)) > 0.5*kMx) test_filter(qx,qz,qy)=0.0
          END DO
        END DO
      END DO    
    CASE(2)
      ALLOCATE(test_filter(0:cx,0:cz,0:cy))
      test_filter=1.0    
      DO qy=0, cy
        IF (ABS(ky(qy)) > 0.25*kMy) test_filter(:,:,qy)=0.0
        DO qz=0, cz
          IF (ABS(kz(qz)) > 0.25*kMz) test_filter(:,qz,qy)=0.0
          DO qx=0, cx
            IF (ABS(kx(qx)) > 0.25*kMx) test_filter(qx,qz,qy)=0.0
          END DO
        END DO
      END DO 
    CASE(3)
      IF(SGS_type.EQ.1)THEN
        ALLOCATE(test_filter(0:cx,0:cz,0:cy))
        test_filter=0.0
        DO qy=0, cy
          DO qz=0, cz
            DO qx=0, cx
              test_filter(qx,qz,qy)=EXP(-ksquare(qx,qz,qy)*3.0*(dx*dy*dz)**(2.0/3.0)/2.0)
            END DO
          END DO
        END DO 
        DO qy=0, cy
          IF (ABS(ky(qy)) > 0.5*kMy) test_filter(:,:,qy)=0.0
          DO qz=0, cz
            IF (ABS(kz(qz)) > 0.5*kMz) test_filter(:,qz,qy)=0.0
            DO qx=0, cx
              IF (ABS(kx(qx)) > 0.5*kMx) test_filter(qx,qz,qy)=0.0
            END DO
          END DO
        END DO    
      ELSE      
        ALLOCATE(LES_filter(0:cx,0:cz,0:cy))
        LES_filter=0.0
        DO qy=0, cy
          DO qz=0, cz
            DO qx=0, cx
              LES_filter(qx,qz,qy)=EXP(-ksquare(qx,qz,qy)*(dx*dy*dz)**(2.0/3.0)/2.0)
            END DO
          END DO
        END DO 
        ALLOCATE(Gauss_filter(0:cx,0:cz,0:cy))
        Gauss_filter=0.0
        DO qy=0, cy
          DO qz=0, cz
            DO qx=0, cx
              Gauss_filter(qx,qz,qy)=EXP(-ksquare(qx,qz,qy)*3.0*(dx*dy*dz)**(2.0/3.0)/2.0)
            END DO
          END DO
        END DO 
        ALLOCATE(Cut_filter(0:cx,0:cz,0:cy))
        Cut_filter=1.0    
        DO qy=0, cy
          IF (ABS(ky(qy)) > 0.5*kMy) Cut_filter(:,:,qy)=0.0
          DO qz=0, cz
            IF (ABS(kz(qz)) > 0.5*kMz) Cut_filter(:,qz,qy)=0.0
            DO qx=0, cx
              IF (ABS(kx(qx)) > 0.5*kMx) Cut_filter(qx,qz,qy)=0.0
            END DO
          END DO
        END DO    
        ALLOCATE(test_filter(0:cx,0:cz,0:cy))
        test_filter=1.0
        test_filter=Gauss_filter*Cut_filter
      END IF
    END SELECT
  END IF    
  
!!! Elements related to Tauij and Force Restore
  IF (field_exchange.EQ.1) THEN
    nxl=nx
    nyl=ny
    nzl=nz
    CALL MPI_BCAST(nxl, 1, MPI_INTEGER, numprocs-1,MPI_COMM_WORLD,ierr)
    CALL MPI_BCAST(nyl, 1, MPI_INTEGER, numprocs-1,MPI_COMM_WORLD,ierr)
    CALL MPI_BCAST(nzl, 1, MPI_INTEGER, numprocs-1,MPI_COMM_WORLD,ierr)
    cxl=nxl/2
    cyl=nyl-1
    czl=nzl-1
    kMxl=kMx*nxl/nx
    kMyl=kMy*nyl/ny
    kMzl=kMz*nzl/nz    
    Ns_ratio=(nxl*nyl*nzl)/REAL(Ns)
    CALL MPI_BCAST(Ns_ratio, 1, MPI_MYREAL, ioid,MPI_COMM_WORLD,ierr)
    
    rest_kfirst=0.5
    rest_klast=FLOOR(2.0*kM)/2.0
    rest_shell_width=1.0
    IF(LES_ident.EQ.1)THEN
      CALL spec_rest_init()    
      out_spa = 300+my_inst
      IF (my_fpart.eq.io_fpart) THEN
        WRITE(inum,'(i2.2)') my_inst
        OPEN(out_spa,FILE=inum//'_'//simname//'_spa.txt',STATUS='UNKNOWN',FORM='FORMATTED',ACCESS='APPEND')
      END IF
      out_spp = 500+my_inst
      IF (my_fpart.eq.io_fpart) THEN
        WRITE(inum,'(i2.2)') my_inst
        OPEN(out_spp,FILE=inum//'_'//simname//'_spp.txt',STATUS='UNKNOWN',FORM='FORMATTED',ACCESS='APPEND')
      END IF
      out_spi = 700+my_inst
      IF (my_fpart.eq.io_fpart) THEN
        WRITE(inum,'(i2.2)') my_inst
        OPEN(out_spi,FILE=inum//'_'//simname//'_spi.txt',STATUS='UNKNOWN',FORM='FORMATTED',ACCESS='APPEND')
      END IF    
    END IF
  END IF
  IF(field_exchange.EQ.1) dt_ext=1.0/(kMx**2+kMy**2+kMz**2)  
  
  IF (my_fpart.eq.io_fpart)  WRITE (out_unit,*) '>> Wave vectors initialized...'
  
END SUBROUTINE init_wavevectors



!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!                                                          !!
!!                    init_vel_field                        !!
!!                                                          !!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
SUBROUTINE init_vel_field()
  USE parameters
  USE variables
  USE mpi_var
  IMPLICIT NONE 

  ALLOCATE(vel(0:cx,0:cz,0:cy,3))
  ALLOCATE(dvel(0:cx,0:cz,0:cy,3))

  SELECT CASE (init_vel)
  CASE(restart)      !! read existing file
    CALL read_vec_divfree(vel,simname//'_u')
    IF (my_fpart.eq.io_fpart) WRITE(out_unit,*) '>> Initial velocity field has been read'
  CASE(random)       !! random field
    CALL init_random_vec_divfree(vel,vel_a,vel_b,vel_c,vel_d)
    IF (my_fpart.eq.io_fpart) WRITE(out_unit,*) '>> Initial random velocity field has been generated'
  CASE(taylor_green) !! Taylor-Green
    CALL init_TaylorGreen(vel,vel_a,vel_b)
    IF (my_fpart.eq.io_fpart) WRITE(out_unit,*) '>> Initial Taylor-Green velocity field has been generated'
  CASE(extern)       !! provided by the user externally
    CALL external_init_vel(vel)
    IF (my_fpart.eq.io_fpart) WRITE(out_unit,*) '>> Initial velocity field provided by user has been generated'
  CASE DEFAULT
    IF (my_fpart.eq.io_fpart) WRITE (out_unit,*) '>> Wrong value of init_vel'
    STOP
  END SELECT
  
  CALL divfree(vel)

END SUBROUTINE init_vel_field



!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!                                                          !!
!!                     init_mag_field                       !!
!!                                                          !!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
SUBROUTINE init_mag_field()
  USE parameters
  USE variables
  USE mpi_var
  IMPLICIT NONE 

  ALLOCATE(mag(0:cx,0:cz,0:cy,3))
  ALLOCATE(dmag(0:cx,0:cz,0:cy,3))

  SELECT CASE (init_mag)
  CASE(restart)      !! read existing file
    CALL read_vec_divfree(mag,simname//'_b')
    IF (my_fpart.eq.io_fpart) WRITE(out_unit,*) '>> Initial magnetic field has been read'
  CASE(random)       !! random field
    CALL init_random_vec_divfree(mag,mag_a,mag_b,mag_c,mag_d)
    IF (my_fpart.eq.io_fpart) WRITE(out_unit,*) '>> Initial random magnetic field has been generated'
  CASE(taylor_green) !! Taylor-Green
    CALL init_TaylorGreen(mag,mag_a,mag_b)
  CASE(extern)       !! provided by the user externally
    CALL external_init_mag(mag)
    IF (my_fpart.eq.io_fpart) WRITE(out_unit,*) '>> Initial magnetic field provided by user has been generated'
  CASE DEFAULT
    IF (my_fpart.eq.io_fpart) WRITE (out_unit,*) '>> Wrong value of init_mag'
    STOP
  END SELECT
  
  CALL divfree(mag)

END SUBROUTINE init_mag_field



!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!                                                          !!
!!                    init_sca_field                        !!
!!                                                          !!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
SUBROUTINE init_sca_field()
  USE parameters
  USE variables
  USE mpi_var
  IMPLICIT NONE 

  INTEGER           :: iscal
  CHARACTER(LEN=2)  :: csc 

  ALLOCATE(scal(0:cx,0:cz,0:cy,NSCAL))
  ALLOCATE(dscal(0:cx,0:cz,0:cy,NSCAL))

  DO iscal=1,NSCAL
    SELECT CASE (init_sca(iscal))
    CASE(0) !! read existing file
      WRITE(csc,'(i2.2)') iscal
      CALL read_scal(scal(:,:,:,iscal),simname//'_s'//csc//'.')
    CASE(1) !! random field
      CALL init_random_scal(scal(:,:,:,iscal),sca_a(iscal),sca_b(iscal),sca_c(iscal),sca_d(iscal))
    CASE(2) !! provided by the user externally
      CALL external_init_scal(scal(:,:,:,iscal),iscal)
    CASE DEFAULT
      IF (my_fpart.eq.io_fpart) WRITE (out_unit,*) '>> Wrong value of init_scal for iscal =', iscal
      STOP
    END SELECT
  END DO

END SUBROUTINE init_sca_field



!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!                                                          !!
!!                   init_TaylorGreen                       !!
!!                                                          !!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
SUBROUTINE init_TaylorGreen(u,angle,amplitude) 
  USE parameters
  USE mpi_var
  IMPLICIT NONE
  
  COMPLEX, DIMENSION(0:cx,0:cz,0:cy,3), INTENT(OUT) :: u
  REAL, INTENT(IN)                                  :: angle, amplitude

  REAL, ALLOCATABLE, DIMENSION(:,:,:)      :: ux,uy,uz
  COMPLEX, ALLOCATABLE, DIMENSION(:,:,:)   :: uc
  INTEGER      :: ix, iy, iz
  
  ALLOCATE(ux(0:rx,0:ry,0:rz),uy(0:rx,0:ry,0:rz),uz(0:rx,0:ry,0:rz))
  ALLOCATE(uc(0:cx,0:cz,0:cy))
    
!! Compute the three components of the Taylor Green vortex in real space

  DO ix=0, rx
    DO iy=0, ry
      DO iz= 0, rz
        ux(ix,iy,iz)=SIN(ix*lx/nx)*COS(iy*ly/ny)*COS((iz+local_z_start)*lz/nz)
        uy(ix,iy,iz)=COS(ix*lx/nx)*SIN(iy*ly/ny)*COS((iz+local_z_start)*lz/nz)
        uz(ix,iy,iz)=COS(ix*lx/nx)*COS(iy*ly/ny)*SIN((iz+local_z_start)*lz/nz)
      END DO
    END DO
  END DO
  ux(nx:nx+1,:,:)=0.0
  uy(nx:nx+1,:,:)=0.0
  uz(nx:nx+1,:,:)=0.0
  ux=ux*2.0/SQRT(3.0)*amplitude*SIN(angle+twopi/3.0)
  uy=uy*2.0/SQRT(3.0)*amplitude*SIN(angle-twopi/3.0)
  uz=uz*2.0/SQRT(3.0)*amplitude*SIN(angle) 
  
!! Fourier transform the three components of the Taylor Green vortex
  
  CALL fft3ds(ux,uc)
  u(:,:,:,1)=uc  
  CALL fft3ds(uy,uc)
  u(:,:,:,2)=uc  
  CALL fft3ds(uz,uc)
  u(:,:,:,3)=uc
 
  DEALLOCATE(ux,uy,uz,uc)
  
END SUBROUTINE init_TaylorGreen



!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!                                                          !!
!!                init_random_vec_divfree                   !!
!!                                                          !!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
SUBROUTINE init_random_vec_divfree(u,pa,pb,pc,pd) 
  USE parameters
  USE mpi_var
  IMPLICIT NONE

  COMPLEX, DIMENSION(0:cx,0:cz,0:cy,3), INTENT(OUT)   :: u
  REAL, INTENT(IN)                                    :: pa, pb, pc, pd
  INTEGER               :: qx, qy, qz
  REAL                  :: theta1, theta2, phi, kxy, k, amp
  COMPLEX               :: al, be
  REAL, EXTERNAL        :: external_energy_spectrum
  

! First step, ensure that the random number will be the same independently of the numfparts.

  IF (my_fpart.ne.io_fpart) THEN
    DO qx=0,3*nx*nz/2*ystart(my_fpart)-1
      CALL RANDOM_NUMBER(theta1)
    END DO
  END IF

! Second step, compute the random field
  
  DO qy=0,cy       ! the order of the loops is important to guarantee equivalence
    DO qx=0,cx     ! between runs using different number of nodes
      DO qz=0,cz
        kxy    = MAX(SQRT(kx(qx)*kx(qx)+ky(qy)*ky(qy)),EPS)
        k      = MAX(knorm(qx,qz,qy),EPS)
        amp    = SQRT(external_energy_spectrum(k,pa,pb,pc,pd)/2.0/pi/k**2)
        CALL RANDOM_NUMBER(theta1)
        CALL RANDOM_NUMBER(theta2)
        CALL RANDOM_NUMBER(phi)
        theta1 = twopi*theta1
        theta2 = twopi*theta2
        phi    = twopi*phi
        al     = amp * (COS(theta1)+I*SIN(theta1)) * COS(phi)
        be     = amp * (COS(theta2)+I*SIN(theta2)) * SIN(phi)
        u(qx,qz,qy,1) = (+ al * k *  ky(qy) + be * kx(qx) * kz(qz)) / k / kxy
        u(qx,qz,qy,2) = (- al * k *  kx(qx) + be * ky(qy) * kz(qz)) / k / kxy
        u(qx,qz,qy,3) =  - be * kxy / k
        IF ((ABS(kx(qx)).lt.EPS).AND.(ABS(ky(qy)).lt.EPS)) THEN
          u(qx,qz,qy,1) = (al + be )/SQRT(2.0)
          u(qx,qz,qy,2) = (be - al )/SQRT(2.0)
          u(qx,qz,qy,3) = 0.0
        END IF
      END DO
    END DO
  END DO

! Third step, ensure that the total number of call to RANDOM_NUMBER is independent of numfparts

  IF((ny-ystart(my_fpart)-ysize(my_fpart)).ne.0)THEN
    DO qx=0,3*nx*nz/2*(ny-ystart(my_fpart)-ysize(my_fpart))-1
      CALL RANDOM_NUMBER(theta1)
    END DO
  END IF

END SUBROUTINE init_random_vec_divfree



!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!                                                          !!
!!                   init_random_scal                       !!
!!                                                          !!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
SUBROUTINE init_random_scal(sc,pa,pb,pc,pd)
  USE parameters
  USE mpi_var
  IMPLICIT NONE
  
  COMPLEX, DIMENSION(0:cx,0:cz,0:cy), INTENT(OUT) :: sc
  REAL, INTENT(IN)                                :: pa, pb, pc, pd
  INTEGER              :: qx, qy, qz 
  REAL                 :: theta, k, amp
  REAL, EXTERNAL       :: external_energy_spectrum
 
! First step, ensure that the random number will be the same independently of the numfparts.

  IF (myid.ne.ioid) THEN
    DO qx=0,(cx+1)*ystart(my_fpart)*(cz+1)-1
      CALL RANDOM_NUMBER(theta)
    END DO
  END IF

! Second step, compute the random field

  DO qy=0,cy       ! the order of the loops is important to guarantee equivalence
    DO qx=0,cx     ! between runs using different number of nodes
      DO qz=0,cz
         CALL RANDOM_NUMBER(theta)
         theta = twopi*theta
         k     = MAX(knorm(qx,qz,qy),EPS)
         amp   = SQRT(external_energy_spectrum(k,pa,pb,pc,pd)/2.0/pi/k**2)
         sc(qx,qz,qy)=amp*EXP(I*theta)
      END DO
    END DO
  END DO

! Third step, ensure that the total number of call to RANDOM_NUMBER is independent of numfparts

  IF((ny-ystart(my_fpart)-ysize(my_fpart)).ne.0)THEN
    DO qx=0,nx*nz/2*(ny-ystart(my_fpart)-ysize(my_fpart))-1
      CALL RANDOM_NUMBER(theta)
    END DO
  END IF

END SUBROUTINE init_random_scal



!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!                                                          !!
!!                        buildphases                       !!
!!                                                          !!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
SUBROUTINE buildphases()
  USE parameters
  USE variables
  IMPLICIT NONE

  INTEGER             :: qx, qy, qz, iscal, iter
  REAL                :: fac, k, amp1, amp2, ener_u, ener_b, kin_heli, mag_heli, cross_heli, ener_s, diss_s
  REAL, EXTERNAL      :: external_energy_spectrum

  DO iter=1,nbuild 
    CALL compute_dt()
    CALL update_variables(iter)

!!  rescale velocity if it is initalized randomly, otherwise reset it to initial value
    IF (VELEQ) THEN
      SELECT CASE(init_vel)
      CASE(1) !! rescale random field
        DO qy=0,cy
          DO qz=0,cz
            DO qx=0,cx
              k      = MAX(knorm(qx,qz,qy),EPS)
              amp1   = SQRT(external_energy_spectrum(k,vel_a,vel_b,vel_c,vel_d)/2.0/pi/k**2)
              amp2   = ABS(vel(qx,qz,qy,1))**2+ABS(vel(qx,qz,qy,2))**2+ABS(vel(qx,qz,qy,3))**2
              fac    = amp1/SQRT(MAX(amp2,EPS))
              vel(qx,qz,qy,1:3) = vel(qx,qz,qy,1:3)*fac
            END DO
          END DO
        END DO
      CASE default !! re-initialise as prescribed by the parameter init_vel
        CALL init_vel_field()
      END SELECT
    END IF
  
!!  rescale magnetic field if it is initalized randomly, otherwise reset it to initial value
    IF (MAGEQ) THEN
      SELECT CASE(init_mag)
      CASE(1) !! rescale random field
        DO qy=0,cy
          DO qz=0,cz
            DO qx=0,cx
              k      = MAX(knorm(qx,qz,qy),EPS)
              amp1   = SQRT(external_energy_spectrum(k,mag_a,mag_b,mag_c,mag_d)/2.0/pi/k**2)
              amp2   = ABS(mag(qx,qz,qy,1))**2+ABS(mag(qx,qz,qy,2))**2+ABS(mag(qx,qz,qy,3))**2
              fac    = amp1/SQRT(MAX(amp2,EPS))
              mag(qx,qz,qy,1:3) = mag(qx,qz,qy,1:3)*fac
            END DO
          END DO
        END DO
      CASE default !! re-initialise as prescribed by the parameter init_mag
        CALL init_mag_field()
      END SELECT
    END IF

!!  rescale scalar fields if they are initalized randomly, otherwise reset them to initial value
    IF (SCAEQ) THEN
      DO iscal=1,NSCAL
        SELECT CASE (init_sca(iscal))
        CASE(1) !! rescale random field
          DO qy=0,cy
            DO qz=0,cz
              DO qx=0,cx
                k      = MAX(knorm(qx,qz,qy),EPS)
                amp1   = SQRT(external_energy_spectrum(k,sca_a(iscal),sca_b(iscal),sca_c(iscal),sca_d(iscal))/2.0/pi/k**2)
                amp2   = ABS(scal(qx,qz,qy,iscal))**2
                fac    = amp1/SQRT(MAX(amp2,EPS))
                scal(qx,qz,qy,iscal) = scal(qx,qz,qy,iscal)*fac
              END DO
            END DO
          END DO
        CASE DEFAULT !! re-initialise as prescribed by the parameter init_sca
          CALL init_sca_field()
        END SELECT
      END DO
    END IF  
    
    CALL external_stats_compute(iter)    
    
  END DO  ! end of the loop on iter
  
END SUBROUTINE buildphases



!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!                                                          !!
!!                      symmetrize_all                      !!
!!                                                          !!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
SUBROUTINE symmetrize_all()
  USE parameters
  USE variables
  IMPLICIT NONE
  
  INTEGER   :: iscal
  
  IF (VELEQ) THEN
    CALL symmetrize_one(vel(:,:,:,1))
    CALL symmetrize_one(vel(:,:,:,2))
    CALL symmetrize_one(vel(:,:,:,3))
  END IF
  
  IF (MAGEQ) THEN
    CALL symmetrize_one(mag(:,:,:,1))
    CALL symmetrize_one(mag(:,:,:,2))
    CALL symmetrize_one(mag(:,:,:,3))
  END IF

  IF (SCAEQ) THEN
    DO iscal=1, NSCAL 
      CALL symmetrize_one(scal(:,:,:,iscal))
    END DO
  END IF
  
END SUBROUTINE symmetrize_all



!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!! !!
!! symmetrize_one !!
!! !!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
SUBROUTINE symmetrize_one(s)
  USE mpi_var
  USE parameters
  IMPLICIT NONE

  COMPLEX, DIMENSION(0:cx,0:cz,0:cy), INTENT(INOUT) :: s
  INTEGER                                           :: qy, qz, qy1, qy2, qqy1, qqy2, pid1, pid2
  COMPLEX, ALLOCATABLE, DIMENSION(:)                :: sb

  ALLOCATE(sb(0:cz))
  DO qy=1,ny/2-1
    qy1=qy
    qy2=ny-qy
    pid1=whereis_ky(qy1,1)
    pid2=whereis_ky(qy2,1)
    qqy1=whereis_ky(qy1,2)
    qqy2=whereis_ky(qy2,2)
    IF (pid1.ne.pid2) THEN
      IF (my_fpart.eq.pid1) CALL MPI_SEND(s(0,:,qqy1),nz,MPI_MYCOMPLEX,pid2,10,MPI_COMM_INSTANCE,ierr)
      IF (my_fpart.eq.pid2) CALL MPI_RECV(sb         ,nz,MPI_MYCOMPLEX,pid1,10,MPI_COMM_INSTANCE,status_mpi,ierr)
    ELSE
      IF (my_fpart.eq.pid2) sb=s(0,:,qqy1)
    END IF
    IF (my_fpart.eq.pid2) THEN
      DO qz=1,cz
        s(0,nz-qz,qqy2) = CONJG(sb(qz))
      END DO
      s(0,0,qqy2) = CONJG(sb(0))
    END IF
  END DO
  DEALLOCATE(sb)
  
  pid1=whereis_ky(0,1)
  qqy1=whereis_ky(0,2)
  IF (my_fpart.eq.pid1) THEN
    DO qz=1,nz/2-1
      s(0,nz-qz,qqy1) = CONJG(s(0,qz,qqy1))
    END DO
  END IF

END SUBROUTINE symmetrize_one



!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!                                                          !!
!!                          divfree                         !!
!!                                                          !!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
SUBROUTINE divfree(a)
  USE parameters
  IMPLICIT NONE

  COMPLEX, DIMENSION(0:cx,0:cz,0:cy,3), INTENT(INOUT) :: a
  REAL       :: k1, k2, k3
  INTEGER    :: qx, qy, qz
  COMPLEX    :: div
  
  DO qy=0, cy
    k2=ky(qy)
    DO qz= 0, cz
      k3=kz(qz)
      DO qx=0, cx
        k1=kx(qx)
        div=(k1*a(qx,qz,qy,1)+k2*a(qx,qz,qy,2)+k3*a(qx,qz,qy,3))*km2(qx,qz,qy)
        a(qx,qz,qy,1)=a(qx,qz,qy,1)-k1*div
        a(qx,qz,qy,2)=a(qx,qz,qy,2)-k2*div
        a(qx,qz,qy,3)=a(qx,qz,qy,3)-k3*div
      END DO
    END DO
  END DO
  
END SUBROUTINE divfree



!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!                                                          !!
!!                       zero_nhalf                         !!
!!                                                          !!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
SUBROUTINE zero_nhalf()
  USE parameters
  USE variables
  USE mpi_var
  IMPLICIT NONE
    
  INTEGER :: qqy, pid
  
  pid = whereis_ky(ny/2,1)
  qqy = whereis_ky(ny/2,2)
  
  IF (VELEQ) THEN
    vel(nx/2,:,:,:)=0
    vel(:,nz/2,:,:)=0
    dvel(nx/2,:,:,:)=0
    dvel(:,nz/2,:,:)=0
    IF (my_fpart.eq.pid) THEN
      vel(:,:,qqy,:) =0
      dvel(:,:,qqy,:)=0
    END IF
  END IF

  IF (MAGEQ) THEN
    mag(nx/2,:,:,:)=0
    mag(:,nz/2,:,:)=0
    dmag(nx/2,:,:,:)=0
    dmag(:,nz/2,:,:)=0
    IF (my_fpart.eq.pid) THEN
      mag(:,:,qqy,:) =0
      dmag(:,:,qqy,:)=0
    END IF
  END IF
  
  IF (SCAEQ) THEN
    scal(nx/2,:,:,:)=0
    scal(:,nz/2,:,:)=0
    dscal(nx/2,:,:,:)=0  
    dscal(:,nz/2,:,:)=0  
    IF (my_fpart.eq.pid) THEN
      scal(:,:,qqy,:) =0
      dscal(:,:,qqy,:)=0
    END IF
  END IF
  
END SUBROUTINE zero_nhalf

