!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!  12 August 2011                                          !!
!!                      tt_io.f90                           !!
!!                                                          !!
!!  output                                                  !!
!!  write_vec_divfree                                       !!
!!  write_scal                                              !!
!!  read_vec_divfree                                        !!
!!  read_scal                                               !!
!!  transfer_slice                                          !!
!!  read_time                                               !!
!!  read_para                                               !!
!!                                             version 2.05 !!
!!                                                          !!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!   



!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!                                                          !!
!!                          output                          !!
!!                                                          !!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
SUBROUTINE output(save_stat)
  USE parameters
  USE variables
  USE mpi_var
  IMPLICIT NONE
  
  INTEGER, INTENT(IN)  :: save_stat
  CHARACTER(LEN=3)     :: cnum
  CHARACTER(LEN=2)     :: csc, inum
  INTEGER              :: iscal

!! First : save fields
  IF (VELEQ) CALL write_vec_divfree(vel,simname//'_u')
  IF (MAGEQ) CALL write_vec_divfree(mag,simname//'_b')
  IF (SCAEQ) THEN
    DO iscal=1,NSCAL
      WRITE(csc,'(i2.2)') iscal
      CALL write_scal(scal(:,:,:,iscal),simname//'_s'//csc//'.')
    END DO
  END IF

!! Second : save files with time and the index of the last saved fields
  IF (my_fpart.eq.io_fpart) THEN  
    WRITE(cnum,'(i3.3)') nfields
    WRITE(inum,'(i2.2)') my_inst
    IF (numinsts.gt.1) THEN
      OPEN (10, FILE=inum//'_'//simname//'_t.'//cnum, STATUS='UNKNOWN', FORM='FORMATTED')
      OPEN (11, FILE=inum//'_'//simname//'.n', STATUS='UNKNOWN', FORM='FORMATTED')      
    ELSE 
      OPEN (10, FILE=simname//'_t.'//cnum, STATUS='UNKNOWN', FORM='FORMATTED')
      OPEN (11, FILE=simname//'.n', STATUS='UNKNOWN', FORM='FORMATTED')     
    END IF
    WRITE(10,NML=time_parameters) 
    WRITE(11,*) nfields  
    CLOSE (10)          
    CLOSE (11)    
  END IF
  
!! Third : save intermediate statistics files
  IF (save_stat.eq.1) THEN
    CALL external_stats_write()
    CALL external_stats_end()
    CALL external_stats_init()
  END IF

!! update the value of nfields
  nfields=nfields+1

END SUBROUTINE output



!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!                                                          !!
!!                     write_vec_divfree                    !!
!!                                                          !!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
SUBROUTINE write_vec_divfree(u,beginname)
  USE parameters
  USE mpi_var
  IMPLICIT NONE
  
  COMPLEX, DIMENSION(0:cx,0:cz,0:cy,3), INTENT(IN)   :: u
  CHARACTER(LEN=namelength+2), INTENT(IN)            :: beginname
  CHARACTER(LEN=namelength+11)                       :: filename
  INTEGER                                            :: ipart, reccount, reclength
  CHARACTER(LEN=3)                                   :: cnum
  CHARACTER(LEN=2)                                   :: partnum, inum
  COMPLEX, ALLOCATABLE, DIMENSION(:,:)               :: utempx, utempz
  INTEGER                                            :: qy, pid, qqy

  WRITE(cnum,'(i3.3)') nfields
  WRITE(inum,'(i2.2)') my_inst
  reclength=2*(cx+1)*(cz+1)*SIZEOF_REAL
  ALLOCATE(utempx(0:cx,0:cz),utempz(0:cx,0:cz))

  DO ipart=0,totpart-1
    IF (my_fpart.eq.io_fpart) THEN
      WRITE(partnum,'(i2.2)') ipart
      IF (numinsts.gt.1) THEN
        filename=inum//'_'//beginname//partnum//'.'//cnum
      ELSE
        filename=beginname//partnum//'.'//cnum
      END IF
      OPEN (UNIT=10, FILE=filename, STATUS='UNKNOWN', ACTION='WRITE',ACCESS='DIRECT',&
          FORM='UNFORMATTED', RecL=reclength)
      reccount=1
    END IF
    
    IF (ipart.eq.0) THEN
      pid=whereis_ky(0,1)
      qqy=whereis_ky(0,2)
      CALL transfer_slice(u(:,:,qqy,2),pid,utempx,io_fpart)
      IF (my_fpart.eq.io_fpart) THEN
        WRITE(10,rec=reccount) utempx
        reccount=2
      END IF
    END IF

    DO qy=ipart*ny/totpart,(ipart+1)*ny/totpart-1
      pid=whereis_ky(qy,1)
      qqy=whereis_ky(qy,2)
      CALL transfer_slice(u(:,:,qqy,1),pid,utempx,io_fpart)
      CALL transfer_slice(u(:,:,qqy,3),pid,utempz,io_fpart)
      IF (my_fpart.eq.io_fpart) THEN
        WRITE(10,rec=reccount) utempx
        WRITE(10,rec=reccount+1) utempz
        reccount=reccount+2
      END IF
    END DO  

    IF (my_fpart.eq.io_fpart) CLOSE(10)
  END DO ! ipart loop

  DEALLOCATE(utempx,utempz)

END SUBROUTINE write_vec_divfree



!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!                                                          !!
!!                     read_vec_divfree                     !!
!!                                                          !!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
SUBROUTINE read_vec_divfree(u,beginname)
  USE parameters
  USE mpi_var
  IMPLICIT NONE

  COMPLEX, DIMENSION(0:cx,0:cz,0:cy,3), INTENT(OUT)  :: u
  CHARACTER(LEN=namelength+2), INTENT(IN)            :: beginname
  CHARACTER(LEN=namelength+11)                       :: filename
  INTEGER                                            :: ipart, reccount, reclength
  CHARACTER(LEN=3)                                   :: cnum
  CHARACTER(LEN=2)                                   :: partnum, inum
  COMPLEX, ALLOCATABLE, DIMENSION(:,:)               :: utempx, utempz
  INTEGER                                            :: qy, qx, qz, pid, qqy

  WRITE(cnum,'(i3.3)') nfields
  WRITE(inum,'(i2.2)') my_inst
  reclength=2*(cx+1)*(cz+1)*SIZEOF_REAL
  ALLOCATE(utempx(0:cx,0:cz),utempz(0:cx,0:cz))

  DO ipart=0,totpart-1
    IF (my_fpart.eq.io_fpart) THEN
      WRITE(partnum,'(i2.2)') ipart
      IF (numinsts.gt.1) THEN
        filename=inum//'_'//beginname//partnum//'.'//cnum
      ELSE
        filename=beginname//partnum//'.'//cnum
      END IF
      OPEN (UNIT=10, FILE=filename, STATUS='UNKNOWN', ACTION='READ',ACCESS='DIRECT',&
          FORM='UNFORMATTED', RecL=reclength)
      reccount=1
      IF (ipart.eq.0) THEN
        READ(10,rec=reccount) utempx
        pid=whereis_ky(0,1)
        qqy=whereis_ky(0,2)
        CALL transfer_slice(utempx,io_fpart,u(:,:,qqy,2),pid)
        reccount=2
      END IF
    END IF
    
    DO qy=ipart*ny/totpart,(ipart+1)*ny/totpart-1
      IF (my_fpart.eq.io_fpart) THEN
        READ(10,rec=reccount) utempx
        READ(10,rec=reccount+1) utempz
        reccount=reccount+2
      END IF
      pid=whereis_ky(qy,1)
      qqy=whereis_ky(qy,2)
      CALL transfer_slice(utempx,io_fpart,u(:,:,qqy,1),pid)
      CALL transfer_slice(utempz,io_fpart,u(:,:,qqy,3),pid)
    END DO  

    IF (my_fpart.eq.io_fpart) CLOSE(10)
  END DO ! ipart loop
  DEALLOCATE(utempx,utempz)
 
  DO qy=0,cy
    IF (ABS(ky(qy)).gt.EPS) THEN
      DO qz=0, cz
        DO qx=0, cx
          u(qx,qz,qy,2)=-(u(qx,qz,qy,1)*kx(qx)+u(qx,qz,qy,3)*kz(qz))/ky(qy)
        END DO
      END DO
    END IF
  END DO
     
END SUBROUTINE read_vec_divfree



!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!                                                          !!
!!                     write_scal                           !!
!!                                                          !!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
SUBROUTINE write_scal(s,beginname)
  USE parameters
  USE mpi_var
  IMPLICIT NONE
  
  COMPLEX, DIMENSION(0:cx,0:cz,0:cy), INTENT(IN)     :: s
  CHARACTER(LEN=namelength+5), INTENT(IN)            :: beginname
  CHARACTER(LEN=namelength+14)                       :: filename
  INTEGER                                            :: ipart, reccount, reclength
  CHARACTER(LEN=3)                                   :: cnum
  CHARACTER(LEN=2)                                   :: partnum, inum
  COMPLEX, ALLOCATABLE, DIMENSION(:,:)               :: stemp
  INTEGER                                            :: qy, pid, qqy
  
  WRITE(cnum,'(i3.3)') nfields
  WRITE(inum,'(i2.2)') my_inst
  reclength=2*(cx+1)*(cz+1)*SIZEOF_REAL
  ALLOCATE(stemp(0:cx,0:cz))

  DO ipart=0,totpart-1
    IF (my_fpart.eq.io_fpart) THEN
      WRITE(partnum,'(i2.2)') ipart
      IF (numinsts.gt.1) THEN
        filename=inum//'_'//beginname//partnum//'.'//cnum
      ELSE
        filename=beginname//partnum//'.'//cnum
      END IF
      OPEN (UNIT=10, FILE=filename, STATUS='UNKNOWN', ACTION='WRITE',ACCESS='DIRECT',&
          FORM='UNFORMATTED', RecL=reclength)
      reccount=1
    END IF
    
    DO qy=ipart*ny/totpart,(ipart+1)*ny/totpart-1
      pid=whereis_ky(qy,1)
      qqy=whereis_ky(qy,2)
      CALL transfer_slice(s(:,:,qqy),pid,stemp,io_fpart)
      IF (my_fpart.eq.io_fpart) THEN
        WRITE(10,rec=reccount) stemp
        reccount=reccount+1
      END IF
    END DO  

    IF (my_fpart.eq.io_fpart) CLOSE(10)
  END DO ! ipart loop
  DEALLOCATE(stemp)
  
END SUBROUTINE write_scal



!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!                                                          !!
!!                       read_scal                          !!
!!                                                          !!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
SUBROUTINE read_scal(s,beginname)
  USE parameters
  USE mpi_var
  IMPLICIT NONE

  COMPLEX, DIMENSION(0:cx,0:cz,0:cy), INTENT(OUT)    :: s
  CHARACTER(LEN=namelength+5), INTENT(IN)            :: beginname
  CHARACTER(LEN=namelength+14)                       :: filename
  INTEGER                                            :: ipart, reccount, reclength
  CHARACTER(LEN=3)                                   :: cnum
  CHARACTER(LEN=2)                                   :: partnum, inum
  COMPLEX, ALLOCATABLE, DIMENSION(:,:)               :: stemp
  INTEGER                                            :: qy, pid, qqy

  WRITE(cnum,'(i3.3)') nfields
  WRITE(inum,'(i2.2)') my_inst
  reclength=2*(cx+1)*(cz+1)*SIZEOF_REAL
  ALLOCATE(stemp(0:cx,0:cz))

  DO ipart=0,totpart-1
    IF (my_fpart.eq.io_fpart) THEN
      WRITE(partnum,'(i2.2)') ipart
      IF (numinsts.gt.1) THEN
        filename=inum//'_'//beginname//partnum//'.'//cnum
      ELSE
        filename=beginname//partnum//'.'//cnum
      END IF
      OPEN (UNIT=10, FILE=filename, STATUS='UNKNOWN', ACTION='READ',ACCESS='DIRECT',&
          FORM='UNFORMATTED', RecL=reclength)
      reccount=1
    END IF
    
    DO qy=ipart*ny/totpart,(ipart+1)*ny/totpart-1
      IF (my_fpart.eq.io_fpart) THEN
        READ(10,rec=reccount) stemp
        reccount=reccount+1
      END IF
      pid=whereis_ky(qy,1)
      qqy=whereis_ky(qy,2)
      CALL transfer_slice(stemp,io_fpart,s(:,:,qqy),pid)
    END DO  

    IF (my_fpart.eq.io_fpart) CLOSE(10)
  END DO ! ipart loop
  DEALLOCATE(stemp)
      
END SUBROUTINE read_scal



!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!                                                          !!
!!                        read_time                         !!
!!                                                          !!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
SUBROUTINE read_time()
  USE parameters
  USE mpi_var
  IMPLICIT NONE

  CHARACTER(LEN=3)                       :: cnum
  CHARACTER(LEN=2)                       :: inum

  WRITE(inum,'(i2.2)') my_inst

  IF (my_fpart.eq.io_fpart) THEN
    IF(time_to_zero.EQ.1) THEN
      t = 0.0
      niter_done = 0
      nfields = 0 
      WRITE(out_unit,*) '>> TEST RUN (!): Time and save parameters are set to zero.'
      WRITE(out_unit,*) '>> WARNING  (!): Existing output will be overwritten.'
      IF (numinsts.gt.1) THEN
        OPEN(10,FILE=inum//'_'//simname//'.n',STATUS='UNKNOWN',FORM='FORMATTED')        
      ELSE
        OPEN(10,FILE=simname//'.n',STATUS='UNKNOWN',FORM='FORMATTED')
      END IF
      WRITE(10,*) nfields
      CLOSE(10)     
    ELSE    
      IF (numinsts.gt.1) THEN
        OPEN(10,FILE=inum//'_'//simname//'.n',STATUS='UNKNOWN',FORM='FORMATTED')        
      ELSE
        OPEN(10,FILE=simname//'.n',STATUS='UNKNOWN',FORM='FORMATTED')
      END IF
      READ(10,*) nfields
      CLOSE(10)
      WRITE(out_unit,*) '>> PRODUCTION RUN: Time and save parameters are read from file number : '
      WRITE(out_unit,*) '>>        ', nfields
     
      WRITE(cnum,'(i3.3)') nfields
      IF (numinsts.gt.1) THEN
        OPEN(10,FILE=inum//'_'//simname//'_t.'//cnum,STATUS='OLD',FORM='FORMATTED')      
      ELSE
        OPEN(10,FILE=simname//'_t.'//cnum,STATUS='OLD',FORM='FORMATTED')        
      END IF
      READ(10,NML=time_parameters)
      CLOSE(10)
    END IF
  END IF
  CALL MPI_BCAST(niter_done,1,MPI_INTEGER,ioid,MPI_COMM_INSTANCE,ierr)
  CALL MPI_BCAST(nfields,1,MPI_INTEGER,ioid,MPI_COMM_INSTANCE,ierr)
  CALL MPI_BCAST(t,1,MPI_MYREAL ,ioid,MPI_COMM_INSTANCE,ierr)
  
END SUBROUTINE read_time


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!                                                          !!
!!                         read_para                        !!
!!                                                          !!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
SUBROUTINE read_para()
  USE parameters
  USE mpi_var
  IMPLICIT NONE
  
  CHARACTER(LEN=namelength+4)                      :: filename
  INTEGER  ::  global_exchange
  filename=simname//'.par'

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! Read and broadcast the equation parameters. Determine and broadcast the
! flags VELEQ, MAGEQ and SCAEQ and allocate and read and broadcast the 
! scalar transport coefficients if required 
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  IF (my_fpart.eq.io_fpart) THEN 
    OPEN(10,FILE=filename,STATUS='OLD',FORM='FORMATTED')
    READ(unit=10,NML=equation_parameters)
    CLOSE(10)
    
    IF (EQTYPE==reac_diff) THEN 
      VELEQ = .FALSE. 
    ELSE
      VELEQ = .TRUE. 
    END IF
    
    IF (EQTYPE==full_mhd) THEN 
      MAGEQ = .TRUE. 
    ELSE 
      MAGEQ = .FALSE.
    END IF
    
    IF (NSCAL>0) THEN
      SCAEQ = .TRUE.
    ELSE
      SCAEQ = .FALSE.
    END IF
    
    SELECT CASE (EQTYPE)
    CASE(0)
      WRITE(out_unit,*)'>> No velocity nor magnetic field equations will be solved. Reaction-Diffusion problem.'
    CASE(1)
      WRITE(out_unit,*)'>> Navier-Stokes run.'
    CASE(2)
      WRITE(out_unit,*)'>> Navier-Stokes run with quasi-static approximation.'
    CASE(3)
      WRITE(out_unit,*)'>> MHD run.'
    CASE DEFAULT
      WRITE(out_unit,*)'>> Wrong value of the EQTYPE parameter in the equation_parameters namelist.'
      STOP
    END SELECT   

    IF (NSCAL==0) WRITE(out_unit,*) '>> No scalar field in this run.'
    IF ((NSCAL<0).or.(NSCAL>99)) THEN
      WRITE(out_unit,*)'>> Wrong value of the NSCAL parameter in the equation_parameters namelist.'
      STOP
    END IF
    IF ((NSCAL>0).or.(NSCAL<100)) THEN
      WRITE(out_unit,*)'>> This run will treat', NSCAL, ' passive scalars.'
      SELECT CASE (CHIM)
      CASE(0)
        WRITE(out_unit,*)'>> No chemistry term added in the passive scalar equations.'
      CASE(1)
        WRITE(out_unit,*)'>> A chemistry term is added in the passive scalar equations.'
      CASE DEFAULT
        WRITE(out_unit,*)'>> Wrong value of the CHIM parameter in the equation_parameters namelist.'
        STOP
      END SELECT   
    END IF
    
    IF (VELEQ) THEN
      SELECT CASE (FORCEPARA)
      CASE(0)
        WRITE(out_unit,*)'>> No forcing term in the Navier-Stokes equation.'
      CASE(1)
        WRITE(out_unit,*)'>> Fixed energy and helicity injection rates in a shell of wave vectors.'
      CASE(2)
        WRITE(out_unit,*)'>> Fixed energy and helicity injection levels in a shell of wave vectors.'
      CASE(3)
        WRITE(out_unit,*)'>> Impose a constant force in time.'
      CASE(4)
        WRITE(out_unit,*)'>> Kolmogorov forcing'
      CASE(5)
        WRITE(out_unit,*)'>> Impose a constant velocity profile in time'
      CASE(6)
        WRITE(out_unit,*)'>> Helical forcing.'
      CASE(7)
        WRITE(out_unit,*)'>> Fixed energy and cross-helicity injection rates in a shell of wave vectors.'
      CASE DEFAULT
        WRITE(out_unit,*)'>> Wrong value of the FORCE parameter in the equation_parameters namelist.'
        STOP
      END SELECT
    END IF
    
  END IF
  
  CALL MPI_BCAST(VELEQ,1,MPI_LOGICAL,ioid,MPI_COMM_INSTANCE,ierr)
  CALL MPI_BCAST(MAGEQ,1,MPI_LOGICAL,ioid,MPI_COMM_INSTANCE,ierr)
  CALL MPI_BCAST(SCAEQ,1,MPI_LOGICAL,ioid,MPI_COMM_INSTANCE,ierr)
  CALL MPI_BCAST(EQTYPE,1,MPI_INTEGER,ioid,MPI_COMM_INSTANCE,ierr)
  CALL MPI_BCAST(FORCEPARA,1,MPI_INTEGER,ioid,MPI_COMM_INSTANCE,ierr)
  CALL MPI_BCAST(CHIM,1,MPI_INTEGER,ioid,MPI_COMM_INSTANCE,ierr)
  CALL MPI_BCAST(NSCAL,1,MPI_INTEGER,ioid,MPI_COMM_INSTANCE,ierr)  


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! Read and broadcast the LES parameters.
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  IF (my_fpart.eq.io_fpart) THEN   
    
    OPEN(10,FILE=filename,STATUS='OLD',FORM='FORMATTED')
    READ(unit=10,NML=les_parameters)
    CLOSE(10)

    WRITE(out_unit,*) '>> LES Informations:'
    SELECT CASE (LESTYPE)
    CASE(0)
      WRITE(out_unit,*) '>> DNS Simulation'
      IF(HV_order.NE.0)  HV_order=0
      IF(SGS_type.NE.0)  SGS_type=0 
      IF(filter_type.NE.0)  filter_type=0  
    CASE(1)
      WRITE(out_unit,*) '>> Hyper-viscosity simulation with exponent of nabla equal to :'
      WRITE(out_unit,*) '>>      ', 2**(HV_order+1)
      IF(SGS_type.NE.0)  SGS_type=0
      IF(filter_type.NE.0)  filter_type=0
    CASE(2)
      WRITE(out_unit,*) '>> LES Simulation with the following SGS model:'
      IF(HV_order.NE.0)  HV_order=0  
      SELECT CASE (SGS_type)
      CASE(0)
        WRITE(out_unit,*) '>> SGS_type parameter = 0 is DNS'
      CASE(1)
        SELECT CASE (Smag_dyn)
        CASE(0)
          WRITE(out_unit,*) '>> Constant Smagorinsky model with constant C=', Smag_const
          filter_type=0
          Smag_averag=0
          Smag_dir=0
          Smag_ens_av=0
        CASE(1)
          SELECT CASE (Smag_averag)
          CASE(0)
            WRITE(out_unit,*) '>> Dynamic Smagorinsky with full C(x,y,z,t) field'
          CASE(1)
            SELECT CASE (Smag_dir)
            CASE(1)
              WRITE(out_unit,*) '>> Dynamic Smagorinsky with C field averaged along lines in X direction' 
            CASE(2)
              WRITE(out_unit,*) '>> Dynamic Smagorinsky with C field averaged along lines in Y direction' 
            CASE(3)
              WRITE(out_unit,*) '>> Dynamic Smagorinsky with C field averaged along lines in Z direction' 
            CASE DEFAULT
              WRITE(out_unit,*) '>> Wrong value of Smag_dir in les_parameter namelist:', Smag_dir
            END SELECT
          CASE(2)
            SELECT CASE (Smag_dir)
            CASE(1)
              WRITE(out_unit,*) '>> Dynamic Smagorinsky with C averaged along plane perpendicular to X direction'
            CASE(2)
              WRITE(out_unit,*) '>> Dynamic Smagorinsky with C averaged along plane perpendicular to Y direction'
            CASE(3)
              WRITE(out_unit,*) '>> Dynamic Smagorinsky with C averaged along plane perpendicular to Z direction'
            CASE DEFAULT
              WRITE(out_unit,*) '>> Wrong value of Smag_dir in les_parameter namelist:', Smag_dir
            END SELECT
          CASE(3)
            WRITE(out_unit,*) '>> Dynamic Smagorinsky with C averaged over the full volume'          
          CASE DEFAULT
            WRITE(out_unit,*) '>> WRONG value of Smag_averag in les_parameter namelist:',Smag_averag
            STOP
          END SELECT
          SELECT CASE (filter_type)
          CASE(1)
            WRITE(out_unit,*) '>> Chosen filter for dynamic Smagorinsky is one half'
          CASE(2)
            WRITE(out_unit,*) '>> Chosen filter for dynamic Smagorinsky is one fourth'
          CASE(3)
            WRITE(out_unit,*) '>> Chosen filter for dynamic Smagorinsky is Gaussian one half'
          CASE DEFAULT
            WRITE(out_unit,*) '>> Wrong value of filter_type in les_parameters namelist : ', filter_type
            STOP
          END SELECT
          SELECT CASE (Smag_ens_av)
          CASE(0)
            WRITE(out_unit,*) '>> Single instance Smagorinsky computation'
          CASE(1)
            WRITE(out_unit,*) '>> Ensemble Average Smagorinsky between all instances'
          CASE DEFAULT
            WRITE(out_unit,*) '>> Wrong value of Smag_ens_av in les_parameters namelist:', Smag_ens_av
            STOP
          END SELECT                              
        CASE DEFAULT
          WRITE(out_unit,*) '>> WRONG value of Smag_dyn in les_parameter namelist:', Smag_dyn
          STOP
        END SELECT
      CASE(2)
        WRITE(out_unit,*) '>> Winckelmans Mixed Model'
        SELECT CASE (filter_type)
        CASE(1)
          WRITE(out_unit,*) '>> Chosen filter for Winckelmans is one half'
        CASE(2)
          WRITE(out_unit,*) '>> Chosen filter for Winckelmans is one fourth'
        CASE(3)
          WRITE(out_unit,*) '>> Chosen filter for Winckelmans is Gaussian one half'
        CASE DEFAULT
          WRITE(out_unit,*) '>> Wrong value of filter_type in les_parameters namelist : ', filter_type
          STOP
        END SELECT  
      CASE(3)
        WRITE(out_unit,*) '>> Bardina Model'
        SELECT CASE (filter_type)
        CASE(1)
          WRITE(out_unit,*) '>> Chosen filter for Bardina is one half'
        CASE(2)
          WRITE(out_unit,*) '>> Chosen filter for Bardina is one fourth'
        CASE(3)
          WRITE(out_unit,*) '>> Chosen filter for Bardina is Gaussian one half'
        CASE DEFAULT
          WRITE(out_unit,*) '>> Wrong value of filter_type in les_parameters namelist : ', filter_type
          STOP
        END SELECT                
      CASE DEFAULT
        WRITE(out_unit,*) '>> Wrong value of SGS_type in les_parameters namelist : ', SGS_type
        STOP
      END SELECT
    CASE DEFAULT
      WRITE(out_unit,*) '>> Wrong value of LESTYPE in les_parameters namelist : ', LESTYPE
      STOP
    END SELECT
  END IF
  
  CALL MPI_BCAST(LESTYPE,1,MPI_INTEGER,ioid,MPI_COMM_INSTANCE,ierr)
  CALL MPI_BCAST(HV_order,1,MPI_INTEGER,ioid,MPI_COMM_INSTANCE,ierr)
  CALL MPI_BCAST(SGS_type,1,MPI_INTEGER,ioid,MPI_COMM_INSTANCE,ierr)
  CALL MPI_BCAST(Smag_dyn,1,MPI_INTEGER,ioid,MPI_COMM_INSTANCE,ierr)
  CALL MPI_BCAST(Smag_averag,1,MPI_INTEGER,ioid,MPI_COMM_INSTANCE,ierr)
  CALL MPI_BCAST(Smag_dir,1,MPI_INTEGER,ioid,MPI_COMM_INSTANCE,ierr)
  CALL MPI_BCAST(Smag_ens_av,1,MPI_INTEGER,ioid,MPI_COMM_INSTANCE,ierr)
  CALL MPI_BCAST(Smag_const,1,MPI_MYREAL,ioid,MPI_COMM_INSTANCE,ierr) 
  CALL MPI_BCAST(filter_type,1,MPI_INTEGER,ioid,MPI_COMM_INSTANCE,ierr)
 
 
 
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! Read and broadcast the force restore parameters.
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  IF (my_fpart.eq.io_fpart) THEN       
    OPEN(10,FILE=filename,STATUS='OLD',FORM='FORMATTED')
    READ(unit=10,NML=force_restore_parameters)
    CLOSE(10)
    
    SELECT CASE (exch_tauij)
    CASE(0)
    
    CASE(1)
      WRITE(out_unit,*) '>> Exchanging elements of Tauij'
      field_exchange=1
    CASE DEFAULT
      WRITE(out_unit,*) '>> Wrong value of exch_tauij in force_restore_parameters namelist : ', exch_tauij
      STOP      
    END SELECT
    
    SELECT CASE (type_restore)
    CASE(0)
      WRITE(out_unit,*) '>> No force restore computed'
      add_force_restore=0
    CASE(1)
      WRITE(out_unit,*) '>> Force restore computed = alpha * (u_dns - u_les)'
      field_exchange=1
      IF (add_force_restore.EQ.0) THEN
        WRITE(out_unit,*) '>>   and force restore not added'
      ELSE
        WRITE(out_unit,*) '>>   and force restore added'
      END IF
    CASE(2)
      WRITE(out_unit,*) '>> Force restore computed = alpha * ksquare * (u_dns - u_les)'
      field_exchange=1
      IF (add_force_restore.EQ.0) THEN
        WRITE(out_unit,*) '>>   and force restore not added'
      ELSE
        WRITE(out_unit,*) '>>   and force restore added'
      END IF
    CASE DEFAULT
      WRITE(out_unit,*) '>> Wrong value of type_restore in force_restore_parameters namelist : ', type_restore
      STOP
    END SELECT
  END IF
  
  CALL MPI_BCAST(exch_tauij,1,MPI_INTEGER,ioid,MPI_COMM_INSTANCE,ierr)
  CALL MPI_BCAST(add_total,1,MPI_INTEGER,ioid,MPI_COMM_INSTANCE,ierr)
  CALL MPI_BCAST(add_1,1,MPI_INTEGER,ioid,MPI_COMM_INSTANCE,ierr)
  CALL MPI_BCAST(add_2,1,MPI_INTEGER,ioid,MPI_COMM_INSTANCE,ierr)
  CALL MPI_BCAST(add_3,1,MPI_INTEGER,ioid,MPI_COMM_INSTANCE,ierr)
  CALL MPI_BCAST(add_4,1,MPI_INTEGER,ioid,MPI_COMM_INSTANCE,ierr)
  CALL MPI_BCAST(add_5,1,MPI_INTEGER,ioid,MPI_COMM_INSTANCE,ierr)
  CALL MPI_BCAST(proj_u,1,MPI_INTEGER,ioid,MPI_COMM_INSTANCE,ierr)
  CALL MPI_BCAST(LES_ident,1,MPI_INTEGER,ioid,MPI_COMM_INSTANCE,ierr)
  CALL MPI_BCAST(type_restore,1,MPI_INTEGER,ioid,MPI_COMM_INSTANCE,ierr)
  CALL MPI_BCAST(add_force_restore,1,MPI_INTEGER,ioid,MPI_COMM_INSTANCE,ierr)
  CALL MPI_BCAST(alpha_restore,1,MPI_MYREAL,ioid,MPI_COMM_INSTANCE,ierr)

  CALL MPI_BCAST(field_exchange,1,MPI_INTEGER,ioid,MPI_COMM_INSTANCE,ierr)
  CALL MPI_ALLREDUCE(field_exchange,global_exchange,1,MPI_INTEGER,MPI_PROD,MPI_COMM_WORLD,ierr)
  IF(field_exchange.NE.global_exchange) THEN
    WRITE(out_unit,*) '>> ERROR in field exchange, not all the nodes are set to exchange field'
    STOP
  END IF
  
  

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! Read and broadcast the rotation parameters
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  IF (my_fpart.eq.io_fpart) THEN 
    OPEN(10,FILE=filename,STATUS='OLD',FORM='FORMATTED')
    READ(unit=10,NML=rot_parameters)
    CLOSE(10)
  END IF
  
  CALL MPI_BCAST(Omega_x,1,MPI_MYREAL,ioid,MPI_COMM_INSTANCE,ierr)
  CALL MPI_BCAST(Omega_y,1,MPI_MYREAL,ioid,MPI_COMM_INSTANCE,ierr)
  CALL MPI_BCAST(Omega_z,1,MPI_MYREAL,ioid,MPI_COMM_INSTANCE,ierr)

  IF((ABS(Omega_x).lt.EPS).AND.(ABS(Omega_y).lt.EPS).AND.(ABS(Omega_z).lt.EPS)) THEN
    ROTEQ = .FALSE.
  ELSE
    ROTEQ = .TRUE.  
  END IF
  
  CALL MPI_BCAST(ROTEQ,1,MPI_LOGICAL,ioid,MPI_COMM_INSTANCE,ierr)
  
  IF (ROTEQ .and. my_fpart.eq.io_fpart) THEN
    WRITE(out_unit,*)'>> SOLID-BODY rotation : ' 
    WRITE(out_unit,*)'>>   angular velocity along X: ', Omega_x
    WRITE(out_unit,*)'>>   angular velocity along Y: ', Omega_y
    WRITE(out_unit,*)'>>   angular velocity along Z: ', Omega_z    
  END IF

    
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! Read and broadcast the velocity field parameters if required.
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  IF (VELEQ) THEN 
    IF (my_fpart.eq.io_fpart) THEN 
      OPEN(10,FILE=filename,STATUS='OLD',FORM='FORMATTED')
      READ(unit=10,NML=vel_parameters)
      CLOSE(10)
    END IF
    
    CALL MPI_BCAST(init_vel,1,MPI_INTEGER,ioid,MPI_COMM_INSTANCE,ierr)  
    CALL MPI_BCAST(nu,1,MPI_MYREAL,ioid,MPI_COMM_INSTANCE,ierr)  
    CALL MPI_BCAST(u0_x,1,MPI_MYREAL,ioid,MPI_COMM_INSTANCE,ierr)
    CALL MPI_BCAST(u0_y,1,MPI_MYREAL,ioid,MPI_COMM_INSTANCE,ierr)
    CALL MPI_BCAST(u0_z,1,MPI_MYREAL,ioid,MPI_COMM_INSTANCE,ierr)
    CALL MPI_BCAST(vel_a,1,MPI_MYREAL,ioid,MPI_COMM_INSTANCE,ierr)  
    CALL MPI_BCAST(vel_b,1,MPI_MYREAL,ioid,MPI_COMM_INSTANCE,ierr)  
    CALL MPI_BCAST(vel_c,1,MPI_MYREAL,ioid,MPI_COMM_INSTANCE,ierr)  
    CALL MPI_BCAST(vel_d,1,MPI_MYREAL,ioid,MPI_COMM_INSTANCE,ierr)

    IF (u0_x.NE.0) WRITE(out_unit,*)'>> Mean velocity field on Ox: ', u0_x 
    IF (u0_y.NE.0) WRITE(out_unit,*)'>> Mean velocity field on Oy: ', u0_y 
    IF (u0_z.NE.0) WRITE(out_unit,*)'>> Mean velocity field on Oz: ', u0_z
    

  END IF


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! Read and broadcast the magnetic field parameters if required.
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  IF (MAGEQ.OR.EQTYPE==qs_mhd) THEN 
    IF (my_fpart.eq.io_fpart) THEN 
      OPEN(10,FILE=filename,STATUS='OLD',FORM='FORMATTED')
      READ(unit=10,NML=mag_parameters)
      CLOSE(10)
    END IF
    
    CALL MPI_BCAST(init_mag,1,MPI_INTEGER,ioid,MPI_COMM_INSTANCE,ierr)  
    CALL MPI_BCAST(eta,1,MPI_MYREAL,ioid,MPI_COMM_INSTANCE,ierr)  
    CALL MPI_BCAST(b0_x,1,MPI_MYREAL,ioid,MPI_COMM_INSTANCE,ierr)
    CALL MPI_BCAST(b0_y,1,MPI_MYREAL,ioid,MPI_COMM_INSTANCE,ierr)
    CALL MPI_BCAST(b0_z,1,MPI_MYREAL,ioid,MPI_COMM_INSTANCE,ierr)
    CALL MPI_BCAST(mag_a,1,MPI_MYREAL,ioid,MPI_COMM_INSTANCE,ierr)  
    CALL MPI_BCAST(mag_b,1,MPI_MYREAL,ioid,MPI_COMM_INSTANCE,ierr)  
    CALL MPI_BCAST(mag_c,1,MPI_MYREAL,ioid,MPI_COMM_INSTANCE,ierr)  
    CALL MPI_BCAST(mag_d,1,MPI_MYREAL,ioid,MPI_COMM_INSTANCE,ierr) 
    
    IF (b0_x.NE.0) WRITE(out_unit,*)'>> Constant external magnetic field on Ox: ', b0_x 
    IF (b0_y.NE.0) WRITE(out_unit,*)'>> Constant external magnetic field on Oy: ', b0_y 
    IF (b0_z.NE.0) WRITE(out_unit,*)'>> Constant external magnetic field on Oz: ', b0_z 
  END IF


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! Read and broadcast the scalar field parameters if required.
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  IF (SCAEQ) THEN 
    ALLOCATE(init_sca(NSCAL),kappa(NSCAL),sca_a(NSCAL),sca_b(NSCAL),sca_c(NSCAL),sca_d(NSCAL))
    IF (my_fpart.eq.io_fpart) THEN 
      OPEN(10,FILE=filename,STATUS='OLD',FORM='FORMATTED')
      READ(unit=10,NML=sca_parameters)
      CLOSE(10)
      init_sca = init_sca_dat(1:NSCAL)
      kappa=kappa_dat(1:NSCAL)
      sca_a = sca_a_dat(1:NSCAL)
      sca_b = sca_b_dat(1:NSCAL)
      sca_c = sca_c_dat(1:NSCAL)
      sca_d = sca_d_dat(1:NSCAL)
    END IF
    
    CALL MPI_BCAST(init_sca,NSCAL,MPI_INTEGER,ioid,MPI_COMM_INSTANCE,ierr)
    CALL MPI_BCAST(kappa,NSCAL,MPI_MYREAL,ioid,MPI_COMM_INSTANCE,ierr)
    CALL MPI_BCAST(sca_a,NSCAL,MPI_MYREAL,ioid,MPI_COMM_INSTANCE,ierr)
    CALL MPI_BCAST(sca_b,NSCAL,MPI_MYREAL,ioid,MPI_COMM_INSTANCE,ierr)
    CALL MPI_BCAST(sca_c,NSCAL,MPI_MYREAL,ioid,MPI_COMM_INSTANCE,ierr)
    CALL MPI_BCAST(sca_d,NSCAL,MPI_MYREAL,ioid,MPI_COMM_INSTANCE,ierr)
  END IF


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!  
! Read and broadcast the computational box sizes and the grid
! sizes.
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  IF (my_fpart.eq.io_fpart) THEN 
    OPEN(10,FILE=filename,STATUS='OLD',FORM='FORMATTED')
    READ(unit=10,NML=dim_and_sizes)
    CLOSE(10)
    WRITE(out_unit,*) '>> Number of points are ', nx,'x',ny,'x',nz
    WRITE(out_unit,*) '>> Box sizes / pi are:'
    WRITE(out_unit,*) '>> ', lx,'x',ly,'x',lz
    lx=lx*pi
    ly=ly*pi
    lz=lz*pi  
  END IF
  
  CALL MPI_BCAST(nx,1,MPI_INTEGER,ioid,MPI_COMM_INSTANCE,ierr)
  CALL MPI_BCAST(ny,1,MPI_INTEGER,ioid,MPI_COMM_INSTANCE,ierr)
  CALL MPI_BCAST(nz,1,MPI_INTEGER,ioid,MPI_COMM_INSTANCE,ierr)
  CALL MPI_BCAST(lx,1,MPI_MYREAL,ioid,MPI_COMM_INSTANCE,ierr)
  CALL MPI_BCAST(ly,1,MPI_MYREAL,ioid,MPI_COMM_INSTANCE,ierr)
  CALL MPI_BCAST(lz,1,MPI_MYREAL,ioid,MPI_COMM_INSTANCE,ierr)


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! Read and broadcast the forcing parameters for random force
! in a shell if and only if the velocity equation has to be solved
! and when FORCEPARA parameter is set to 1 (constant injection rate),
! 2 (constant levels) and 4 Kologorov forcing
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  IF ((my_fpart.eq.io_fpart).and.VELEQ) THEN 
    SELECT CASE (FORCEPARA)
    CASE(1,2,4,6)
      OPEN(10,FILE=filename,STATUS='OLD',FORM='FORMATTED')
      READ(unit=10,NML=forcing_parameters)
      CLOSE(10)
      IF (helicity_para**2>1) THEN
        WRITE(out_unit,*)'>> Wrong value of the helicity_para parameter in the forcing_parameters namelist.'
        STOP
      END IF
      IF ((FORCEPARA==4).AND.((ny/2.LT.Fmode).or.(MOD(ny/4,Fmode).NE.0))) THEN
        WRITE(out_unit,*) '>> Kolmogorov forced mode incompatible with the resolution'
        STOP
      END IF
    CASE DEFAULT
    END SELECT
  END IF

  CALL MPI_BCAST(kinf,1,MPI_MYREAL,ioid,MPI_COMM_INSTANCE,ierr)
  CALL MPI_BCAST(ksup,1,MPI_MYREAL,ioid,MPI_COMM_INSTANCE,ierr)
  CALL MPI_BCAST(energy_forcing,1,MPI_MYREAL,ioid,MPI_COMM_INSTANCE,ierr)
  CALL MPI_BCAST(helicity_para,1,MPI_MYREAL,ioid,MPI_COMM_INSTANCE,ierr)
  CALL MPI_BCAST(Famp,1,MPI_MYREAL,ioid,MPI_COMM_INSTANCE,ierr)
  CALL MPI_BCAST(Fmode,1,MPI_INTEGER,ioid,MPI_COMM_INSTANCE,ierr)


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! Read and broadcast the forcing_ub parameters 
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  IF ((my_fpart.eq.io_fpart).and.VELEQ) THEN 
    SELECT CASE (FORCEPARA)
    CASE(7)
      OPEN(10,FILE=filename,STATUS='OLD',FORM='FORMATTED')
      READ(unit=10,NML=forcing_ub_parameters)
      CLOSE(10)
      IF (kin_hel_forcing**2>1) THEN
        WRITE(6,*)'>> Wrong value of the helicity_para parameter in the forcing_parameters namelist.'
        STOP
      END IF
      IF (mag_hel_forcing**2>1) THEN
        WRITE(6,*)'>> Wrong value of the crosshel_para parameter in the forcing_parameters namelist.'
        STOP
      END IF  
      IF (croshel_forcing**2>1) THEN
        WRITE(6,*)'>> Wrong value of the crosshel_para parameter in the forcing_parameters namelist.'
        STOP
      END IF 	      
    CASE DEFAULT
    END SELECT
  END IF

  CALL MPI_BCAST(kinf,1,MPI_MYREAL,ioid,MPI_COMM_INSTANCE,ierr)
  CALL MPI_BCAST(ksup,1,MPI_MYREAL,ioid,MPI_COMM_INSTANCE,ierr)
  CALL MPI_BCAST(kin_energy_forcing,1,MPI_MYREAL,ioid,MPI_COMM_INSTANCE,ierr)
  CALL MPI_BCAST(mag_energy_forcing,1,MPI_MYREAL,ioid,MPI_COMM_INSTANCE,ierr)
  CALL MPI_BCAST(kin_hel_forcing,1,MPI_MYREAL,ioid,MPI_COMM_INSTANCE,ierr)
  CALL MPI_BCAST(mag_hel_forcing,1,MPI_MYREAL,ioid,MPI_COMM_INSTANCE,ierr)  
  CALL MPI_BCAST(croshel_forcing,1,MPI_MYREAL,ioid,MPI_COMM_INSTANCE,ierr)  


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! Read and broadcast the time numerics parameters 
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  IF (my_fpart.eq.io_fpart) THEN 
    OPEN(10,FILE=filename,STATUS='OLD',FORM='FORMATTED')
    READ(unit=10,NML=numerics_parameters)
    CLOSE(10)
    SELECT CASE (dealias)
    CASE(shifts)
      WRITE(out_unit,*)'>> This run uses a third order, 4 step RK method with phase shift dealiasing'
    CASE(twothird)
      WRITE(out_unit,*)'>> This run uses a third order, 3 step RK method with 2/3 dealiasing method'
    CASE DEFAULT
      WRITE(out_unit,*)'>> Wrong value of the dealias parameter in the numerics_parameters namelist.'
      STOP
    END SELECT
    SELECT CASE (fixed_dt)
    CASE(0)
      WRITE(out_unit,*)'>> This run automatically computes the time increment'
    CASE(1)
      WRITE(out_unit,*)'>> This run uses a fixed time increment : ', dt_dat
    CASE DEFAULT
      WRITE(out_unit,*)'>> Wrong value of the dealias parameter in the numerics_parameters namelist.'
      STOP
    END SELECT
  END IF

  CALL MPI_BCAST(dt_dat  ,1,MPI_MYREAL ,ioid,MPI_COMM_INSTANCE,ierr)
  CALL MPI_BCAST(cfl     ,1,MPI_MYREAL ,ioid,MPI_COMM_INSTANCE,ierr)
  CALL MPI_BCAST(fixed_dt,1,MPI_INTEGER,ioid,MPI_COMM_INSTANCE,ierr)
  CALL MPI_BCAST(dealias ,1,MPI_INTEGER,ioid,MPI_COMM_INSTANCE,ierr)
  CALL MPI_BCAST(userseed,1,MPI_INTEGER,ioid,MPI_COMM_INSTANCE,ierr)


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! Read and broadcast the save-stat parameters 
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  IF (my_fpart.eq.io_fpart) THEN 
    OPEN(10,FILE=filename,STATUS='OLD',FORM='FORMATTED')
    READ(unit=10,NML=save_parameters)
    CLOSE(10)
    WRITE(out_unit,*) '>> Total number of iterations in the run = ', niter_todo
    WRITE(out_unit,*) '>> Number of iterations in this job      = ', nstep
    WRITE(out_unit,*) '>> Save file and statistics every ', nsave,' iterations'
    WRITE(out_unit,*) '>> Compute statistics every       ', nstat,' iterations' 
    WRITE(out_unit,*) '>> Uses       ', nbuild,' iterations for building the phases' 
    
    IF (MOD(ny,totpart).ne.0) THEN 
      WRITE(out_unit,*) '>> STOP : ny is not an integrer times totpart '
      STOP
    END IF
    WRITE(out_unit,*) '>> Number of files used to store a frame = ', totpart
    
    IF (MOD(nstep,nsave).ne.0) THEN 
      WRITE(out_unit,*) '>> Please chose nstep as an integer times nsave '
      STOP
    END IF
    
    IF (MOD(nstep,nstat).ne.0) THEN 
      WRITE(out_unit,*) '>> Please chose nstep as an integer times nstat '
      STOP
    END IF
  END IF

  CALL MPI_BCAST(niter_todo,1,MPI_INTEGER,ioid,MPI_COMM_INSTANCE,ierr)
  CALL MPI_BCAST(nstep     ,1,MPI_INTEGER,ioid,MPI_COMM_INSTANCE,ierr)
  CALL MPI_BCAST(nsave     ,1,MPI_INTEGER,ioid,MPI_COMM_INSTANCE,ierr)
  CALL MPI_BCAST(nstat     ,1,MPI_INTEGER,ioid,MPI_COMM_INSTANCE,ierr)
  CALL MPI_BCAST(nbuild    ,1,MPI_INTEGER,ioid,MPI_COMM_INSTANCE,ierr)
  CALL MPI_BCAST(totpart   ,1,MPI_INTEGER,ioid,MPI_COMM_INSTANCE,ierr)
  
END SUBROUTINE read_para



!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!                                                          !!
!!              transfer_slice_world                        !!
!!                                                          !!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
SUBROUTINE transfer_slice_world(slice,fromid,destination,destid)
  USE parameters
  USE mpi_var
  IMPLICIT NONE

  COMPLEX, DIMENSION(0:cx,0:cz), INTENT(IN)    :: slice
  COMPLEX, DIMENSION(0:cx,0:cz), INTENT(OUT)   :: destination
  INTEGER, INTENT(IN)                          :: fromid, destid

  IF (destid.ne.fromid) THEN
    IF (myid.eq.fromid) CALL MPI_SEND(slice,(cx+1)*(cz+1),MPI_MYCOMPLEX,destid,10,MPI_COMM_WORLD,ierr)
    IF (myid.eq.destid) CALL MPI_RECV(destination,(cx+1)*(cz+1),MPI_MYCOMPLEX,fromid,10,MPI_COMM_WORLD,status_mpi,ierr)
  ELSE
    IF (myid.eq.destid) destination=slice
  END IF

END SUBROUTINE transfer_slice_world



!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!                                                          !!
!!                     transfer_slice                       !!
!!                                                          !!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
SUBROUTINE transfer_slice(slice,fromid,destination,destid)
  USE parameters
  USE mpi_var
  IMPLICIT NONE

  COMPLEX, DIMENSION(0:cx,0:cz), INTENT(IN)    :: slice
  COMPLEX, DIMENSION(0:cx,0:cz), INTENT(OUT)   :: destination
  INTEGER, INTENT(IN)                          :: fromid, destid

  IF (destid.ne.fromid) THEN
    IF (my_fpart.eq.fromid) CALL MPI_SEND(slice      ,(cx+1)*(cz+1),MPI_MYCOMPLEX,destid,10,MPI_COMM_INSTANCE,ierr)
    IF (my_fpart.eq.destid) CALL MPI_RECV(destination,(cx+1)*(cz+1),MPI_MYCOMPLEX,fromid,10,MPI_COMM_INSTANCE,status_mpi,ierr)
  ELSE
    IF (my_fpart.eq.destid) destination=slice
  END IF

END SUBROUTINE transfer_slice



