!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!  12 August 2011                                          !!
!!                      tt_les.f90                          !!
!!                                                          !!
!!  compute_sgs                                             !!
!!  sgs_smagorinsky                                         !!
!!  sgs_winckelmans                                         !!
!!  sgs_bardina                                             !!
!!  compute_strain                                          !!
!!  compute_rotation                                        !!
!!  filter                                                  !!
!!  compute_leonard                                         !!
!!  do_diuj                                                 !!
!!  sgs_exchanging_tauij                                    !!
!!  reduce_send_and_receive                                 !!
!!  compute_force_restore                                   !!
!!  exchange_large_scales                                   !!
!!                                             version 2.05 !!
!!                                                          !!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!



!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!                                                          !!
!!                      compute_sgs                         !!
!!                                                          !!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
SUBROUTINE compute_sgs(force_sgs)
  USE parameters
  IMPLICIT NONE

  COMPLEX,  DIMENSION(0:cx,0:cz,0:cy,3), INTENT(OUT) ::  force_sgs

  SELECT CASE (SGS_type)
  CASE(1)
    CALL sgs_smagorinsky(force_sgs)
  CASE(2)
    CALL sgs_winckelmans(force_sgs)
  CASE(3)
    CALL sgs_bardina(force_sgs)
  END SELECT
  
  CALL divfree(force_sgs)
  
END SUBROUTINE compute_sgs



!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!                                                          !!
!!                    sgs_smagorinsky                       !!
!!                                                          !!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
SUBROUTINE sgs_smagorinsky(a)
  USE parameters
  USE variables
  USE mpi_var
  IMPLICIT NONE

  COMPLEX,  DIMENSION(0:cx,0:cz,0:cy,3), INTENT(OUT)   ::  a
  COMPLEX,  DIMENSION(:,:,:,:),  ALLOCATABLE           ::  af
  COMPLEX,  DIMENSION(:,:,:,:),  ALLOCATABLE           ::  tauij
  REAL,     DIMENSION(:,:,:,:),  ALLOCATABLE           ::  Sfij, Sij, Mij, Lij
  REAL,     DIMENSION(:,:,:),    ALLOCATABLE           ::  Snorm, Sfnorm, Sup, Inf, Sup_fpart, Inf_fpart
  REAL,     DIMENSION(:,:,:),    ALLOCATABLE           ::  C                
  REAL,     DIMENSION(:,:,:),    ALLOCATABLE           ::  r1
  COMPLEX,  DIMENSION(:,:,:),    ALLOCATABLE           ::  c1
  REAL      ::  alph1, Delta, AverageSup, AverageInf, Sup_node, Inf_node, nu_tot
  INTEGER   ::  qx, qy, qz, j  

  Delta = MIN(dx,dy,dz)
  !Delta=(dx*dy*dz)^(1.0/3.0)
  ALLOCATE(Sij(0:rx,0:ry,0:rz,6))
  CALL compute_strain(vel, Sij)  
  
  ALLOCATE(Snorm(0:rx,0:ry,0:rz))                    
  Snorm  = Sij(:,:,:,1)**2 + 2.0*Sij(:,:,:,2)**2 + 2.0 * Sij(:,:,:,3)**2 + Sij(:,:,:,4)**2 + 2.0*Sij(:,:,:,5)**2 + Sij(:,:,:,6)**2
  Snorm  = SQRT(2.0*Snorm)
  
  IF (Smag_dyn.EQ.0) THEN
  
    ALLOCATE(r1(0:rx,0:ry,0:rz))
    ALLOCATE(c1(0:cx,0:cz,0:cy))
    ALLOCATE(tauij(0:cx,0:cz,0:cy,6))
    ALLOCATE(C(1,1,1))
    Delta = (dx*dy*dz)**(1.0/3.0) 
    C(1,1,1) = Smag_const
    DO j =1, 6
      r1 = -2.0*(C(1,1,1)*Delta)**2*Snorm*Sij(:,:,:,j)          
      CALL fft3ds(r1,c1)
      tauij(:,:,:,j)=c1
    END DO 
    DEALLOCATE(C)
    DEALLOCATE(r1)
    DEALLOCATE(c1)
    
  ELSE    

    ALLOCATE(c1(0:cx,0:cz,0:cy))
    ALLOCATE(r1(0:rx,0:ry,0:rz))
    ALLOCATE(af(0:cx,0:cz,0:cy,3))
    CALL filter(vel(:,:,:,1), af(:,:,:,1))
    CALL filter(vel(:,:,:,2), af(:,:,:,2))
    CALL filter(vel(:,:,:,3), af(:,:,:,3))
    ALLOCATE(Sfij(0:rx,0:ry,0:rz,6))
    CALL compute_strain(af,Sfij)
    DEALLOCATE(af)
    ALLOCATE(Sfnorm(0:rx,0:ry,0:rz))
    Sfnorm = Sfij(:,:,:,1)**2 + 2.0*Sfij(:,:,:,2)**2 + 2.0*Sfij(:,:,:,3)**2 + Sfij(:,:,:,4)**2 + 2.0*Sfij(:,:,:,5)**2 + Sfij(:,:,:,6)**2
    Sfnorm = SQRT(2.0*Sfnorm)

  SELECT CASE (filter_type)
  CASE(1)
    alph1=2.0
  CASE(2)
    alph1=4.0
  CASE(3)
    alph1=2.0
  END SELECT
    ALLOCATE(Mij(0:rx,0:ry,0:rz,6))
    DO j =1, 6 
      r1 = Snorm*Sij(:,:,:,j) 
      CALL fft3ds(r1,c1)
      CALL filter(c1,c1)
      CALL ifft3ds(c1,r1)
      Mij(:,:,:,j)   = 2.0*(r1 - alph1**2 * Sfnorm*Sfij(:,:,:,j))
    END DO        
    DEALLOCATE(Sfnorm)
    DEALLOCATE(Sfij)

    ALLOCATE(Lij(0:rx,0:ry,0:rz,6))
    CALL compute_leonard(vel,Lij)
    
    ALLOCATE(Sup(0:rx,0:ry,0:rz))
    ALLOCATE(Inf(0:rx,0:ry,0:rz))  
    
    IF (Smag_ens_av.EQ.1) THEN
      ALLOCATE(Sup_fpart(0:rx,0:ry,0:rz))
      ALLOCATE(Inf_fpart(0:rx,0:ry,0:rz))    
      Sup_fpart = Lij(:,:,:,1)*Mij(:,:,:,1) + 2.0*Lij(:,:,:,2)*Mij(:,:,:,2) + 2.0*Lij(:,:,:,3)*Mij(:,:,:,3) + Lij(:,:,:,4)*Mij(:,:,:,4) + 2.0*Lij(:,:,:,5)*Mij(:,:,:,5) + Lij(:,:,:,6)*Mij(:,:,:,6)
      Inf_fpart = Mij(:,:,:,1)**2 + 2.0*Mij(:,:,:,2)**2 + 2.0*Mij(:,:,:,3)**2 + Mij(:,:,:,4)**2 + 2.0*Mij(:,:,:,5)**2 + Mij(:,:,:,6)**2    
      CALL mpi_allreduce(Sup_fpart, Sup, (rx+1)*(ry+1)*(rz+1),MPI_MYREAL,MPI_SUM,MPI_COMM_FPART,ierr)
      CALL mpi_allreduce(Inf_fpart, Inf, (rx+1)*(ry+1)*(rz+1),MPI_MYREAL,MPI_SUM,MPI_COMM_FPART,ierr)
      DEALLOCATE(Sup_fpart)
      DEALLOCATE(Inf_fpart)
    ELSE
      Sup = Lij(:,:,:,1)*Mij(:,:,:,1) + 2.0*Lij(:,:,:,2)*Mij(:,:,:,2) + 2.0*Lij(:,:,:,3)*Mij(:,:,:,3) + Lij(:,:,:,4)*Mij(:,:,:,4) + 2.0*Lij(:,:,:,5)*Mij(:,:,:,5) + Lij(:,:,:,6)*Mij(:,:,:,6)
      Inf = Mij(:,:,:,1)**2 + 2.0*Mij(:,:,:,2)**2 + 2.0*Mij(:,:,:,3)**2 + Mij(:,:,:,4)**2 + 2.0*Mij(:,:,:,5)**2 + Mij(:,:,:,6)**2  
    END IF
    
    DEALLOCATE(Mij)
    DEALLOCATE(Lij)    

    SELECT CASE (Smag_averag)
    CASE(0)                  !    Full Field
      DO qz=0, rz
        DO qy=0, ry
          DO qx=0, rx
            IF(ABS(Inf(qx,qy,qz)).LT.EPS) THEN
              IF(Inf(qx,qy,qz).LT.0.0) THEN
                Inf(qx,qy,qz)=-EPS
              ELSE
                Inf(qx,qy,qz)=EPS
              END IF
            END IF
          END DO
        END DO
      END DO    
      ALLOCATE(C(0:rx,0:ry,0:rz))
      C=Sup/Inf  
      DEALLOCATE(Sup)
      DEALLOCATE(Inf)
      DO qz=0, rz ! Clipping if nu_tot .LT. 0
        DO qy=0, ry
          DO qx=0, rx
            nu_tot=nu+2*C(qx,qy,qz)*Snorm(qx,qy,qz)
            IF(nu_tot.LT.0.0) C(qx,qy,qz) = - nu / (2*Snorm(qx,qy,qz))
          END DO
        END DO
      END DO

      ALLOCATE(tauij(0:cx,0:cz,0:cy,6))
      DO j =1, 6
        r1=-2.0*C*Snorm*Sij(:,:,:,j)
        CALL fft3ds(r1,c1)
        tauij(:,:,:,j)=c1
      END DO        
    
    CASE(1)                  !    Line averaged
      SELECT CASE (Smag_dir)
      CASE(1)  ! averaged along direction X
        ALLOCATE(C(1,0:ry,0:rz))
        DO qz=0,rz
          DO qy=0,ry
            AverageSup=SUM(Sup(:,qy,qz))
            AverageInf=SUM(Inf(:,qy,qz))
            C(1,qy,qz)= AverageSup/AverageInf
          END DO
        END DO
        DEALLOCATE(Sup)
        DEALLOCATE(Inf)
        ALLOCATE(tauij(0:cx,0:cz,0:cy,6))
        DO j =1, 6
          DO qz=0,rz
            DO qy=0,ry
              r1(:,qy,qz)=-2.0*C(1,qy,qz)*Snorm(:,qy,qz)*Sij(:,qy,qz,j)
            END DO
          END DO
          CALL fft3ds(r1,c1)
          tauij(:,:,:,j)=c1(:,:,:)
        END DO
        
      CASE(2)  ! averaged along direction Y      
        ALLOCATE(C(0:rx,1,0:rz))
        DO qz=0,rz
          DO qx=0,rx
            AverageSup=SUM(Sup(qx,:,qz))
            AverageInf=SUM(Inf(qx,:,qz))
            C(qx,1,qz)= AverageSup/AverageInf
          END DO
        END DO
        DEALLOCATE(Sup)
        DEALLOCATE(Inf)
        ALLOCATE(tauij(0:cx,0:cz,0:cy,6))
        DO j =1, 6
          DO qz=0,rz
            DO qx=0,rx
              r1(qx,:,qz)=-2.0*C(qx,1,qz)*Snorm(qx,:,qz)*Sij(qx,:,qz,j)
            END DO
          END DO
          CALL fft3ds(r1,c1)
          tauij(:,:,:,j)=c1(:,:,:)
        END DO
        
      CASE(3)  ! averaged along direction Z      
        ALLOCATE(C(0:rx,0:ry,1))
        DO qy=0,ry
          DO qx=0,rx
            Sup_node=SUM(Sup(qx,qy,:))
            Inf_node=SUM(Inf(qx,qy,:))
            CALL mpi_allreduce(Sup_node,AverageSup,1,MPI_MYREAL,MPI_SUM,MPI_COMM_INSTANCE,ierr)
            CALL mpi_allreduce(Inf_node,AverageInf,1,MPI_MYREAL,MPI_SUM,MPI_COMM_INSTANCE,ierr)             
            C(qx,qy,1)= AverageSup/AverageInf
          END DO
        END DO
        DEALLOCATE(Sup)
        DEALLOCATE(Inf)
        ALLOCATE(tauij(0:cx,0:cz,0:cy,6))
        DO j =1, 6
          DO qy=0,ry
            DO qx=0,rx
              r1(qx,qy,:)=-2.0*C(qx,qy,1)*Snorm(qx,qy,:)*Sij(qx,qy,:,j)
            END DO
          END DO
          CALL fft3ds(r1,c1)
          tauij(:,:,:,j)=c1(:,:,:)
        END DO
      END SELECT
    
    CASE(2)                  !    Plane averaged
      SELECT CASE (Smag_dir)
      CASE(1)  ! averaged perpendicularly to direction X
        ALLOCATE(C(0:rx,1,1))
        DO qx=0,rx
          Sup_node=SUM(Sup(qx,:,:))
          Inf_node=SUM(Inf(qx,:,:))
          CALL mpi_allreduce(Sup_node,AverageSup,1,MPI_MYREAL,MPI_SUM,MPI_COMM_INSTANCE,ierr)
          CALL mpi_allreduce(Inf_node,AverageInf,1,MPI_MYREAL,MPI_SUM,MPI_COMM_INSTANCE,ierr)             
          C(qx,1,1)= AverageSup/AverageInf
        END DO
        DEALLOCATE(Sup)
        DEALLOCATE(Inf)
        ALLOCATE(tauij(0:cx,0:cz,0:cy,6))
        DO j =1, 6
          DO qx=0,rx
            r1(qx,:,:)=-2.0*C(qx,1,1)*Snorm(qx,:,:)*Sij(qx,:,:,j)
          END DO
          CALL fft3ds(r1,c1)
          tauij(:,:,:,j)=c1(:,:,:)
        END DO
      
      CASE(2)  ! averaged perpendicularly to direction Y
        ALLOCATE(C(1,0:ry,1))
        DO qy=0,ry
          Sup_node=SUM(Sup(:,qy,:))
          Inf_node=SUM(Inf(:,qy,:))
          CALL mpi_allreduce(Sup_node,AverageSup,1,MPI_MYREAL,MPI_SUM,MPI_COMM_INSTANCE,ierr)
          CALL mpi_allreduce(Inf_node,AverageInf,1,MPI_MYREAL,MPI_SUM,MPI_COMM_INSTANCE,ierr)             
          C(1,qy,1)= AverageSup/AverageInf
        END DO
        DEALLOCATE(Sup)
        DEALLOCATE(Inf)
        ALLOCATE(tauij(0:cx,0:cz,0:cy,6))
        DO j =1, 6
          DO qy=0,ry
            r1(:,qy,:)=-2.0*C(1,qy,1)*Snorm(:,qy,:)*Sij(:,qy,:,j)
          END DO
          CALL fft3ds(r1,c1)
          tauij(:,:,:,j)=c1(:,:,:)
        END DO
      
      CASE(3)  ! averaged perpendicularly to direction Z
        ALLOCATE(C(1,1,0:rz))
        DO qz=0,rz
          AverageSup=SUM(Sup(:,:,qz))
          AverageInf=SUM(Inf(:,:,qz))
          C(1,1,qz)= AverageSup/AverageInf
        END DO
        DEALLOCATE(Sup)
        DEALLOCATE(Inf)
        ALLOCATE(tauij(0:cx,0:cz,0:cy,6))
        DO j =1, 6
          DO qz=0,rz
            r1(:,:,qz)=-2.0*C(1,1,qz)*Snorm(:,:,qz)*Sij(:,:,qz,j)
          END DO
          CALL fft3ds(r1,c1)
          tauij(:,:,:,j)=c1(:,:,:)
        END DO
      
      END SELECT    
    
    CASE(3)                  !    Volume averaged
      Sup_node=SUM(Sup)
      CALL mpi_allreduce(Sup_node,AverageSup,1,MPI_MYREAL,MPI_SUM,MPI_COMM_INSTANCE,ierr)
      Inf_node=SUM(Inf)
      CALL mpi_allreduce(Inf_node,AverageInf,1,MPI_MYREAL,MPI_SUM,MPI_COMM_INSTANCE,ierr) 
      ALLOCATE(C(1,1,1))
      C(1,1,1)= AverageSup/AverageInf  
            
      DEALLOCATE(Sup)
      DEALLOCATE(Inf)
    
      ALLOCATE(tauij(0:cx,0:cz,0:cy,6))
      DO j =1, 6
        r1=-2.0*C(1,1,1)*Snorm*Sij(:,:,:,j)
        CALL fft3ds(r1,c1) 
        tauij(:,:,:,j)=c1
      END DO
    
    END SELECT
  
    DEALLOCATE(C)
    DEALLOCATE(r1)
    DEALLOCATE(c1)
  END IF

  DO qy=0, cy
    DO qz=0, cz
      DO qx=0, cx
        a(qx,qz,qy,1) = - ikx(qx)*tauij(qx,qz,qy,1) - iky(qy)*tauij(qx,qz,qy,2) - ikz(qz)*tauij(qx,qz,qy,3)
        a(qx,qz,qy,2) = - ikx(qx)*tauij(qx,qz,qy,2) - iky(qy)*tauij(qx,qz,qy,4) - ikz(qz)*tauij(qx,qz,qy,5)
        a(qx,qz,qy,3) = - ikx(qx)*tauij(qx,qz,qy,3) - iky(qy)*tauij(qx,qz,qy,5) - ikz(qz)*tauij(qx,qz,qy,6)
      END DO
    END DO
  END DO
  
  DEALLOCATE(Sij)
  DEALLOCATE(Snorm)
  DEALLOCATE(tauij)  
  
END SUBROUTINE sgs_smagorinsky



!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!                                                          !!
!!                    sgs_winckelmans                       !!
!!                                                          !!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
SUBROUTINE sgs_winckelmans(a)
  USE parameters
  USE variables
  USE mpi_var
  IMPLICIT NONE

  COMPLEX,  DIMENSION(0:cx,0:cz,0:cy,3), INTENT(OUT)   ::  a
  COMPLEX,  DIMENSION(:,:,:,:),  ALLOCATABLE           ::  af, u_filt
  COMPLEX,  DIMENSION(:,:,:,:),  ALLOCATABLE           ::  tauij
  REAL,     DIMENSION(:,:,:,:),  ALLOCATABLE           ::  Sfij, Sij, pij, qij, Lij, Mij
  REAL,     DIMENSION(:,:,:,:,:),ALLOCATABLE           ::  diuj, diufj
  REAL,     DIMENSION(:,:,:),    ALLOCATABLE           ::  Snorm, Sfnorm, Sup, Inf, Sup_fpart, Inf_fpart
  REAL,     DIMENSION(:,:,:),    ALLOCATABLE           ::  C                
  REAL,     DIMENSION(:,:,:),    ALLOCATABLE           ::  r1
  COMPLEX,  DIMENSION(:,:,:),    ALLOCATABLE           ::  c1
  REAL      ::  alph1, Delta, AverageSup, AverageInf, Sup_node, Inf_node, nu_tot
  INTEGER   ::  qx, qy, qz, j  

  Delta = MIN(dx,dy,dz)
  !Delta=(dx*dy*dz)^(1.0/3.0)
  ALLOCATE(Sij(0:rx,0:ry,0:rz,6))
  ALLOCATE(u_filt(0:cx,0:cz,0:cy,3))
  DO j=1,3
    u_filt(:,:,:,j)=LES_filter(:,:,:)*vel(:,:,:,j)
  END DO
  CALL compute_strain(u_filt, Sij)  
  
  ALLOCATE(Snorm(0:rx,0:ry,0:rz))                    
  Snorm  = Sij(:,:,:,1)**2 + 2.0*Sij(:,:,:,2)**2 + 2.0 * Sij(:,:,:,3)**2 + Sij(:,:,:,4)**2 + 2.0*Sij(:,:,:,5)**2 + Sij(:,:,:,6)**2
  Snorm  = SQRT(2.0*Snorm)
  
  ALLOCATE(c1(0:cx,0:cz,0:cy))
  ALLOCATE(r1(0:rx,0:ry,0:rz))
  ALLOCATE(af(0:cx,0:cz,0:cy,3))
  DO j=1,3
    af(:,:,:,j)=u_filt(:,:,:,j)*Cut_filter(:,:,:)*Gauss_filter(:,:,:)
  END DO
  ALLOCATE(Sfij(0:rx,0:ry,0:rz,6))
  CALL compute_strain(af,Sfij)
  ALLOCATE(Sfnorm(0:rx,0:ry,0:rz))
  Sfnorm = Sfij(:,:,:,1)**2 + 2.0*Sfij(:,:,:,2)**2 + 2.0*Sfij(:,:,:,3)**2 + Sfij(:,:,:,4)**2 + 2.0*Sfij(:,:,:,5)**2 + Sfij(:,:,:,6)**2
  Sfnorm = SQRT(2.0*Sfnorm)

  SELECT CASE (filter_type)
  CASE(1)
    alph1=2.0
  CASE(2)
    alph1=4.0
  CASE(3)
    alph1=2.0
  END SELECT

  ALLOCATE(Lij(0:rx,0:ry,0:rz,6))
  CALL compute_leonard(u_filt,Lij)

  ALLOCATE(qij(0:rx,0:ry,0:rz,6))
  DO j=1,6
    r1=Snorm*Sij(:,:,:,j)
    CALL fft3ds(r1,c1)
    c1(:,:,:)=c1(:,:,:)*Gauss_filter(:,:,:)*Cut_filter(:,:,:)
    CALL ifft3ds(c1,r1)    
    qij(:,:,:,j)=LES_filter(:,:,:)*( 2*Delta*Delta*(r1-alph1*alph1*Sfnorm*Sfij(:,:,:,j)) )
  END DO
  DO j=1,6
    qij(:,:,:,j)=qij(:,:,:,j)-1.0/3.0*(qij(:,:,:,1)+qij(:,:,:,4)+qij(:,:,:,6))
  END DO   
  
  ALLOCATE(pij(0:rx,0:ry,0:rz,6))
  ALLOCATE(diuj(0:rx,0:ry,0:rz,3,3))
  ALLOCATE(diufj(0:rx,0:ry,0:rz,3,3))
  CALL do_diuj(u_filt,diuj)
  CALL do_diuj(af,diufj)
  DEALLOCATE(af)  
  pij(:,:,:,1) = diuj(:,:,:,1,1)*diuj(:,:,:,1,1) + diuj(:,:,:,2,1)*diuj(:,:,:,2,1) + diuj(:,:,:,3,1)*diuj(:,:,:,3,1)
  pij(:,:,:,2) = diuj(:,:,:,1,1)*diuj(:,:,:,1,2) + diuj(:,:,:,2,1)*diuj(:,:,:,2,2) + diuj(:,:,:,3,1)*diuj(:,:,:,3,2)
  pij(:,:,:,3) = diuj(:,:,:,1,1)*diuj(:,:,:,1,3) + diuj(:,:,:,2,1)*diuj(:,:,:,2,3) + diuj(:,:,:,3,1)*diuj(:,:,:,3,3)
  pij(:,:,:,4) = diuj(:,:,:,1,2)*diuj(:,:,:,1,2) + diuj(:,:,:,2,2)*diuj(:,:,:,2,2) + diuj(:,:,:,3,2)*diuj(:,:,:,3,2)
  pij(:,:,:,5) = diuj(:,:,:,1,2)*diuj(:,:,:,1,3) + diuj(:,:,:,2,2)*diuj(:,:,:,2,3) + diuj(:,:,:,3,2)*diuj(:,:,:,3,3)
  pij(:,:,:,6) = diuj(:,:,:,1,3)*diuj(:,:,:,1,3) + diuj(:,:,:,2,3)*diuj(:,:,:,2,3) + diuj(:,:,:,3,3)*diuj(:,:,:,3,3)
  DEALLOCATE(diuj)
  ALLOCATE(Mij(0:rx,0:ry,0:rz,6))  
  Mij=pij  
  DO j=1,6
    Mij(:,:,:,j)=Mij(:,:,:,j)-1.0/3.0*(Mij(:,:,:,1)+Mij(:,:,:,4)+Mij(:,:,:,6))
  END DO
  DO j=1,6
    r1=pij(:,:,:,j)
    CALL fft3ds(r1,c1)
    c1(:,:,:)=c1(:,:,:)*Gauss_filter(:,:,:)*Cut_filter(:,:,:)
    CALL ifft3ds(c1,r1)
    pij(:,:,:,j)=r1
    r1=Mij(:,:,:,j)
    CALL fft3ds(r1,c1)
    c1(:,:,:)=c1(:,:,:)*Gauss_filter(:,:,:)
    CALL ifft3ds(c1,r1)
    Mij(:,:,:,j)=r1
  END DO
  pij(:,:,:,1) = pij(:,:,:,1) - alph1*alph1 * ( diufj(:,:,:,1,1)*diufj(:,:,:,1,1) + diufj(:,:,:,2,1)*diufj(:,:,:,2,1) + diufj(:,:,:,3,1)*diufj(:,:,:,3,1) )
  pij(:,:,:,2) = pij(:,:,:,2) - alph1*alph1 * ( diufj(:,:,:,1,1)*diufj(:,:,:,1,2) + diufj(:,:,:,2,1)*diufj(:,:,:,2,2) + diufj(:,:,:,3,1)*diufj(:,:,:,3,2) )
  pij(:,:,:,3) = pij(:,:,:,3) - alph1*alph1 * ( diufj(:,:,:,1,1)*diufj(:,:,:,1,3) + diufj(:,:,:,2,1)*diufj(:,:,:,2,3) + diufj(:,:,:,3,1)*diufj(:,:,:,3,3) )
  pij(:,:,:,4) = pij(:,:,:,4) - alph1*alph1 * ( diufj(:,:,:,1,2)*diufj(:,:,:,1,2) + diufj(:,:,:,2,2)*diufj(:,:,:,2,2) + diufj(:,:,:,3,2)*diufj(:,:,:,3,2) )
  pij(:,:,:,5) = pij(:,:,:,5) - alph1*alph1 * ( diufj(:,:,:,1,2)*diufj(:,:,:,1,3) + diufj(:,:,:,2,2)*diufj(:,:,:,2,3) + diufj(:,:,:,3,2)*diufj(:,:,:,3,3) )
  pij(:,:,:,6) = pij(:,:,:,6) - alph1*alph1 * ( diufj(:,:,:,1,3)*diufj(:,:,:,1,3) + diufj(:,:,:,2,3)*diufj(:,:,:,2,3) + diufj(:,:,:,3,3)*diufj(:,:,:,3,3) )
  DEALLOCATE(diufj)
  DO j=1,6
    pij(:,:,:,j)=Delta*Delta*pij(:,:,:,j)+Lij(:,:,:,j)
  END DO
  DEALLOCATE(Lij)
  DO j=1,6
    pij(:,:,:,j)=pij(:,:,:,j)-1.0/3.0*(pij(:,:,:,1)+pij(:,:,:,4)+pij(:,:,:,6))
  END DO 
  DO j=1,6
    r1=pij(:,:,:,j)
    CALL fft3ds(r1,c1)
    c1(:,:,:)=c1(:,:,:)*Cut_filter(:,:,:)
    CALL ifft3ds(c1,r1)
    pij(:,:,:,j)=r1
  END DO
  
    
  ALLOCATE(Sup(0:rx,0:ry,0:rz))
  ALLOCATE(Inf(0:rx,0:ry,0:rz))    
  Sup = pij(:,:,:,1)*qij(:,:,:,1) + 2.0*pij(:,:,:,2)*qij(:,:,:,2) + 2.0*pij(:,:,:,3)*qij(:,:,:,3) + pij(:,:,:,4)*qij(:,:,:,4) + 2.0*pij(:,:,:,5)*qij(:,:,:,5) + pij(:,:,:,6)*qij(:,:,:,6)
  Inf = qij(:,:,:,1)**2 + 2.0*qij(:,:,:,2)**2 + 2.0*qij(:,:,:,3)**2 + qij(:,:,:,4)**2 + 2.0*qij(:,:,:,5)**2 + qij(:,:,:,6)**2      
  DEALLOCATE(pij)
  DEALLOCATE(qij)    

  Sup_node=SUM(Sup)
  CALL mpi_allreduce(Sup_node,AverageSup,1,MPI_MYREAL,MPI_SUM,MPI_COMM_INSTANCE,ierr)
  Inf_node=SUM(Inf)
  CALL mpi_allreduce(Inf_node,AverageInf,1,MPI_MYREAL,MPI_SUM,MPI_COMM_INSTANCE,ierr) 
  ALLOCATE(C(1,1,1))  
  C(1,1,1)= AverageSup/AverageInf 
    
  DEALLOCATE(Sup)
  DEALLOCATE(Inf)    
    
  ALLOCATE(tauij(0:cx,0:cz,0:cy,6))
  DO j =1, 6
    r1= Delta*Delta * ( Mij(:,:,:,j) - 2.0*C(1,1,1)*Snorm*Sij(:,:,:,j) )
    CALL fft3ds(r1,c1) 
    tauij(:,:,:,j)=c1
  END DO
  DO j=1,6
    tauij(:,:,:,j)=tauij(:,:,:,j)-1.0/3.0*(tauij(:,:,:,1)+tauij(:,:,:,4)+tauij(:,:,:,6))
  END DO 
      
  DEALLOCATE(C)
  DEALLOCATE(r1)
  DEALLOCATE(c1)
  DEALLOCATE(Mij)

  DO qy=0, cy
    DO qz=0, cz
      DO qx=0, cx
        a(qx,qz,qy,1) = - ikx(qx)*tauij(qx,qz,qy,1) - iky(qy)*tauij(qx,qz,qy,2) - ikz(qz)*tauij(qx,qz,qy,3)
        a(qx,qz,qy,2) = - ikx(qx)*tauij(qx,qz,qy,2) - iky(qy)*tauij(qx,qz,qy,4) - ikz(qz)*tauij(qx,qz,qy,5)
        a(qx,qz,qy,3) = - ikx(qx)*tauij(qx,qz,qy,3) - iky(qy)*tauij(qx,qz,qy,5) - ikz(qz)*tauij(qx,qz,qy,6)
      END DO
    END DO
  END DO
    
  DEALLOCATE(Sij)
  DEALLOCATE(Snorm)
  DEALLOCATE(Sfnorm)
  DEALLOCATE(Sfij)  
  DEALLOCATE(tauij)  
  
END SUBROUTINE sgs_winckelmans



!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!                                                          !!
!!                      sgs_bardina                         !!
!!                                                          !!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
SUBROUTINE sgs_bardina(a)
  USE parameters
  USE variables
  USE mpi_var
  IMPLICIT NONE

  COMPLEX,  DIMENSION(0:cx,0:cz,0:cy,3), INTENT(OUT)   ::  a
  COMPLEX,  DIMENSION(:,:,:,:),  ALLOCATABLE           ::  u_filt
  COMPLEX,  DIMENSION(:,:,:,:),  ALLOCATABLE           ::  tauij
  REAL,     DIMENSION(:,:,:,:),  ALLOCATABLE           ::  Mij
  REAL,     DIMENSION(:,:,:,:,:),ALLOCATABLE           ::  diuj 
  REAL,     DIMENSION(:,:,:),    ALLOCATABLE           ::  r1
  COMPLEX,  DIMENSION(:,:,:),    ALLOCATABLE           ::  c1
  REAL      ::  alph1, Delta, AverageSup, AverageInf, Sup_node, Inf_node, nu_tot
  INTEGER   ::  qx, qy, qz, j  

  Delta = MIN(dx,dy,dz)
  ALLOCATE(u_filt(0:cx,0:cz,0:cy,3))
  DO j=1,3
    u_filt(:,:,:,j)=LES_filter(:,:,:)*vel(:,:,:,j)
  END DO
  
  ALLOCATE(c1(0:cx,0:cz,0:cy))
  ALLOCATE(r1(0:rx,0:ry,0:rz))

  SELECT CASE (filter_type)
  CASE(1)
    alph1=2.0
  CASE(2)
    alph1=4.0
  CASE(3)
    alph1=2.0
  END SELECT
  
  ALLOCATE(Mij(0:rx,0:ry,0:rz,6))
  ALLOCATE(diuj(0:rx,0:ry,0:rz,3,3))
  CALL do_diuj(u_filt,diuj)
  Mij(:,:,:,1) = diuj(:,:,:,1,1)*diuj(:,:,:,1,1) + diuj(:,:,:,2,1)*diuj(:,:,:,2,1) + diuj(:,:,:,3,1)*diuj(:,:,:,3,1)
  Mij(:,:,:,2) = diuj(:,:,:,1,1)*diuj(:,:,:,1,2) + diuj(:,:,:,2,1)*diuj(:,:,:,2,2) + diuj(:,:,:,3,1)*diuj(:,:,:,3,2)
  Mij(:,:,:,3) = diuj(:,:,:,1,1)*diuj(:,:,:,1,3) + diuj(:,:,:,2,1)*diuj(:,:,:,2,3) + diuj(:,:,:,3,1)*diuj(:,:,:,3,3)
  Mij(:,:,:,4) = diuj(:,:,:,1,2)*diuj(:,:,:,1,2) + diuj(:,:,:,2,2)*diuj(:,:,:,2,2) + diuj(:,:,:,3,2)*diuj(:,:,:,3,2)
  Mij(:,:,:,5) = diuj(:,:,:,1,2)*diuj(:,:,:,1,3) + diuj(:,:,:,2,2)*diuj(:,:,:,2,3) + diuj(:,:,:,3,2)*diuj(:,:,:,3,3)
  Mij(:,:,:,6) = diuj(:,:,:,1,3)*diuj(:,:,:,1,3) + diuj(:,:,:,2,3)*diuj(:,:,:,2,3) + diuj(:,:,:,3,3)*diuj(:,:,:,3,3)
  DEALLOCATE(diuj)
  DO j=1,6
    Mij(:,:,:,j)=Mij(:,:,:,j)-1.0/3.0*(Mij(:,:,:,1)+Mij(:,:,:,4)+Mij(:,:,:,6))
  END DO
  DO j=1,6
    r1=Mij(:,:,:,j)
    CALL fft3ds(r1,c1)
    c1(:,:,:)=c1(:,:,:)*Gauss_filter(:,:,:)
    CALL ifft3ds(c1,r1)
    Mij(:,:,:,j)=r1
  END DO
    
  ALLOCATE(tauij(0:cx,0:cz,0:cy,6))
  DO j =1, 6
    r1= Delta*Delta * Mij(:,:,:,j)
    CALL fft3ds(r1,c1) 
    tauij(:,:,:,j)=c1
  END DO
  DO j=1,6
    tauij(:,:,:,j)=tauij(:,:,:,j)-1.0/3.0*(tauij(:,:,:,1)+tauij(:,:,:,4)+tauij(:,:,:,6))
  END DO 
      
  DEALLOCATE(r1)
  DEALLOCATE(c1)
  DEALLOCATE(Mij)

  DO qy=0, cy
    DO qz=0, cz
      DO qx=0, cx
        a(qx,qz,qy,1) = - ikx(qx)*tauij(qx,qz,qy,1) - iky(qy)*tauij(qx,qz,qy,2) - ikz(qz)*tauij(qx,qz,qy,3)
        a(qx,qz,qy,2) = - ikx(qx)*tauij(qx,qz,qy,2) - iky(qy)*tauij(qx,qz,qy,4) - ikz(qz)*tauij(qx,qz,qy,5)
        a(qx,qz,qy,3) = - ikx(qx)*tauij(qx,qz,qy,3) - iky(qy)*tauij(qx,qz,qy,5) - ikz(qz)*tauij(qx,qz,qy,6)
      END DO
    END DO
  END DO
    
  DEALLOCATE(tauij)  
  
END SUBROUTINE sgs_bardina



!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!                                                          !!
!!                     compute_strain                       !!
!!                                                          !!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
SUBROUTINE compute_strain(a, Sij)
  USE parameters
  IMPLICIT NONE
  
  COMPLEX,  DIMENSION(0:cx,0:cz,0:cy,3),  INTENT(IN)   ::  a
  REAL,     DIMENSION(0:rx,0:ry,0:rz,6),  INTENT(OUT)  ::  Sij
  
  COMPLEX,  ALLOCATABLE,  DIMENSION(:,:,:)  :: c1
  REAL,     ALLOCATABLE,  DIMENSION(:,:,:)  :: r1
  INTEGER   ::  qx, qy, qz

  ALLOCATE(c1(0:cx,0:cz,0:cy),r1(0:rx,0:ry,0:rz))
  
  DO qx=0, cx
    c1(qx,:,:) =  ikx(qx)*a(qx,:,:,1)
  END DO
  CALL ifft3ds(c1,r1)
  Sij(:,:,:,1)=r1

  DO qx=0, cx
    c1(qx,:,:) = ikx(qx)*a(qx,:,:,2)
  END DO
  DO qy=0, cy
    c1(:,:,qy) = (c1(:,:,qy) + iky(qy)*a(:,:,qy,1))*0.5
  END DO
  CALL ifft3ds(c1,r1)
  Sij(:,:,:,2)=r1

  DO qx=0, cx
    c1(qx,:,:) = ikx(qx)*a(qx,:,:,3)
  END DO
  DO qz=0, cz
    c1(:,qz,:) = (c1(:,qz,:) + ikz(qz)*a(:,qz,:,1))*0.5
  END DO
  CALL ifft3ds(c1,r1)
  Sij(:,:,:,3)=r1

  DO qy=0, cy
    c1(:,:,qy) =  iky(qy)*a(:,:,qy,2)
  END DO
  CALL ifft3ds(c1,r1)
  Sij(:,:,:,4)=r1

  DO qy=0, cy
    c1(:,:,qy) = iky(qy)*a(:,:,qy,3)
  END DO
  DO qz=0, cz
    c1(:,qz,:) = (c1(:,qz,:) + ikz(qz)*a(:,qz,:,2))*0.5
  END DO
  CALL ifft3ds(c1,r1)
  Sij(:,:,:,5)=r1

  DO qz=0, cz
    c1(:,qz,:) = ikz(qz)*a(:,qz,:,3)
  END DO
  CALL ifft3ds(c1,r1)
  Sij(:,:,:,6)=r1
  
  DEALLOCATE(c1,r1)

END SUBROUTINE compute_strain



!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!                                                          !!
!!                    compute_rotation                      !!
!!                                                          !!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
SUBROUTINE compute_rotation(a, Rij)
  USE parameters
  IMPLICIT NONE
  
  COMPLEX,  DIMENSION(0:cx,0:cz,0:cy,3),  INTENT(IN)   ::  a
  REAL,     DIMENSION(0:rx,0:ry,0:rz,3),  INTENT(OUT)  ::  Rij
  
  COMPLEX,  ALLOCATABLE,  DIMENSION(:,:,:)  :: c1
  REAL,     ALLOCATABLE,  DIMENSION(:,:,:)  :: r1
  INTEGER   ::  qx, qy, qz

  ALLOCATE(c1(0:cx,0:cz,0:cy),r1(0:rx,0:ry,0:rz))
  
  DO qx=0, cx
    c1(qx,:,:) = ikx(qx)*a(qx,:,:,2)
  END DO
  DO qy=0, cy
    c1(:,:,qy) = (c1(:,:,qy) - iky(qy)*a(:,:,qy,1))*0.5
  END DO
  CALL ifft3ds(c1,r1)
  Rij(:,:,:,1)=r1

  DO qx=0, cx
    c1(qx,:,:) = ikx(qx)*a(qx,:,:,3)
  END DO
  DO qz=0, cz
    c1(:,qz,:) = (c1(:,qz,:) - ikz(qz)*a(:,qz,:,1))*0.5
  END DO
  CALL ifft3ds(c1,r1)
  Rij(:,:,:,2)=r1

  DO qy=0, cy
    c1(:,:,qy) = iky(qy)*a(:,:,qy,3)
  END DO
  DO qz=0, cz
    c1(:,qz,:) = (c1(:,qz,:) - ikz(qz)*a(:,qz,:,2))*0.5
  END DO
  CALL ifft3ds(c1,r1)
  Rij(:,:,:,3)=r1
  
  DEALLOCATE(c1,r1)

END SUBROUTINE compute_rotation



!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!                                                          !!
!!                         filter                           !!
!!                                                          !!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
SUBROUTINE filter(a, af)
  USE parameters
  IMPLICIT NONE
  
  COMPLEX,  DIMENSION(0:cx,0:cz,0:cy),  INTENT(IN)   ::  a
  COMPLEX,  DIMENSION(0:cx,0:cz,0:cy),  INTENT(OUT)  ::  af
 
  SELECT CASE (filter_type)
  CASE(1,2,3)            
    af=test_filter*a
  END SELECT
    
END SUBROUTINE filter



!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!                                                          !!
!!                    compute_leonard                       !!
!!                                                          !!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
SUBROUTINE compute_leonard(a, Lij)
  USE parameters
  IMPLICIT NONE
  
  COMPLEX,  DIMENSION(0:cx,0:cz,0:cy,3),  INTENT(IN)   ::  a
  REAL,     DIMENSION(0:rx,0:ry,0:rz,6),  INTENT(OUT)  ::  Lij
  COMPLEX,  DIMENSION(:,:,:),  ALLOCATABLE  ::  c1
  REAL,     DIMENSION(:,:,:),  ALLOCATABLE  ::  a1, a2, a3
  INTEGER  ::  j

  ALLOCATE(c1(0:cx,0:cz,0:cy))
  ALLOCATE(a1(0:rx,0:ry,0:rz),a2(0:rx,0:ry,0:rz),a3(0:rx,0:ry,0:rz))
  c1=a(:,:,:,1)
  CALL ifft3ds(c1,a1)
  c1=a(:,:,:,2)
  CALL ifft3ds(c1,a2)
  c1=a(:,:,:,3)
  CALL ifft3ds(c1,a3)

  Lij(:,:,:,1)=a1*a1
  Lij(:,:,:,2)=a1*a2
  Lij(:,:,:,3)=a1*a3
  Lij(:,:,:,4)=a2*a2
  Lij(:,:,:,5)=a2*a3
  Lij(:,:,:,6)=a3*a3
  
  c1=a(:,:,:,1)
  CALL filter(c1,c1)
  CALL ifft3ds(c1,a1)
  c1=a(:,:,:,2)
  CALL filter(c1,c1)
  CALL ifft3ds(c1,a2)
  c1=a(:,:,:,3)
  CALL filter(c1,c1)
  CALL ifft3ds(c1,a3)

  Lij(:,:,:,1)=Lij(:,:,:,1)-a1*a1
  Lij(:,:,:,2)=Lij(:,:,:,2)-a1*a2
  Lij(:,:,:,3)=Lij(:,:,:,3)-a1*a3
  Lij(:,:,:,4)=Lij(:,:,:,4)-a2*a2
  Lij(:,:,:,5)=Lij(:,:,:,5)-a2*a3
  Lij(:,:,:,6)=Lij(:,:,:,6)-a3*a3

  DEALLOCATE(a2,a3)
  
  DO j=1,6
    a1=Lij(:,:,:,j)
    CALL fft3ds(a1,c1)
    CALL filter(c1,c1)
    CALL ifft3ds(c1,a1)
    Lij(:,:,:,j)=a1
  END DO  
  
  DEALLOCATE(a1,c1)
  
END SUBROUTINE compute_leonard



!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!                                                          !!
!!                        do_diuj                           !!
!!                                                          !!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
SUBROUTINE do_diuj(a, du)
  USE parameters
  IMPLICIT NONE
  
  COMPLEX,  DIMENSION(0:cx,0:cz,0:cy,3),  INTENT(IN)   ::  a
  REAL,     DIMENSION(0:rx,0:ry,0:rz,3,3),INTENT(OUT)  ::  du
  COMPLEX,  DIMENSION(:,:,:),  ALLOCATABLE  ::  c1
  REAL,     DIMENSION(:,:,:),  ALLOCATABLE  ::  r1
  INTEGER  ::  qx, qy, qz
  
  ALLOCATE(c1(0:cx,0:cz,0:cy))
  ALLOCATE(r1(0:rx,0:ry,0:rz))
  
  DO qx=1,cx
    c1(qx,:,:)=ikx(qx)*a(qx,:,:,1)
  END DO
  CALL ifft3ds(c1,r1)
  du(:,:,:,1,1)=r1
  DO qx=1,cx
    c1(qx,:,:)=ikx(qx)*a(qx,:,:,2)
  END DO
  CALL ifft3ds(c1,r1)
  du(:,:,:,1,2)=r1
  DO qx=1,cx
    c1(qx,:,:)=ikx(qx)*a(qx,:,:,3)
  END DO
  CALL ifft3ds(c1,r1)
  du(:,:,:,1,3)=r1
  DO qy=1,cy
    c1(:,:,qy)=iky(qy)*a(:,:,qy,1)
  END DO
  CALL ifft3ds(c1,r1)
  du(:,:,:,2,1)=r1
  DO qy=1,cy
    c1(:,:,qy)=iky(qy)*a(:,:,qy,2)
  END DO
  CALL ifft3ds(c1,r1)
  du(:,:,:,2,2)=r1
  DO qy=1,cy
    c1(:,:,qy)=iky(qy)*a(:,:,qy,3)
  END DO
  CALL ifft3ds(c1,r1)
  du(:,:,:,2,3)=r1
  DO qy=1,cz
    c1(:,qz,:)=ikz(qz)*a(:,qz,:,1)
  END DO
  CALL ifft3ds(c1,r1)
  du(:,:,:,3,1)=r1
  DO qy=1,cz
    c1(:,qz,:)=ikz(qz)*a(:,qz,:,2)
  END DO
  CALL ifft3ds(c1,r1)
  du(:,:,:,3,2)=r1
  DO qy=1,cz
    c1(:,qz,:)=ikz(qz)*a(:,qz,:,3)
  END DO
  CALL ifft3ds(c1,r1)
  du(:,:,:,3,3)=r1

  DEALLOCATE(c1)
  DEALLOCATE(r1)
  
END SUBROUTINE do_diuj



!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!                                                          !!
!!                 sgs_exchanging_tauij                     !!
!!                                                          !!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
SUBROUTINE sgs_exchanging_tauij(a)
  USE parameters
  USE variables
  IMPLICIT NONE
  
  COMPLEX,  DIMENSION(0:cx,0:cz,0:cy,3), INTENT(OUT)  ::  a
  COMPLEX,  DIMENSION(:,:,:,:),          ALLOCATABLE  ::  tauij, subtotal
  COMPLEX,  DIMENSION(:,:,:),            ALLOCATABLE  ::  c1, filtering  
  REAL,     DIMENSION(:,:,:),            ALLOCATABLE  ::  ur1, ur2, ur3
  REAL,     DIMENSION(:,:,:),            ALLOCATABLE  ::  urf1, urf2, urf3
  REAL,     DIMENSION(:,:,:,:),          ALLOCATABLE  ::  Sij, Rij, Pij, tauijr
  INTEGER        ::  qx, qy, qz, j
  
  SELECT CASE(LES_ident)
  CASE(0)  
    ALLOCATE(tauij(0:cx,0:cz,0:cy,6))
    ALLOCATE(c1(0:cx,0:cz,0:cy))
    ALLOCATE(filtering(0:cx,0:cz,0:cy))
    
    filtering=1.0    
    DO qy=0, cy
      IF (ABS(ky(qy)) > kMyl) filtering(:,:,qy)=0.0
      DO qz=0, cz
        IF (ABS(kz(qz)) > kMzl) filtering(:,qz,qy)=0.0
        DO qx=0, cx
          IF (ABS(kx(qx)) > kMxl) filtering(qx,qz,qy)=0.0
        END DO
      END DO
    END DO    

    ALLOCATE(ur1(0:rx,0:ry,0:rz),ur2(0:rx,0:ry,0:rz),ur3(0:rx,0:ry,0:rz))
    c1=vel(:,:,:,1)
    CALL ifft3ds(c1,ur1)
    c1=vel(:,:,:,2)
    CALL ifft3ds(c1,ur2)
    c1=vel(:,:,:,3)
    CALL ifft3ds(c1,ur3)

    ALLOCATE(urf1(0:rx,0:ry,0:rz),urf2(0:rx,0:ry,0:rz),urf3(0:rx,0:ry,0:rz))
    c1=vel(:,:,:,1)*filtering
    CALL ifft3ds(c1,urf1)
    c1=vel(:,:,:,2)*filtering
    CALL ifft3ds(c1,urf2)
    c1=vel(:,:,:,3)*filtering
    CALL ifft3ds(c1,urf3) 
    
    ! Tau(1,1)  
    CALL fft3ds(ur1*ur1,tauij(:,:,:,1))
    CALL fft3ds(urf1*urf1,c1)
    tauij(:,:,:,1)=filtering*(tauij(:,:,:,1)-c1)
  
    ! Tau(1,2)  
    CALL fft3ds(ur1*ur2,tauij(:,:,:,2))
    CALL fft3ds(urf1*urf2,c1)
    tauij(:,:,:,2)=filtering*(tauij(:,:,:,2)-c1)
  
    ! Tau(1,3)  
    CALL fft3ds(ur1*ur3,tauij(:,:,:,3))
    CALL fft3ds(urf1*urf3,c1)
    tauij(:,:,:,3)=filtering*(tauij(:,:,:,3)-c1)
  
    ! Tau(2,2)  
    CALL fft3ds(ur2*ur2,tauij(:,:,:,4))
    CALL fft3ds(urf2*urf2,c1)
    tauij(:,:,:,4)=filtering*(tauij(:,:,:,4)-c1)
  
    ! Tau(2,3)  
    CALL fft3ds(ur2*ur3,tauij(:,:,:,5))
    CALL fft3ds(urf2*urf3,c1)
    tauij(:,:,:,5)=filtering*(tauij(:,:,:,5)-c1)
    
    ! Tau(3,3)  
    CALL fft3ds(ur3*ur3,tauij(:,:,:,6))
    CALL fft3ds(urf3*urf3,c1)
    tauij(:,:,:,6)=filtering*(tauij(:,:,:,6)-c1)
      
    DO j=1,6
      CALL reduce_send_and_receive(tauij(:,:,:,j))
    END DO
    
    DEALLOCATE(tauij)
    DEALLOCATE(ur1,ur2,ur3)
    DEALLOCATE(urf1,urf2,urf3)
    DEALLOCATE(c1)
    DEALLOCATE(filtering)
    a=0.0 
  CASE(1)
    ALLOCATE(tauij(0:cx,0:cz,0:cy,6))

    DO j=1,6
      CALL reduce_send_and_receive(tauij(:,:,:,j))
    END DO
    
    IF (add_total.EQ.0) THEN
      ALLOCATE(subtotal(0:cx,0:cz,0:cy,6))
      subtotal=0.0
      ALLOCATE(Sij(0:rx,0:ry,0:rz,6))
      ALLOCATE(Rij(0:rx,0:ry,0:rz,3))
      ALLOCATE(Pij(0:rx,0:ry,0:rz,6))
      ALLOCATE(tauijr(0:rx,0:ry,0:rz,6))
      ALLOCATE(c1(0:cx,0:cz,0:cy))      
      ALLOCATE(ur1(0:rx,0:ry,0:rz))      
      CALL compute_strain(vel, Sij) 
      CALL compute_rotation(vel,Rij)
      DO j=1,6
        c1=tauij(:,:,:,j)
        CALL ifft3ds(c1,ur1)
        tauijr(:,:,:,j)=ur1
      END DO
      
      IF (add_1.EQ.1) THEN
        Pij=Sij
        ur1=0.0
        ur1 = (tauijr(:,:,:,1)*Pij(:,:,:,1) + 2*tauijr(:,:,:,2)*Pij(:,:,:,2) + 2*tauijr(:,:,:,3)*Pij(:,:,:,3) + tauijr(:,:,:,4)*Pij(:,:,:,4) + 2*tauijr(:,:,:,5)*Pij(:,:,:,5) + tauijr(:,:,:,6)*Pij(:,:,:,6)) / (Pij(:,:,:,1)*Pij(:,:,:,1) + 2*Pij(:,:,:,2)*Pij(:,:,:,2) + 2*Pij(:,:,:,3)*Pij(:,:,:,3) + Pij(:,:,:,4)*Pij(:,:,:,4) + 2*Pij(:,:,:,5)*Pij(:,:,:,5) + Pij(:,:,:,6)*Pij(:,:,:,6))
        DO j=1,6
          tauijr(:,:,:,j)=ur1*Pij(:,:,:,j)
        END DO
        DO j=1,6
          ur1=tauijr(:,:,:,j)
          CALL fft3ds(ur1,c1)
          subtotal(:,:,:,j)=subtotal(:,:,:,j)+c1
        END DO
      END IF
      IF (add_2.EQ.1) THEN
        Pij(:,:,:,1)= Sij(:,:,:,1)*Sij(:,:,:,1) + Sij(:,:,:,2)*Sij(:,:,:,2) + Sij(:,:,:,3)*Sij(:,:,:,3)
        Pij(:,:,:,2)= Sij(:,:,:,1)*Sij(:,:,:,2) + Sij(:,:,:,2)*Sij(:,:,:,4) + Sij(:,:,:,3)*Sij(:,:,:,5)
        Pij(:,:,:,3)= Sij(:,:,:,1)*Sij(:,:,:,3) + Sij(:,:,:,2)*Sij(:,:,:,5) + Sij(:,:,:,3)*Sij(:,:,:,6)
        Pij(:,:,:,4)= Sij(:,:,:,2)*Sij(:,:,:,2) + Sij(:,:,:,4)*Sij(:,:,:,4) + Sij(:,:,:,5)*Sij(:,:,:,5)
        Pij(:,:,:,5)= Sij(:,:,:,2)*Sij(:,:,:,3) + Sij(:,:,:,4)*Sij(:,:,:,5) + Sij(:,:,:,5)*Sij(:,:,:,6)
        Pij(:,:,:,6)= Sij(:,:,:,3)*Sij(:,:,:,3) + Sij(:,:,:,5)*Sij(:,:,:,5) + Sij(:,:,:,6)*Sij(:,:,:,6)
        ur1=(Pij(:,:,:,1)+Pij(:,:,:,4)+Pij(:,:,:,6))/3.0
        Pij(:,:,:,1)=Pij(:,:,:,1)-ur1
        Pij(:,:,:,4)=Pij(:,:,:,4)-ur1
        Pij(:,:,:,6)=Pij(:,:,:,6)-ur1
        ur1=0.0
        ur1 = (tauijr(:,:,:,1)*Pij(:,:,:,1) + 2*tauijr(:,:,:,2)*Pij(:,:,:,2) + 2*tauijr(:,:,:,3)*Pij(:,:,:,3) + tauijr(:,:,:,4)*Pij(:,:,:,4) + 2*tauijr(:,:,:,5)*Pij(:,:,:,5) + tauijr(:,:,:,6)*Pij(:,:,:,6)) / (Pij(:,:,:,1)*Pij(:,:,:,1) + 2*Pij(:,:,:,2)*Pij(:,:,:,2) + 2*Pij(:,:,:,3)*Pij(:,:,:,3) + Pij(:,:,:,4)*Pij(:,:,:,4) + 2*Pij(:,:,:,5)*Pij(:,:,:,5) + Pij(:,:,:,6)*Pij(:,:,:,6))
        DO j=1,6
          tauijr(:,:,:,j)=ur1*Pij(:,:,:,j)
        END DO
        DO j=1,6
          ur1=tauijr(:,:,:,j)
          CALL fft3ds(ur1,c1)
          subtotal(:,:,:,j)=subtotal(:,:,:,j)+c1
        END DO
      END IF
      IF (add_3.EQ.1) THEN
        Pij(:,:,:,1)= - Rij(:,:,:,1)*Rij(:,:,:,1) - Rij(:,:,:,2)*Rij(:,:,:,2)
        Pij(:,:,:,2)= - Rij(:,:,:,2)*Rij(:,:,:,3)
        Pij(:,:,:,3)=   Rij(:,:,:,1)*Rij(:,:,:,3)
        Pij(:,:,:,4)= - Rij(:,:,:,1)*Rij(:,:,:,1) - Rij(:,:,:,3)*Rij(:,:,:,3)
        Pij(:,:,:,5)= - Rij(:,:,:,1)*Rij(:,:,:,2)
        Pij(:,:,:,6)= - Rij(:,:,:,2)*Rij(:,:,:,2) - Rij(:,:,:,3)*Rij(:,:,:,3)
        ur1=(Pij(:,:,:,1)+Pij(:,:,:,4)+Pij(:,:,:,6))/3.0
        Pij(:,:,:,1)=Pij(:,:,:,1)-ur1
        Pij(:,:,:,4)=Pij(:,:,:,4)-ur1
        Pij(:,:,:,6)=Pij(:,:,:,6)-ur1
        ur1=0.0
        ur1 = (tauijr(:,:,:,1)*Pij(:,:,:,1) + 2*tauijr(:,:,:,2)*Pij(:,:,:,2) + 2*tauijr(:,:,:,3)*Pij(:,:,:,3) + tauijr(:,:,:,4)*Pij(:,:,:,4) + 2*tauijr(:,:,:,5)*Pij(:,:,:,5) + tauijr(:,:,:,6)*Pij(:,:,:,6)) / (Pij(:,:,:,1)*Pij(:,:,:,1) + 2*Pij(:,:,:,2)*Pij(:,:,:,2) + 2*Pij(:,:,:,3)*Pij(:,:,:,3) + Pij(:,:,:,4)*Pij(:,:,:,4) + 2*Pij(:,:,:,5)*Pij(:,:,:,5) + Pij(:,:,:,6)*Pij(:,:,:,6))
        DO j=1,6
          tauijr(:,:,:,j)=ur1*Pij(:,:,:,j)
        END DO
        DO j=1,6
          ur1=tauijr(:,:,:,j)
          CALL fft3ds(ur1,c1)
          subtotal(:,:,:,j)=subtotal(:,:,:,j)+c1
        END DO
      END IF
      IF (add_4.EQ.1) THEN
        Pij(:,:,:,1)= -2.0 * ( Sij(:,:,:,2)*Rij(:,:,:,1) + Sij(:,:,:,3)*Rij(:,:,:,2) )
        Pij(:,:,:,2)= Sij(:,:,:,1)*Rij(:,:,:,1) - Sij(:,:,:,4)*Rij(:,:,:,1) - Sij(:,:,:,3)*Rij(:,:,:,3) - Sij(:,:,:,5)*Rij(:,:,:,2)
        Pij(:,:,:,3)= Sij(:,:,:,1)*Rij(:,:,:,1) + Sij(:,:,:,2)*Rij(:,:,:,3) - Sij(:,:,:,5)*Rij(:,:,:,1) - Sij(:,:,:,6)*Rij(:,:,:,2)
        Pij(:,:,:,4)=  2.0 * ( Sij(:,:,:,2)*Rij(:,:,:,1) - Sij(:,:,:,5)*Rij(:,:,:,3) )
        Pij(:,:,:,5)= Sij(:,:,:,2)*Rij(:,:,:,2) + Sij(:,:,:,3)*Rij(:,:,:,1) + Sij(:,:,:,4)*Rij(:,:,:,3) - Sij(:,:,:,6)*Rij(:,:,:,3)
        Pij(:,:,:,6)=  2.0 * ( Sij(:,:,:,3)*Rij(:,:,:,2) + Sij(:,:,:,5)*Rij(:,:,:,3) )
        ur1=0.0
        ur1 = (tauijr(:,:,:,1)*Pij(:,:,:,1) + 2*tauijr(:,:,:,2)*Pij(:,:,:,2) + 2*tauijr(:,:,:,3)*Pij(:,:,:,3) + tauijr(:,:,:,4)*Pij(:,:,:,4) + 2*tauijr(:,:,:,5)*Pij(:,:,:,5) + tauijr(:,:,:,6)*Pij(:,:,:,6)) / (Pij(:,:,:,1)*Pij(:,:,:,1) + 2*Pij(:,:,:,2)*Pij(:,:,:,2) + 2*Pij(:,:,:,3)*Pij(:,:,:,3) + Pij(:,:,:,4)*Pij(:,:,:,4) + 2*Pij(:,:,:,5)*Pij(:,:,:,5) + Pij(:,:,:,6)*Pij(:,:,:,6))
        DO j=1,6
          tauijr(:,:,:,j)=ur1*Pij(:,:,:,j)
        END DO
        DO j=1,6
          ur1=tauijr(:,:,:,j)
          CALL fft3ds(ur1,c1)
          subtotal(:,:,:,j)=subtotal(:,:,:,j)+c1
        END DO
      END IF
      IF (add_5.EQ.1) THEN
        Pij(:,:,:,1)= -2.0 * ( Sij(:,:,:,1)*Sij(:,:,:,2)*Rij(:,:,:,1) + Sij(:,:,:,1)*Sij(:,:,:,3)*Rij(:,:,:,2) + Sij(:,:,:,2)*Sij(:,:,:,4)*Rij(:,:,:,1) + Sij(:,:,:,2)*Sij(:,:,:,5)*Rij(:,:,:,2) + Sij(:,:,:,3)*Sij(:,:,:,5)*Rij(:,:,:,1) + Sij(:,:,:,3)*Sij(:,:,:,6)*Rij(:,:,:,2) )
        Pij(:,:,:,2)= Sij(:,:,:,1)*Sij(:,:,:,1)*Rij(:,:,:,1) - Sij(:,:,:,1)*Sij(:,:,:,3)*Rij(:,:,:,3) - Sij(:,:,:,2)*Sij(:,:,:,3)*Rij(:,:,:,2) - Sij(:,:,:,2)*Sij(:,:,:,5)*Rij(:,:,:,3) + Sij(:,:,:,3)*Sij(:,:,:,3)*Rij(:,:,:,1) - Sij(:,:,:,3)*Sij(:,:,:,6)*Rij(:,:,:,3) - Sij(:,:,:,4)*Sij(:,:,:,4)*Rij(:,:,:,1) - Sij(:,:,:,4)*Sij(:,:,:,5)*Rij(:,:,:,2) - Sij(:,:,:,5)*Sij(:,:,:,5)*Rij(:,:,:,1) - Sij(:,:,:,5)*Sij(:,:,:,6)*Rij(:,:,:,2)
        Pij(:,:,:,3)= Sij(:,:,:,1)*Sij(:,:,:,1)*Rij(:,:,:,2) + Sij(:,:,:,1)*Sij(:,:,:,2)*Rij(:,:,:,3) + Sij(:,:,:,2)*Sij(:,:,:,2)*Rij(:,:,:,2) - Sij(:,:,:,2)*Sij(:,:,:,3)*Rij(:,:,:,1) + Sij(:,:,:,2)*Sij(:,:,:,4)*Rij(:,:,:,3) + Sij(:,:,:,3)*Sij(:,:,:,5)*Rij(:,:,:,3) - Sij(:,:,:,4)*Sij(:,:,:,5)*Rij(:,:,:,1) - Sij(:,:,:,5)*Sij(:,:,:,5)*Rij(:,:,:,2) - Sij(:,:,:,5)*Sij(:,:,:,6)*Rij(:,:,:,1) - Sij(:,:,:,6)*Sij(:,:,:,6)*Rij(:,:,:,2)
        Pij(:,:,:,4)= 2.0 * ( Sij(:,:,:,1)*Sij(:,:,:,2)*Rij(:,:,:,1) + Sij(:,:,:,2)*Sij(:,:,:,4)*Rij(:,:,:,1) + Sij(:,:,:,3)*Sij(:,:,:,5)*Rij(:,:,:,1) - Sij(:,:,:,2)*Sij(:,:,:,3)*Rij(:,:,:,3) - Sij(:,:,:,4)*Sij(:,:,:,5)*Rij(:,:,:,3) - Sij(:,:,:,5)*Sij(:,:,:,6)*Rij(:,:,:,3) )
        Pij(:,:,:,5)= Sij(:,:,:,1)*Sij(:,:,:,2)*Rij(:,:,:,2) + Sij(:,:,:,1)*Sij(:,:,:,3)*Rij(:,:,:,1) + Sij(:,:,:,2)*Sij(:,:,:,2)*Rij(:,:,:,3) + Sij(:,:,:,2)*Sij(:,:,:,4)*Rij(:,:,:,2) + Sij(:,:,:,2)*Sij(:,:,:,5)*Rij(:,:,:,1) - Sij(:,:,:,3)*Sij(:,:,:,3)*Rij(:,:,:,3) + Sij(:,:,:,3)*Sij(:,:,:,5)*Rij(:,:,:,2) + Sij(:,:,:,3)*Sij(:,:,:,6)*Rij(:,:,:,1) + Sij(:,:,:,4)*Sij(:,:,:,4)*Rij(:,:,:,3) - Sij(:,:,:,6)*Sij(:,:,:,6)*Rij(:,:,:,3)
        Pij(:,:,:,6)= 2.0 * ( Sij(:,:,:,1)*Sij(:,:,:,3)*Rij(:,:,:,2) + Sij(:,:,:,2)*Sij(:,:,:,3)*Rij(:,:,:,3) + Sij(:,:,:,2)*Sij(:,:,:,5)*Rij(:,:,:,2) + Sij(:,:,:,3)*Sij(:,:,:,6)*Rij(:,:,:,2) + Sij(:,:,:,4)*Sij(:,:,:,5)*Rij(:,:,:,3) + Sij(:,:,:,5)*Sij(:,:,:,6)*Rij(:,:,:,3) )
        ur1=0.0
        ur1 = (tauijr(:,:,:,1)*Pij(:,:,:,1) + 2*tauijr(:,:,:,2)*Pij(:,:,:,2) + 2*tauijr(:,:,:,3)*Pij(:,:,:,3) + tauijr(:,:,:,4)*Pij(:,:,:,4) + 2*tauijr(:,:,:,5)*Pij(:,:,:,5) + tauijr(:,:,:,6)*Pij(:,:,:,6)) / (Pij(:,:,:,1)*Pij(:,:,:,1) + 2*Pij(:,:,:,2)*Pij(:,:,:,2) + 2*Pij(:,:,:,3)*Pij(:,:,:,3) + Pij(:,:,:,4)*Pij(:,:,:,4) + 2*Pij(:,:,:,5)*Pij(:,:,:,5) + Pij(:,:,:,6)*Pij(:,:,:,6))
        DO j=1,6
          tauijr(:,:,:,j)=ur1*Pij(:,:,:,j)
        END DO
        DO j=1,6
          ur1=tauijr(:,:,:,j)
          CALL fft3ds(ur1,c1)
          subtotal(:,:,:,j)=subtotal(:,:,:,j)+c1
        END DO
      END IF
            
      tauij=subtotal
      DEALLOCATE(Sij)
      DEALLOCATE(subtotal)
      DEALLOCATE(Rij)
      DEALLOCATE(Pij)
      DEALLOCATE(tauijr)
      DEALLOCATE(c1) 
      DEALLOCATE(ur1)
    END IF
    
    DO qy=0, cy
      DO qz=0, cz
        DO qx=0, cx
          a(qx,qz,qy,1) = - ikx(qx)*tauij(qx,qz,qy,1) - iky(qy)*tauij(qx,qz,qy,2) - ikz(qz)*tauij(qx,qz,qy,3)
          a(qx,qz,qy,2) = - ikx(qx)*tauij(qx,qz,qy,2) - iky(qy)*tauij(qx,qz,qy,4) - ikz(qz)*tauij(qx,qz,qy,5)
          a(qx,qz,qy,3) = - ikx(qx)*tauij(qx,qz,qy,3) - iky(qy)*tauij(qx,qz,qy,5) - ikz(qz)*tauij(qx,qz,qy,6)   
        END DO
      END DO
    END DO  
    
    SELECT CASE (proj_u)
    CASE(1)
      ALLOCATE(ur1(0:rx,0:ry,0:rz))
      ALLOCATE(Sij(0:rx,0:ry,0:rz,3))
      ALLOCATE(Rij(0:rx,0:ry,0:rz,3))
      ALLOCATE(c1(0:cx,0:cz,0:cy))
      DO j=1,3
        c1=a(:,:,:,j)
        CALL ifft3ds(c1,ur1)
        Sij(:,:,:,j)=ur1
        c1=vel(:,:,:,j)
        CALL ifft3ds(c1,ur1)
        Rij(:,:,:,j)=ur1
      END DO
      ur1=0.0
      ur1 = ( Sij(:,:,:,1)*Rij(:,:,:,1) + Sij(:,:,:,2)*Rij(:,:,:,2) + Sij(:,:,:,3)*Rij(:,:,:,3) ) / ( Rij(:,:,:,1)*Rij(:,:,:,1) + Rij(:,:,:,2)*Rij(:,:,:,2) + Rij(:,:,:,3)*Rij(:,:,:,3) )
      Sij(:,:,:,1)=ur1*Rij(:,:,:,1)
      Sij(:,:,:,2)=ur1*Rij(:,:,:,2)
      Sij(:,:,:,3)=ur1*Rij(:,:,:,3)
      DO j=1,3
        ur1=Sij(:,:,:,j)
        CALL fft3ds(ur1,c1)
        a(:,:,:,j)=c1
      END DO
      DEALLOCATE(ur1)
      DEALLOCATE(c1)
      DEALLOCATE(Sij)
      DEALLOCATE(Rij)
    CASE(2)
      ALLOCATE(c1(0:cx,0:cz,0:cy))
      c1 = ( CONJG(vel(:,:,:,1))*a(:,:,:,1) +CONJG(vel(:,:,:,2))*a(:,:,:,2) + CONJG(vel(:,:,:,3))*a(:,:,:,3) ) / ( vel(:,:,:,1)*CONJG(vel(:,:,:,1)) + vel(:,:,:,2)*CONJG(vel(:,:,:,2)) + vel(:,:,:,3)*CONJG(vel(:,:,:,3)) )
      a(:,:,:,1)=c1*vel(:,:,:,1)
      a(:,:,:,2)=c1*vel(:,:,:,2)
      a(:,:,:,3)=c1*vel(:,:,:,3)
      DEALLOCATE(c1)
    END SELECT
    
    DEALLOCATE(tauij)
  END SELECT
  
END SUBROUTINE sgs_exchanging_tauij



!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!                                                          !!
!!               reduce_send_and_receive                    !!
!!                                                          !!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
SUBROUTINE reduce_send_and_receive(a)
  USE parameters
  USE mpi_var
  IMPLICIT NONE
    
  COMPLEX,  DIMENSION(0:cx,0:cz,0:cy),  INTENT(INOUT)  ::  a
  COMPLEX,  DIMENSION(0:cxl,0:czl)      ::  source_slice, destination_slice
  
  INTEGER  ::  qy, qqy, pid, qj
  
  SELECT CASE(LES_ident)
  CASE(0)
    DO qy=0,nyl/2
      pid=whereis_ky(qy,1)
      qqy=whereis_ky(qy,2)
      source_slice(:,0:nzl/2)    =a(0:cxl,0:nzl/2      ,qqy)
      source_slice(:,nzl/2+1:czl)=a(0:cxl,nz-nzl/2+1:cz,qqy)
      IF (io_fpart.ne.pid) THEN
        IF (my_fpart.eq.pid)      CALL MPI_SEND(source_slice     ,(cxl+1)*(czl+1),MPI_MYCOMPLEX,io_fpart,10,MPI_COMM_INSTANCE,ierr)
        IF (my_fpart.eq.io_fpart) CALL MPI_RECV(destination_slice,(cxl+1)*(czl+1),MPI_MYCOMPLEX,pid,10,MPI_COMM_INSTANCE,status_mpi,ierr)
      ELSE
        IF (my_fpart.eq.io_fpart) destination_slice=source_slice
      END IF
      IF(myid.EQ.ioid) THEN
        DO qj=numfparts,numprocs-1
          CALL MPI_SEND(destination_slice,(cxl+1)*(czl+1),MPI_MYCOMPLEX,qj,10,MPI_COMM_WORLD,ierr)
        END DO
      END IF
    END DO
    DO qy=nyl/2+1,cyl
      pid=whereis_ky(qy+ny-nyl,1)
      qqy=whereis_ky(qy+ny-nyl,2)
      source_slice(:,0:nzl/2)    =a(0:cxl,0:nzl/2      ,qqy)
      source_slice(:,nzl/2+1:czl)=a(0:cxl,nz-nzl/2+1:cz,qqy)
      IF (io_fpart.ne.pid) THEN
        IF (my_fpart.eq.pid)      CALL MPI_SEND(source_slice     ,(cxl+1)*(czl+1),MPI_MYCOMPLEX,io_fpart,10,MPI_COMM_INSTANCE,ierr)
        IF (my_fpart.eq.io_fpart) CALL MPI_RECV(destination_slice,(cxl+1)*(czl+1),MPI_MYCOMPLEX,pid,10,MPI_COMM_INSTANCE,status_mpi,ierr)
      ELSE
        IF (my_fpart.eq.io_fpart) destination_slice=source_slice
      END IF
      IF(myid.EQ.ioid) THEN
        DO qj=numfparts,numprocs-1
          CALL MPI_SEND(destination_slice,(cxl+1)*(czl+1),MPI_MYCOMPLEX,qj,10,MPI_COMM_WORLD,ierr)
        END DO
      END IF
    END DO
  CASE(1)
    DO qy=0,ny-1
      CALL MPI_RECV(a(:,:,qy),(cx+1)*(cz+1),MPI_MYCOMPLEX,ioid,10,MPI_COMM_WORLD,status_mpi,ierr)
    END DO
    a=Ns_ratio*a
  END SELECT
  
END SUBROUTINE reduce_send_and_receive



!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!                                                          !!
!!                compute_force_restore                     !!
!!                                                          !!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
SUBROUTINE compute_force_restore(force_restore)
  USE parameters
  USE variables
  USE mpi_var
  IMPLICIT NONE
  
  COMPLEX,  DIMENSION(0:cx,0:cz,0:cy,3),  INTENT(INOUT)  ::  force_restore
  COMPLEX,  DIMENSION(0:cx,0:cz,0:cy)  ::  c1
  COMPLEX,  DIMENSION(0:cx,0:cz,0:cy,3)  ::  vel_dns  
  INTEGER  ::  qj
  REAL     ::  fu, ff
  
  SELECT CASE (type_restore)
  CASE(1)
    DO qj=1,3
      vel_dns(:,:,:,qj)=vel(:,:,:,qj)
      CALL reduce_send_and_receive(vel_dns(:,:,:,qj))
    END DO
    force_restore=alpha_restore*(vel_dns-vel)
  CASE(2)
    DO qj=1,3
      vel_dns(:,:,:,qj)=vel(:,:,:,qj)
      CALL reduce_send_and_receive(vel_dns(:,:,:,qj))
      force_restore(:,:,:,qj)=alpha_restore*ksquare*(vel_dns(:,:,:,qj)-vel(:,:,:,qj))
    END DO
  END SELECT
  
  CALL divfree(force_restore)
    
END SUBROUTINE compute_force_restore



!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!                                                          !!
!!                   spec_rest_init                         !!
!!                                                          !!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
SUBROUTINE spec_rest_init()
  USE parameters
  USE mpi_var
  USE spec_rest_mod
  IMPLICIT NONE
  
  INTEGER   :: qx, qy, qz, ishell

  !! Determining the number of shells (cilinders)
  rest_nshell=CEILING((rest_klast-rest_kfirst)/rest_shell_width)

  !! Determining the shells boundaries
  ALLOCATE(rest_kshell(0:rest_nshell+2))
  rest_kshell(0)=0.0
  rest_kshell(1)=rest_kfirst
  DO ishell = 2, rest_nshell
    rest_kshell(ishell)=rest_kfirst+rest_shell_width*(ishell-1)
  END DO
  rest_kshell(rest_nshell+1)=rest_klast
  rest_kshell(rest_nshell+2)=REAL(MIN(nx,ny,nx))

  !! Determining the shells midpoint wavevector
  ALLOCATE(rest_kc(0:rest_nshell))
  rest_kc(0)=0.0
  DO ishell = 1, rest_nshell
	rest_kc(ishell)=rest_kfirst+rest_shell_width*(REAL(ishell)-0.5)
	IF(rest_kc(ishell) > rest_klast) rest_kc(ishell) = rest_klast
  END DO

  !! Determining the number of modes in a shell
  ALLOCATE(rest_nkshell(0:rest_nshell+1),rest_nkshell_node(0:rest_nshell+1)) 
  rest_nkshell_node=0.0
  DO qy=0,cy
   DO qz=0,cz 
      DO qx=0,cx
        DO ishell = 0, rest_nshell+1
          SELECT CASE(spec_rest_type)
          CASE(1)   !! 1D
            IF( ( ABS(kx(qx)) > rest_kshell(ishell)).AND.(ABS(kx(qx)) <= rest_kshell(ishell+1)) ) THEN
              rest_nkshell_node(ishell)=rest_nkshell_node(ishell)+MM(qx)
            END IF
          CASE(2,5)   !! 2D
            IF( ( knorm(qx,0,qy) > rest_kshell(ishell)).AND.(knorm(qx,0,qy) <= rest_kshell(ishell+1)) ) THEN
              rest_nkshell_node(ishell)=rest_nkshell_node(ishell)+MM(qx)
            END IF
          CASE(3,4) !! 3D
            IF( ( knorm(qx,qz,qy) > rest_kshell(ishell)).AND.(knorm(qx,qz,qy) <= rest_kshell(ishell+1)) ) THEN
              rest_nkshell_node(ishell)=rest_nkshell_node(ishell)+MM(qx)
            END IF
          CASE DEFAULT
          END SELECT 
        END DO
      END DO
    END DO
  END DO
  CALL mpi_allreduce(rest_nkshell_node,rest_nkshell,(rest_nshell+2),MPI_MYREAL,MPI_SUM,MPI_COMM_INSTANCE,ierr)
  
  rest_nkshell(0)= rest_nkshell(0) + 1.0

END SUBROUTINE spec_rest_init



!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!                                                          !!
!!                   spec_rest_write                        !!
!!                                                          !!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
SUBROUTINE spec_rest_write(spec_rest_a,filename)
  USE parameters
  USE mpi_var
  USE spec_rest_mod
  IMPLICIT NONE

  REAL, DIMENSION(0:rest_nshell+1,6), INTENT(IN)     :: spec_rest_a
  CHARACTER(LEN=namelength+11),  INTENT(IN)     :: filename
  CHARACTER(LEN=namelength+8)                   :: shortname
  INTEGER   :: ishell

  IF (my_fpart.eq.io_fpart) THEN
    IF(numinsts.GT.1)THEN
      OPEN (UNIT=10, FILE=filename, STATUS='UNKNOWN', FORM='FORMATTED')
    ELSE
      shortname=filename(1:namelength+8)
      OPEN (UNIT=10, FILE=shortname, STATUS='UNKNOWN', FORM='FORMATTED')
    END IF
    DO ishell = 0, rest_nshell
      WRITE(10,'(I10,2F15.2,8F21.8)')    ishell, rest_kc(ishell), rest_nkshell(ishell), &
                      spec_rest_a(ishell,1),  spec_rest_a(ishell,2),  spec_rest_a(ishell,3), &
                      spec_rest_a(ishell,1) + spec_rest_a(ishell,2) + spec_rest_a(ishell,3), &
                      spec_rest_a(ishell,4),  spec_rest_a(ishell,5),  spec_rest_a(ishell,6), &
                      spec_rest_a(ishell,4) + spec_rest_a(ishell,5) + spec_rest_a(ishell,6)
    END DO
    CLOSE (UNIT=10)
  END IF

END SUBROUTINE spec_rest_write



!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!                                                          !!
!!                 spec_rest_compute_3D                     !!
!!                                                          !!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
SUBROUTINE spec_rest_compute_3D(a, spec)
  USE parameters
  USE mpi_var
  USE spec_rest_mod
  IMPLICIT NONE
  
  COMPLEX, DIMENSION(0:cx,0:cz,0:cy,3), INTENT(IN)   :: a
  REAL, DIMENSION(0:rest_nshell+1,6), INTENT(OUT)    :: spec
  REAL, DIMENSION(0:rest_nshell+1,6)                 :: spec_rest_node
  INTEGER    :: qx, qy, qz, ishell
  REAL       :: u2, v2, w2

  spec_rest_node=0.0
  spec=0.0
  DO qy=0,cy
    DO qx=0,cx
      DO qz=0,cz
  		DO ishell=0,rest_nshell+1
          IF( (knorm(qx,qz,qy) > rest_kshell(ishell)).AND.(knorm(qx,qz,qy) <= rest_kshell(ishell+1)) ) THEN
            u2=ABS(a(qx,qz,qy,1))**2*MM(qx)
            v2=ABS(a(qx,qz,qy,2))**2*MM(qx)
            w2=ABS(a(qx,qz,qy,3))**2*MM(qx)
            spec_rest_node(ishell,1)=spec_rest_node(ishell,1)+u2*0.5
            spec_rest_node(ishell,2)=spec_rest_node(ishell,2)+v2*0.5
            spec_rest_node(ishell,3)=spec_rest_node(ishell,3)+w2*0.5
            spec_rest_node(ishell,4)=spec_rest_node(ishell,4)+u2*ksquare(qx,qz,qy)
            spec_rest_node(ishell,5)=spec_rest_node(ishell,5)+v2*ksquare(qx,qz,qy)
            spec_rest_node(ishell,6)=spec_rest_node(ishell,6)+w2*ksquare(qx,qz,qy)
          END IF
        END DO
      END DO
    END DO
  END DO  
  spec_rest_node=spec_rest_node/(Ck*Ns)**2
  CALL mpi_allreduce(spec_rest_node,spec,6*(rest_nshell+2),MPI_MYREAL,MPI_SUM,MPI_COMM_INSTANCE,ierr)

END SUBROUTINE spec_rest_compute_3D



!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!                                                          !!
!!               spec_rest_compute_3DN                      !!
!!                                                          !!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
SUBROUTINE spec_rest_compute_3DN(a, spec)
  USE parameters
  USE mpi_var
  USE spec_rest_mod
  IMPLICIT NONE
  
  COMPLEX, DIMENSION(0:cx,0:cz,0:cy,3), INTENT(IN)   :: a
  REAL, DIMENSION(0:rest_nshell+1,6), INTENT(OUT)    :: spec
  REAL, DIMENSION(0:rest_nshell+1,6)                 :: spec_rest_node
  INTEGER    :: qx, qy, qz, ishell, j
  REAL       :: u2, v2, w2

  spec = 0.0
  spec_rest_node = 0.0
  DO qy=0,cy
    DO qx=0,cx
      DO qz=0,cz
  		DO ishell=0,rest_nshell
          IF( (knorm(qx,qz,qy) > rest_kshell(ishell)).AND.(knorm(qx,qz,qy) <= rest_kshell(ishell+1)) ) THEN
            u2=ABS(a(qx,qz,qy,1))**2*MM(qx)
            v2=ABS(a(qx,qz,qy,2))**2*MM(qx)
            w2=ABS(a(qx,qz,qy,3))**2*MM(qx)
            spec_rest_node(ishell,1)=spec_rest_node(ishell,1)+u2*0.5
            spec_rest_node(ishell,2)=spec_rest_node(ishell,2)+v2*0.5
            spec_rest_node(ishell,3)=spec_rest_node(ishell,3)+w2*0.5
            spec_rest_node(ishell,4)=spec_rest_node(ishell,4)+u2*ksquare(qx,qz,qy)
            spec_rest_node(ishell,5)=spec_rest_node(ishell,5)+v2*ksquare(qx,qz,qy)
            spec_rest_node(ishell,6)=spec_rest_node(ishell,6)+w2*ksquare(qx,qz,qy)
          END IF
        END DO
      END DO
    END DO
  END DO  
  spec_rest_node=spec_rest_node/(Ck*Ns)**2
  CALL mpi_allreduce(spec_rest_node,spec,6*(rest_nshell+2),MPI_MYREAL,MPI_SUM,MPI_COMM_INSTANCE,ierr)

  DO j=1,6
    DO ishell=0,rest_nshell
      spec(ishell,j)=spec(ishell,j)*(4*pi*rest_kc(ishell)**2*rest_shell_width)/rest_nkshell(ishell)
    END DO
  END DO

END SUBROUTINE spec_rest_compute_3DN



!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!                                                          !!
!!               exchange_large_scales                      !!
!!                                                          !!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
SUBROUTINE exchange_large_scales()
  USE parameters
  USE variables
  USE mpi_var
  IMPLICIT NONE
  
  INTEGER   ::  qx, qy, qz, j, inst_dest
  
  DO qy=0,cy
    DO qz=0,cz
      DO qx=0,cx
        IF(knorm(qx,qz,qy).LE.(kM/Smag_const))THEN
          DO j=1,3
            IF(my_inst.EQ.0)THEN
              DO inst_dest=1,numinsts-1
                IF (VELEQ) CALL MPI_SEND(vel(qx,qz,qy,j), 1, MPI_MYCOMPLEX, inst_dest, 10, MPI_COMM_FPART, ierr)
                IF (MAGEQ) CALL MPI_SEND(mag(qx,qz,qy,j), 1, MPI_MYCOMPLEX, inst_dest, 20, MPI_COMM_FPART, ierr)
              END DO
            ELSE
              IF (VELEQ)   CALL MPI_RECV(vel(qx,qz,qy,j), 1, MPI_MYCOMPLEX, 0,         10, MPI_COMM_FPART, status_mpi, ierr)
              IF (MAGEQ)   CALL MPI_RECV(mag(qx,qz,qy,j), 1, MPI_MYCOMPLEX, 0,         20, MPI_COMM_FPART, status_mpi, ierr)
            END IF
          END DO
        END IF
      END DO
    END DO
  END DO
            
END SUBROUTINE exchange_large_scales