!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!  12 August 2011                                          !!
!!                       TURBO                              !!
!!         Solver for TURbulent flows with                  !!
!!           periodic BOundary conditions                   !!
!!                                             version 2.05 !!
!!                                                          !!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
PROGRAM turbo
  USE parameters
  USE variables
  USE mpi_var
  IMPLICIT NONE  

  INTEGER   :: iter, ninit, nfinal

!!! Initializations
  CALL mpi_environment_init()

  IF (myid.eq.ioid) WRITE (6,*) '>> Starting...' 
  IF (my_fpart.eq.io_fpart) WRITE(out_unit,*) '>> Simulation name is: ', simname, ' for instance ', my_inst 

  CALL read_para()  
  CALL read_time()
  
  ninit  = niter_done+1
  nfinal = MIN(niter_done+nstep, niter_todo)
  
  CALL init_variables()
  CALL external_stats_compute(niter_done)
  CALL output(0)

!!! Main loop
  DO iter=ninit,nfinal
    IF(field_exchange.EQ.1) THEN
      CALL compute_dt()
    ELSE
      IF (MOD(iter-1,2).eq.0) CALL compute_dt()
    END IF
    t=t+dt
    CALL update_variables(iter)
    niter_done=iter
    IF (MOD(iter-(ninit-1),nstat).eq.0) CALL external_stats_compute(iter)
    IF (MOD(iter-(ninit-1),nsave).eq.0) CALL output(1)
  END DO

!!! End of run
  IF (MOD(nfinal,nsave).ne.0) THEN
    CALL external_stats_compute(nfinal)
    CALL output(1)
  END IF
  
  IF (my_fpart.eq.io_fpart) THEN
    IF (nfinal.eq.niter_todo) THEN
      WRITE(out_unit,*) '>> The requested total number of time steps have been done'
      IF (myid.eq.ioid) WRITE(6,*) '>> The requested total number of time steps have been done'
    ELSE
      OPEN(10,FILE="turbo.go",STATUS='UNKNOWN',FORM='FORMATTED')
      CLOSE(10)
    END IF  
  END IF

  IF (VELEQ) DEALLOCATE(vel,dvel)
  IF (MAGEQ) DEALLOCATE(mag,dmag)
  IF (SCAEQ) DEALLOCATE(scal,dscal)

  CALL mpi_environment_end()

END PROGRAM turbo

