!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!  12 August 2011                                          !!
!!                      tt_mod.F90                          !!
!!                                                          !!
!!  parameters                                              !!
!!  variables                                               !!
!!  fftwmodule                                              !!
!!  mpi_var                                                 !!
!!                                             version 2.05 !!
!!                                                          !!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! 



!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!                                                          !!
!!                      parameters                          !!
!!                                                          !!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
MODULE parameters

!!!! 
!!!!    Useful constant parameters
!!!!
  REAL, PARAMETER         :: pi=3.1415926535897932384626433832795028841971693993751E0
  REAL, PARAMETER         :: EPS=1.0E-10
  REAL, PARAMETER         :: twopi = 2.0*pi
  COMPLEX, PARAMETER      :: I=(0,1.0E0)
  INTEGER, PARAMETER      :: namelength = 6
  INTEGER, PARAMETER      :: SIZEOF_REAL = KIND(1.0) 

!!!! 
!!!!    Useful variables and arrays
!!!!
  REAL                        :: rsx, rsy, rsz, sx, sy, sz
  INTEGER                     :: odd_even
  INTEGER(8)                  :: plan_CtoR, plan_RtoC
  CHARACTER(LEN=namelength)   :: simname
  
  INTEGER, ALLOCATABLE, DIMENSION (:)     ::  MM
  REAL, ALLOCATABLE, DIMENSION (:)        ::  kx, ky, kz
  COMPLEX, ALLOCATABLE, DIMENSION (:)     ::  ikx, iky, ikz
  REAL, ALLOCATABLE, DIMENSION (:,:,:)    ::  ksquare, knorm, km2, zeros_twothird, rot_omega

!!!!
!!!! Namelist dim_and_sizes and related quantities
!!!!
  REAL      :: lx, ly, lz
  INTEGER   :: nx, ny, nz
  NAMELIST /dim_and_sizes/ nx, ny, nz, lx, ly, lz
  
  REAL      :: kM, kMx, kMy, kMz
  REAL      :: k0, k0x, k0y, k0z 
  REAL      :: dx, dy, dz, Cr, Ck
  INTEGER   :: cx, cy, cz, rx, ry, rz, Ns
  
!!!!
!!!! Namelist equation_parameters and related quantities
!!!!
  INTEGER   :: FORCEPARA, CHIM, NSCAL, EQTYPE
  NAMELIST /equation_parameters/ EQTYPE, FORCEPARA, CHIM, NSCAL
  
  !! Possible values for the flag <<EQTYPE>>  
  INTEGER, PARAMETER :: reac_diff     = 0
  INTEGER, PARAMETER :: navier_stokes = 1
  INTEGER, PARAMETER :: qs_mhd        = 2
  INTEGER, PARAMETER :: full_mhd      = 3
  
  LOGICAL   :: VELEQ, MAGEQ, SCAEQ

!!!!
!!!! Namelist rot_parameters and related quantities
!!!!
  REAL      :: Omega_x, Omega_y, Omega_z
  NAMELIST /rot_parameters/ Omega_x, Omega_y, Omega_z
  
  LOGICAL   :: ROTEQ

!!!!
!!!! Namelist vel_parameters and related quantities
!!!!
  INTEGER   :: init_vel
  REAL      :: nu, u0_x, u0_y, u0_z, vel_a, vel_b, vel_c, vel_d
  NAMELIST /vel_parameters/ init_vel, nu, u0_x, u0_y, u0_z, vel_a, vel_b, vel_c, vel_d

!!!!
!!!! Namelist mag_parameters and related quantities
!!!!
  INTEGER   :: init_mag
  REAL      :: eta, b0_x, b0_y, b0_z, mag_a, mag_b, mag_c, mag_d
  NAMELIST /mag_parameters/ init_mag, eta, b0_x, b0_y, b0_z, mag_a, mag_b, mag_c, mag_d

  !! Possible values for the flag <<init_vel>> and <<init_mag>>
  INTEGER, PARAMETER :: restart      = 0
  INTEGER, PARAMETER :: random       = 1
  INTEGER, PARAMETER :: taylor_green = 2
  INTEGER, PARAMETER :: extern       = 3

!!!!
!!!! Namelist sca_parameters  
!!!!
  INTEGER, DIMENSION(1:99)   :: init_sca_dat
  REAL, DIMENSION(1:99)      :: kappa_dat, sca_a_dat, sca_b_dat, sca_c_dat, sca_d_dat
  NAMELIST /sca_parameters/ init_sca_dat, kappa_dat, sca_a_dat, sca_b_dat, sca_c_dat, sca_d_dat

  INTEGER, ALLOCATABLE, DIMENSION (:)  :: init_sca
  REAL, ALLOCATABLE, DIMENSION (:)     :: kappa, sca_a, sca_b, sca_c, sca_d

!!!!
!!!! Namelist forcing_parameters  
!!!!
  REAL      :: kinf, ksup, energy_forcing, helicity_para, Famp
  INTEGER   :: Fmode
  NAMELIST /forcing_parameters/ kinf, ksup, energy_forcing, helicity_para, Famp, Fmode

  INTEGER   :: modes_layer
  REAL      :: inj_energy, inj_helicity, inj_cross, inj_maghel, dt_f

!!!!
!!!! Namelist forcing_ub_parameters  
!!!!
  REAL      :: kin_energy_forcing, mag_energy_forcing, kin_hel_forcing, mag_hel_forcing, croshel_forcing
  NAMELIST /forcing_ub_parameters/ kinf, ksup, kin_energy_forcing, mag_energy_forcing, kin_hel_forcing, mag_hel_forcing, croshel_forcing

!!!!
!!!! Namelist numerics_parameters  
!!!!
  REAL      :: dt_dat, cfl
  INTEGER   :: fixed_dt, time_to_zero, dealias, userseed
  NAMELIST /numerics_parameters/ dt_dat, cfl, fixed_dt, time_to_zero, dealias, userseed

  !! valid values for the flag <<dealias>>  
  INTEGER, PARAMETER :: shifts   = 1
  INTEGER, PARAMETER :: twothird = 2
  
  REAL      :: dt
!!!!
!!!! Namelist time_parameters  
!!!!
  REAL      :: t
  INTEGER   :: niter_done, nfields
  NAMELIST /time_parameters/ t, niter_done, nfields

!!!!
!!!! Namelist save_parameters  
!!!!
  INTEGER   :: niter_todo, nstep, nsave, nstat, nbuild, totpart
  NAMELIST /save_parameters/  niter_todo, nstep, nsave, nstat, nbuild, totpart

!!!!
!!!! Namelist les_parameters and related LES quantities
!!!!
  INTEGER   :: LESTYPE, HV_order, SGS_type, Smag_dyn, Smag_averag, Smag_dir, Smag_ens_av, filter_type
  REAL      :: Smag_const
  NAMELIST /les_parameters/  LESTYPE, HV_order, SGS_type, Smag_dyn, Smag_averag, Smag_dir, Smag_ens_av, Smag_const, filter_type
  
  COMPLEX, DIMENSION(:,:,:), ALLOCATABLE   ::  test_filter, LES_filter, Gauss_filter, Cut_filter
  
!!!!
!!!! Namelist force_restore_parameters and related quantities
!!!!
  INTEGER   :: exch_tauij, LES_ident, type_restore, add_force_restore, add_total, add_1, add_2, add_3, add_4, add_5, proj_u
  REAL      :: alpha_restore
  NAMELIST /force_restore_parameters/  exch_tauij, add_total, add_1, add_2, add_3, add_4, add_5, proj_u, LES_ident, type_restore, add_force_restore, alpha_restore
  
  INTEGER  ::  nxl, nyl, nzl, cxl, cyl, czl, field_exchange
  REAL     ::  dt_ext
  REAL     ::  kMxl, kMyl, kMzl, Ns_ratio
  

END MODULE parameters



!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!                                                          !!
!!                      variables                           !!
!!                                                          !!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
MODULE variables

  COMPLEX, ALLOCATABLE, DIMENSION (:,:,:,:)     :: vel, mag, scal
  COMPLEX, ALLOCATABLE, DIMENSION (:,:,:,:)     :: dvel, dmag, dscal

END MODULE variables



!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!                                                          !!
!!                      fftwmodule                          !!
!!                                                          !!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
MODULE fftwmodule

!     This file contains PARAMETER statements for various constants
!     that can be passed to FFTW routines.  You should include
!     this file in any FORTRAN program that calls the fftw_f77
!     routines (either directly or with an #include statement
!     if you use the C preprocessor).

      integer FFTW_FORWARD,FFTW_BACKWARD
      parameter (FFTW_FORWARD=-1,FFTW_BACKWARD=1)

      integer FFTW_REAL_TO_COMPLEX,FFTW_COMPLEX_TO_REAL
      parameter (FFTW_REAL_TO_COMPLEX=-1,FFTW_COMPLEX_TO_REAL=1)

      integer FFTW_ESTIMATE,FFTW_MEASURE
      parameter (FFTW_ESTIMATE=0,FFTW_MEASURE=1)

      integer FFTW_OUT_OF_PLACE,FFTW_IN_PLACE,FFTW_USE_WISDOM
      parameter (FFTW_OUT_OF_PLACE=0)
      parameter (FFTW_IN_PLACE=8,FFTW_USE_WISDOM=16)

      integer FFTW_THREADSAFE
      parameter (FFTW_THREADSAFE=128)

!     Constants for the MPI wrappers:
      integer FFTW_TRANSPOSED_ORDER, FFTW_NORMAL_ORDER
      integer FFTW_SCRAMBLED_INPUT, FFTW_SCRAMBLED_OUTPUT
      parameter(FFTW_TRANSPOSED_ORDER=1, FFTW_NORMAL_ORDER=0)
      parameter(FFTW_SCRAMBLED_INPUT=8192)
      parameter(FFTW_SCRAMBLED_OUTPUT=16384)

END MODULE fftwmodule



!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!                                                          !!
!!                      mpi_var                             !!
!!                                                          !!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
MODULE mpi_var

  include 'mpif.h' 
   
  INTEGER, PARAMETER                   :: ioid = 0
  INTEGER, PARAMETER                   :: io_fpart = 0
  INTEGER                              :: myid, my_inst, my_fpart  
  INTEGER                              :: numprocs, numfparts
  INTEGER                              :: ierr  
  INTEGER, PARAMETER                   :: undef_id = -1
  INTEGER, PARAMETER                   :: out_unit_first = 100   
  INTEGER                              :: out_unit     
  INTEGER, DIMENSION(MPI_STATUS_SIZE)  :: status_mpi 
    
  INTEGER                              :: numinsts, same_numfp, dt_global   
  CHARACTER(LEN=6), DIMENSION(0:99)    :: instance_name
  INTEGER, DIMENSION(0:99)             :: fparts
  NAMELIST /ensemble/ numinsts, instance_name, same_numfp, fparts, dt_global

  INTEGER                              :: MPI_COMM_INSTANCE
  INTEGER                              :: MPI_COMM_FPART
  INTEGER, ALLOCATABLE, DIMENSION(:,:) :: which_id         
                                                               
  INTEGER :: local_nz, local_z_start
  INTEGER :: local_ny_after_transpose, local_y_start_after_transpose, local_nr
  
  INTEGER, ALLOCATABLE, DIMENSION (:)    ::  nodey, ystart, ysize, zstart, zsize
  INTEGER, ALLOCATABLE, DIMENSION (:,:)  ::  whereis_ky, whereis_z

#if SINGLE 
  INTEGER, PARAMETER      :: MPI_MYREAL=MPI_REAL            ! for simple precision runs only!
  INTEGER, PARAMETER      :: MPI_MYCOMPLEX=MPI_COMPLEX      ! for simple precision runs only!
#endif 

#if DOUBLE 
  INTEGER, PARAMETER      :: MPI_MYREAL=MPI_DOUBLE_PRECISION       ! for double precision runs only! 
  INTEGER, PARAMETER      :: MPI_MYCOMPLEX=MPI_DOUBLE_COMPLEX      ! for double precision runs only! 
#endif   
  
END MODULE mpi_var

