!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!  12 August 2011                                          !!
!!                     tt_stat.f90                          !!
!!                                                          !!
!!  correl_vector                                           !!
!!  correl_scalar                                           !!
!!  correl_helicity                                         !!
!!  compute_diss_QS                                         !!
!!  compute_energy_real                                     !!
!!  compute_Rlambda                                         !!
!!  compute_kmaxeta                                         !!
!!                                             version 2.05 !!
!!                                                          !!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! 



!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!                                                          !!
!!                      correl_vector                       !!
!!                                                          !!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
SUBROUTINE correl_vector(a1,a2,j,factor,result)
  USE mpi_var
  USE parameters
  IMPLICIT NONE
  
  COMPLEX, DIMENSION(0:cx,0:cz,0:cy,3), INTENT(IN) :: a1,a2
  INTEGER, INTENT(IN)     :: j
  REAL, INTENT(IN)        :: factor
  REAL, INTENT(OUT)       :: result
  DOUBLE PRECISION        :: result_dbl, result_node
  COMPLEX                 :: prx, pry, prz
  INTEGER                 :: qx, qy, qz
  
  result_node=0.0

  DO qy=0,cy
    DO qz=0,cz
      DO qx=0,cx
        prx=a1(qx,qz,qy,1)*CONJG(a2(qx,qz,qy,1))
        pry=a1(qx,qz,qy,2)*CONJG(a2(qx,qz,qy,2))
        prz=a1(qx,qz,qy,3)*CONJG(a2(qx,qz,qy,3))
        result_node=result_node+REAL(prx+pry+prz)*MM(qx)*MAX(ksquare(qx,qz,qy),EPS)**j
      END DO
    END DO
  END DO
  
  result_node=result_node*factor*Cr**2
  CALL mpi_allreduce(result_node,result_dbl,1,MPI_DOUBLE_PRECISION,MPI_SUM,MPI_COMM_INSTANCE,ierr)
  result=REAL(result_dbl)
  
END SUBROUTINE correl_vector



!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!                                                          !!
!!                      correl_scalar                       !!
!!                                                          !!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
SUBROUTINE correl_scalar(a1,a2,j,factor,result)
  USE mpi_var
  USE parameters
  IMPLICIT NONE

  COMPLEX, DIMENSION(0:cx,0:cz,0:cy), INTENT(IN) :: a1,a2
  INTEGER, INTENT(IN)     :: j  
  REAL, INTENT(IN)        :: factor
  REAL, INTENT(OUT)       :: result
  DOUBLE PRECISION        :: result_dbl, result_node
  COMPLEX                 :: prx
  INTEGER                 :: qx, qy, qz
  
  result_node=0.0
  
  DO qy=0,cy
    DO qz=0,cz
      DO qx=0,cx
        prx=a1(qx,qz,qy)*CONJG(a2(qx,qz,qy))
        result_node=result_node+REAL(prx)*MM(qx)*MAX(ksquare(qx,qz,qy),EPS)**j
      END DO
    END DO
  END DO
  
  result_node=result_node*factor*Cr**2
  CALL mpi_allreduce(result_node,result_dbl,1,MPI_DOUBLE_PRECISION,MPI_SUM,MPI_COMM_INSTANCE,ierr)
  result=REAL(result_dbl)
  
END SUBROUTINE correl_scalar



!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!                                                          !!
!!                     correl_helicity                      !!
!!                                                          !!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
SUBROUTINE correl_helicity(a1,a2,j,factor,result)
  USE mpi_var
  USE parameters
  IMPLICIT NONE

  COMPLEX, DIMENSION(0:cx,0:cz,0:cy,3), INTENT(IN) :: a1,a2
  INTEGER, INTENT(IN)     :: j  
  REAL, INTENT(IN)        :: factor
  REAL, INTENT(OUT)       :: result
  DOUBLE PRECISION        :: result_dbl, result_node
  COMPLEX                 :: prx, pry, prz
  INTEGER                 :: qx, qy, qz
  
  result_node=0.0
  
  DO qy=0,cy
    DO qz=0,cz
      DO qx=0,cx
        prx=(iky(qy)*a1(qx,qz,qy,3)-ikz(qz)*a1(qx,qz,qy,2))*CONJG(a2(qx,qz,qy,1))
        pry=(ikz(qz)*a1(qx,qz,qy,1)-ikx(qx)*a1(qx,qz,qy,3))*CONJG(a2(qx,qz,qy,2))
        prz=(ikx(qx)*a1(qx,qz,qy,2)-iky(qy)*a1(qx,qz,qy,1))*CONJG(a2(qx,qz,qy,3))
        result_node=result_node+REAL(prx+pry+prz)*MM(qx)*MAX(ksquare(qx,qz,qy),EPS)**j
      END DO
    END DO
  END DO
  
  result_node=result_node*factor*Cr**2
  CALL mpi_allreduce(result_node,result_dbl,1,MPI_DOUBLE_PRECISION,MPI_SUM,MPI_COMM_INSTANCE,ierr)
  result=REAL(result_dbl) 
  
END SUBROUTINE correl_helicity



!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!                                                          !!
!!                   compute_diss_QS                        !!
!!                                                          !!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
SUBROUTINE compute_diss_QS(u,omegaQS)
  USE mpi_var
  USE parameters
  IMPLICIT NONE

  COMPLEX, DIMENSION(0:cx,0:cz,0:cy,3), INTENT(IN) :: u
  REAL, INTENT(OUT)       :: omegaQS
  DOUBLE PRECISION        :: omega_dbl, omega_node
  INTEGER                 :: qx, qy, qz
  REAL                    :: k2
  
  omega_node=0.0
  DO qy=0,cy
    DO qz=0,cz
      DO qx=0,cx
        omega_node=omega_node+(CABS(u(qx,qz,qy,1))**2+CABS(u(qx,qz,qy,2))**2+  &
             CABS(u(qx,qz,qy,3))**2)*MM(qx)*(b0_x*kx(qx)+b0_y*ky(qy)+b0_z*kz(qz))**2*km2(qx,qz,qy)
      END DO
    END DO
  END DO

  omega_node = omega_node*Cr**2
  CALL mpi_allreduce(omega_node,omega_dbl,1,MPI_DOUBLE_PRECISION,MPI_SUM,MPI_COMM_INSTANCE,ierr)
  omegaQS=REAL(omega_dbl)

END SUBROUTINE compute_diss_QS



!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!                                                          !!
!!                 compute_energy_real                      !!
!!                                                          !!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
SUBROUTINE compute_energy_real(u,ener)
  USE parameters
  USE mpi_var
  IMPLICIT NONE

  COMPLEX, DIMENSION(0:cx,0:cz,0:cy,3), INTENT(IN) :: u
  REAL, INTENT(OUT)                                :: ener
  DOUBLE PRECISION                                 :: ener_node, ener_dbl
  REAL, ALLOCATABLE, DIMENSION(:,:,:)    :: r1
  COMPLEX, ALLOCATABLE, DIMENSION(:,:,:) :: c1
  INTEGER  :: j
    
  ALLOCATE(c1(0:cx,0:cz,0:cy),r1(0:rx,0:ry,0:rz))
  ener_node=0.0
  
  DO j=1,3
    c1=u(:,:,:,j)
    CALL ifft3ds(c1,r1)
    r1(nx:nx+1,:,:)=0.0
    ener_node=ener_node+SUM(r1*r1)/2.0/Ns
  END DO
  DEALLOCATE(c1,r1)
  
  CALL mpi_allreduce(ener_node,ener_dbl,1,MPI_DOUBLE_PRECISION,MPI_SUM,MPI_COMM_INSTANCE,ierr) 
  ener=REAL(ener_dbl)

END SUBROUTINE compute_energy_real



!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!                                                          !!
!!                      compute_Rlambda                     !!
!!                                                          !!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
SUBROUTINE compute_Rlambda(energy,dissipation,Rlambda)
  USE parameters
  IMPLICIT NONE

  REAL, INTENT(IN)         :: energy,dissipation
  REAL, INTENT(OUT)        :: Rlambda
  REAL                     :: uprime, lambda
  
  uprime         = SQRT(energy*2.0/3.0)
  lambda         = SQRT(15.0*nu*uprime*uprime/MAX(dissipation,EPS))
  Rlambda        = uprime*lambda/MAX(nu,EPS)
  
END SUBROUTINE compute_Rlambda



!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!                                                          !!
!!                    compute_kmaxeta                       !!
!!                                                          !!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
SUBROUTINE compute_kmaxeta(dissipation,kmaxeta)
  USE parameters
  IMPLICIT NONE

  REAL, INTENT(IN)      :: dissipation
  REAL, INTENT(OUT)     :: kmaxeta
  REAL                  :: ldiss
  
  ldiss   = SQRT(SQRT((nu*nu*nu/MAX(dissipation,EPS))))
  kmaxeta = kM * ldiss

END SUBROUTINE compute_kmaxeta
