!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!  12 August 2011                                          !!
!!                     tt_update.f90                        !!
!!                                                          !!
!!  update_variables                                        !!
!!  rk3_twothird                                            !!
!!  rk4_shifts                                              !!
!!  rk_nlstep                                               !!
!!  rk_shiftgrid                                            !!
!!  rk_nonlin                                               !!
!!  rk_nonlin_add                                           !!
!!  rk_expon                                                !!
!!  rk_endstep                                              !!
!!  compute_dt                                              !!
!!                                             version 2.05 !!
!!                                                          !!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! 



!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!                                                          !!
!!                     update_variables                     !!
!!                                                          !!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
SUBROUTINE update_variables(iter)
  USE parameters
  USE variables
  IMPLICIT NONE

  INTEGER, INTENT(IN) :: iter
  
  IF (VELEQ) THEN 
    dvel=0.0
    inj_energy   = 0.0
    inj_helicity = 0.0
  END IF
  IF (MAGEQ) dmag=0.0
  IF (SCAEQ) dscal=0.0
  IF(field_exchange.EQ.1) dt_ext=1e10

  SELECT CASE(dealias)
  CASE(shifts)
    CALL rk4_shifts(iter)
  CASE(twothird)
    CALL rk3_twothird()
  CASE DEFAULT
    WRITE(6,*) 'Wrong value of the dealiasing parameter',dealias
    STOP
  END SELECT
  
END SUBROUTINE update_variables



!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!                                                          !!
!!                       rk3_twothird                       !!
!!                                                          !!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
SUBROUTINE rk3_twothird()
  USE parameters
  USE variables
  IMPLICIT NONE

  REAL, PARAMETER :: al1=+1.0/2.0, al2=-2.0/9.0, al3=+1.0
  REAL, PARAMETER :: be1=+1.0/3.0, be2=+2.0/3.0, be3=-1.0/2.0
  REAL, PARAMETER :: ga1=-1.0/2.0, ga2=+4.0/3.0, ga3=+1.0/2.0
  
!!!!!!   First Runge-Kutta step
  CALL rk_nlstep(al1,ga1,1)
  CALL rk_expon(dt/3.0)
  CALL rk_endstep(be1)

!!!!!!   Second Runge-Kutta step
  CALL rk_nlstep(al2,ga2,2)
  CALL rk_expon(dt/3.0)
  CALL rk_endstep(be2)

!!!!!!   Third Runge Kutta step
  CALL rk_nlstep(al3,ga3,3)
  CALL rk_expon(dt/3.0)
  CALL rk_endstep(be3)

END SUBROUTINE rk3_twothird



!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!                                                          !!
!!                        rk4_shifts                        !!
!!                                                          !!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
SUBROUTINE rk4_shifts(iter)
  USE parameters
  USE variables
  USE mpi_var
  IMPLICIT NONE

  INTEGER, INTENT(IN) :: iter
  REAL                :: sh_x, sh_y, sh_z
 
  REAL, PARAMETER :: al1=+1.0/3.0, al2=-3.0/8.0, al3=+1.0/4.0, al4=+1.0/4.0
  REAL, PARAMETER :: be1=+1.0/3.0, be2=-3.0/4.0, be3=+1.0/2.0, be4=-1.0/6.0
  REAL, PARAMETER :: ga1=1.0, ga2=-3.0/2.0, ga3=0.0, ga4=0.0


!!!!!!   COMPUTE THE SHIFTS
  odd_even=MOD(iter-1,2)
  IF (odd_even.eq.0) THEN
    IF (my_fpart.eq.io_fpart) THEN
      CALL RANDOM_NUMBER(rsx)
      rsx=(0.5-rsx)*dx
      CALL RANDOM_NUMBER(rsy)
      rsy=(0.5-rsy)*dy
      CALL RANDOM_NUMBER(rsz)
      rsz=(0.5-rsz)*dz
    END IF
    CALL MPI_BCAST(rsx,1,MPI_MYREAL,io_fpart,MPI_COMM_INSTANCE,ierr)
    CALL MPI_BCAST(rsy,1,MPI_MYREAL,io_fpart,MPI_COMM_INSTANCE,ierr)
    CALL MPI_BCAST(rsz,1,MPI_MYREAL,io_fpart,MPI_COMM_INSTANCE,ierr)
  END IF
  

!!!!!!   First Runge-Kutta step
  CALL rk_nlstep(al1,ga1,1)
  CALL rk_expon(2.0*dt/3.0)
  CALL rk_endstep(be1)

!!!!!!   Second Runge-Kutta step (no rk_expon since xi2=0)
  CALL rk_nlstep(al2,ga2,2)
  CALL rk_endstep(be2)

!!!!!!   Third Runge Kutta step (no rk_expon since xi3=0)
  CALL rk_nlstep(al3,ga3,3)
  CALL rk_endstep(be3)

!!!!!!   Fourth Runge Kutta step
  CALL rk_nlstep(al4,ga4,4)
  CALL rk_expon(dt/3.0)  
  CALL rk_endstep(be4)

END SUBROUTINE rk4_shifts



!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!                                                          !!
!!                        rk_nlstep                         !!
!!                                                          !!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
SUBROUTINE rk_nlstep(al,ga,substep)
  USE parameters
  USE variables
  USE mpi_var
  IMPLICIT NONE

  REAL, INTENT(IN)    :: al, ga
  INTEGER, INTENT(IN) :: substep
  REAL                :: sh_x, sh_y, sh_z
  REAL                :: maxvec_node, maxvec  
  INTEGER             :: ii
  REAL, ALLOCATABLE, DIMENSION(:,:,:)      :: vecx, vecy, vecz  
  COMPLEX, ALLOCATABLE, DIMENSION(:,:,:)   :: vecc  
  COMPLEX, ALLOCATABLE, DIMENSION(:,:,:,:) :: force, force_sgs, force_restore
  COMPLEX, ALLOCATABLE, DIMENSION(:,:,:,:) :: force_u, force_b
  

!!!  Prepare the shifts if the grid-shifting procedure is used for alias removal
  IF (dealias.eq.shifts) THEN
    sh_x=(2.0*odd_even-1.0)*dx*0.25
    sh_y=(2.0*odd_even-1.0)*dy*0.25
    sh_z=(2.0*odd_even-1.0)*dz*0.25
    SELECT CASE (substep)
    CASE(1)      
      sx=+sh_x; sy=+sh_y; sz=+sh_z
    CASE(2)     
      sx=-sh_x; sy=+sh_y; sz=+sh_z
    CASE(3)
      sx=+sh_x; sy=-sh_y; sz=+sh_z
    CASE(4)
      sx=+sh_x; sy=+sh_y; sz=-sh_z
    END SELECT  
    sx=sx+rsx; sy=sy+rsy; sz=sz+rsz
    CALL rk_shiftgrid(sx,sy,sz)
  END IF
  
  
!!!  Exchange the large scales
  CALL exchange_large_scales()


!!!  Compute the nonlinear term and add it to the variables and the work arrays
  CALL rk_nonlin(al,ga)


!!! Add the SGS forcing term    
  SELECT CASE (SGS_type)
  CASE(1,2,3,4)
    ALLOCATE(force_sgs(0:cx,0:cz,0:cy,3))
    CALL compute_sgs(force_sgs)
    IF(dealias.EQ.twothird) THEN
      DO ii=1,3
        force_sgs(:,:,:,ii)=force_sgs(:,:,:,ii)*zeros_twothird
      END DO
    END IF    
    dvel = dvel + force_sgs*ga
    vel  =  vel + force_sgs*al*dt 
    DEALLOCATE(force_sgs)
  END SELECT  


!!! Exchanging Tauij
  IF(exch_tauij.GT.0) THEN
    ALLOCATE(force_sgs(0:cx,0:cz,0:cy,3))
    CALL sgs_exchanging_tauij(force_sgs)
    IF(dealias.EQ.twothird) THEN
      DO ii=1,3
        force_sgs(:,:,:,ii)=force_sgs(:,:,:,ii)*zeros_twothird
      END DO
    END IF        
    IF ((dealias==shifts.AND.substep==4).OR.(dealias==twothird.AND.substep==3)) THEN
      IF(LES_ident.EQ.0)THEN
        maxvec_node=cfl/dt_ext
      ELSE
        ALLOCATE(vecx(0:rx,0:ry,0:rz),vecy(0:rx,0:ry,0:rz),vecz(0:rx,0:ry,0:rz),vecc(0:cx,0:cz,0:cy))
        vecc=force_sgs(:,:,:,1)
        CALL ifft3ds(vecc,vecx)
        vecc=force_sgs(:,:,:,2)
        CALL ifft3ds(vecc,vecy)
        vecc=force_sgs(:,:,:,3)
        CALL ifft3ds(vecc,vecz)
        maxvec_node=MAX(EPS,pi*MAXVAL(ABS(vecx)/dx+ABS(vecy)/dy+ABS(vecz)/dz))
        DEALLOCATE(vecx,vecy,vecz,vecc)
      END IF
      CALL mpi_allreduce(maxvec_node,maxvec,1,MPI_MYREAL,MPI_MAX,MPI_COMM_WORLD,ierr)
      dt_ext=MIN(dt_ext,cfl/maxvec)    
    END IF    
    IF (LES_ident.EQ.1) THEN
      dvel = dvel + force_sgs*ga
      vel  =  vel + force_sgs*al*dt 
    END IF
    DEALLOCATE(force_sgs)
    
  END IF

  
!!! Add the force restore term
  IF(add_force_restore.EQ.1) THEN
    ALLOCATE(force_restore(0:cx,0:cz,0:cy,3))  
    CALL compute_force_restore(force_restore) 
    IF(dealias.EQ.twothird) THEN
      DO ii=1,3
        force_restore(:,:,:,ii)=force_restore(:,:,:,ii)*zeros_twothird
      END DO
    END IF
    IF ((dealias==shifts.AND.substep==4).OR.(dealias==twothird.AND.substep==3)) THEN
      IF(LES_ident.EQ.0)THEN
        maxvec_node=cfl/dt_ext
      ELSE
        ALLOCATE(vecx(0:rx,0:ry,0:rz),vecy(0:rx,0:ry,0:rz),vecz(0:rx,0:ry,0:rz),vecc(0:cx,0:cz,0:cy))
        vecc=force_restore(:,:,:,1)
        CALL ifft3ds(vecc,vecx)
        vecc=force_restore(:,:,:,2)
        CALL ifft3ds(vecc,vecy)
        vecc=force_restore(:,:,:,3)
        CALL ifft3ds(vecc,vecz)
        maxvec_node=MAX(EPS,pi*MAXVAL(ABS(vecx)/dx+ABS(vecy)/dy+ABS(vecz)/dz))
        DEALLOCATE(vecx,vecy,vecz,vecc)
      END IF
      CALL mpi_allreduce(maxvec_node,maxvec,1,MPI_MYREAL,MPI_MAX,MPI_COMM_WORLD,ierr)
      dt_ext=MIN(dt_ext,cfl/maxvec)  
    END IF
    IF ((add_force_restore.EQ.1).AND.(LES_ident.EQ.1)) THEN    
      dvel = dvel + force_restore*ga
      vel  =  vel + force_restore*al*dt
    END IF
    DEALLOCATE(force_restore)
  END IF

    
!!!  Dealiasing : grid shifting or two-third truncation
  SELECT CASE(dealias)
  CASE(shifts)
    CALL rk_shiftgrid(-sx,-sy,-sz)
  CASE(twothird)
    IF (VELEQ) THEN
      DO ii=1,3
        vel(:,:,:,ii)=vel(:,:,:,ii)*zeros_twothird
        dvel(:,:,:,ii)=dvel(:,:,:,ii)*zeros_twothird
      END DO
    END IF
    IF (MAGEQ) THEN
      DO ii=1,3
        mag(:,:,:,ii)=mag(:,:,:,ii)*zeros_twothird
        dmag(:,:,:,ii)=dmag(:,:,:,ii)*zeros_twothird
      END DO
    END IF
    IF (SCAEQ) THEN
      DO ii=1,NSCAL
        scal(:,:,:,ii)=scal(:,:,:,ii)*zeros_twothird
        dscal(:,:,:,ii)=dscal(:,:,:,ii)*zeros_twothird
     END DO
    END IF
  END SELECT
  

!!!!! Impose divergence free condition
  IF (VELEQ) THEN
    CALL divfree(vel)
    CALL divfree(dvel)
  ENDIF
  IF (MAGEQ) THEN
    CALL divfree(mag)
    CALL divfree(dmag)
  ENDIF


!!! Add the forcing term
  IF (VELEQ) THEN
    SELECT CASE (FORCEPARA)
    CASE(1,3,4,6)
      ALLOCATE(force(0:cx,0:cz,0:cy,3))
      CALL forcing(force)
      IF ((dealias==shifts.AND.substep==4).OR.(dealias==twothird.AND.substep==3)) THEN
        CALL correl_vector(vel,force,0,1.0,inj_energy)
        CALL correl_helicity(vel,force,0,2.0,inj_helicity)
      END IF
      dvel = dvel + force*ga
      vel  =  vel + force*al*dt
      DEALLOCATE(force)
    CASE(2,5)
      CALL modify_vel(vel,inj_energy,inj_helicity)
    CASE(7,8,9,10)
      ALLOCATE(force_u(0:cx,0:cz,0:cy,3),force_b(0:cx,0:cz,0:cy,3))
      CALL forcing_ub(force_u, force_b)
      IF ((dealias==shifts.AND.substep==4).OR.(dealias==twothird.AND.substep==3)) THEN
        CALL correl_vector(mag,force_u,0,1.0,inj_energy)   
        CALL correl_vector(vel,force_b,0,1.0,inj_cross)      
        inj_cross=inj_cross+inj_energy
	    CALL correl_vector(vel,force_u,0,1.0,inj_helicity) 
        CALL correl_vector(mag,force_b,0,1.0,inj_energy) 
        inj_energy=inj_energy+inj_helicity
	    CALL correl_helicity(force_u,vel,0,2.0,inj_helicity) 
	    CALL correl_helicity(force_b,mag,-1,2.0,inj_maghel)  
	  END IF
      dvel = dvel + force_u*ga
      dmag = dmag + force_b*ga
      vel  =  vel + force_u*al*dt
      mag  =  mag + force_b*al*dt
      DEALLOCATE(force_u,force_b)  
    END SELECT
  END IF
    
END SUBROUTINE rk_nlstep



!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!                                                          !!
!!                      rk_shiftgrid                        !!
!!                                                          !!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
SUBROUTINE rk_shiftgrid(shiftx,shifty,shiftz)
  USE parameters
  USE variables
  IMPLICIT NONE

  REAL , INTENT(IN)  :: shiftx, shifty, shiftz
  INTEGER            :: qx, qy, qz, iscal, ii
  COMPLEX, ALLOCATABLE, DIMENSION(:)     ::  ex, ey, ez
  COMPLEX, ALLOCATABLE, DIMENSION(:,:,:) ::  expo

  ALLOCATE(ex(0:cx),ey(0:cy),ez(0:cz))
  ALLOCATE(expo(0:cx,0:cz,0:cy))
  ex=CMPLX(cos(kx*shiftx),sin(kx*shiftx))
  ey=CMPLX(cos(ky*shifty),sin(ky*shifty))
  ez=CMPLX(cos(kz*shiftz),sin(kz*shiftz))
  
  DO qy=0, cy
    DO qz=0, cz
      DO qx=0, cx
        expo(qx,qz,qy)=ex(qx)*ey(qy)*ez(qz)
      END DO
    END DO
  END DO
       
  IF (VELEQ) THEN
    DO ii=1,3
      vel(:,:,:,ii)  =  vel(:,:,:,ii)*expo
      dvel(:,:,:,ii) = dvel(:,:,:,ii)*expo
    END DO
  END IF
        
  IF (MAGEQ) THEN
    DO ii=1,3
      mag(:,:,:,ii)  =  mag(:,:,:,ii)*expo
      dmag(:,:,:,ii) = dmag(:,:,:,ii)*expo
    END DO
  END IF
        
  IF (SCAEQ) THEN
    DO iscal=1,NSCAL
      scal(:,:,:,iscal)  =  scal(:,:,:,iscal)*expo
      dscal(:,:,:,iscal) = dscal(:,:,:,iscal)*expo
    END DO
  END IF
  DEALLOCATE(ex,ey,ez,expo)  
    
END SUBROUTINE rk_shiftgrid



!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!                                                          !!
!!                       rk_nonlin                          !!
!!                                                          !!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
SUBROUTINE rk_nonlin(al,ga)
  USE parameters
  USE variables
  IMPLICIT NONE
  
  REAL,    INTENT(IN)                       :: al, ga
  REAL,    ALLOCATABLE, DIMENSION (:,:,:)   :: ur1, ur2, ur3, br1, br2, br3, sr, prod
  COMPLEX, ALLOCATABLE, DIMENSION (:,:,:)   :: c1
  INTEGER                                   :: iscal


!!! Allocate arrays for the expression of the velocity and the magnetic field
!!! in real space if required and make the IFTT
  ALLOCATE(c1(0:cx,0:cz,0:cy))
  IF (VELEQ)  THEN 
    ALLOCATE(ur1(0:rx,0:ry,0:rz),ur2(0:rx,0:ry,0:rz),ur3(0:rx,0:ry,0:rz))
    c1=vel(:,:,:,1)
    CALL ifft3ds(c1,ur1)
    c1=vel(:,:,:,2)
    CALL ifft3ds(c1,ur2)
    c1=vel(:,:,:,3)
    CALL ifft3ds(c1,ur3)
  END IF
  IF (MAGEQ) THEN 
    ALLOCATE(br1(0:rx,0:ry,0:rz),br2(0:rx,0:ry,0:rz),br3(0:rx,0:ry,0:rz))
    c1=mag(:,:,:,1)
    CALL ifft3ds(c1,br1)
    c1=mag(:,:,:,2)
    CALL ifft3ds(c1,br2)
    c1=mag(:,:,:,3)
    CALL ifft3ds(c1,br3)
  END IF
 

!!!  Compute rk_nonlin_add products for the velocity field
  ALLOCATE(prod(0:rx,0:ry,0:rz))  

  IF (VELEQ) THEN
  !!!  Term 11 - 22 u
    prod = ur1*ur1-ur2*ur2
    IF (MAGEQ) prod=prod-br1*br1+br2*br2
    CALL fft3ds(prod,c1)
    CALL rk_nonlin_add(c1,vel(:,:,:,1),dvel(:,:,:,1),al*dt,ga,1)
     
  !!!  Term 12 u
    prod = ur1*ur2
    IF (MAGEQ) prod=prod-br1*br2
    CALL fft3ds(prod,c1)
    CALL rk_nonlin_add(c1,vel(:,:,:,1),dvel(:,:,:,1),al*dt,ga,2)
    CALL rk_nonlin_add(c1,vel(:,:,:,2),dvel(:,:,:,2),al*dt,ga,1)

  !!!  Term 13 u  
    prod = ur1*ur3
    IF (MAGEQ) prod=prod-br1*br3
    CALL fft3ds(prod,c1)
    CALL rk_nonlin_add(c1,vel(:,:,:,1),dvel(:,:,:,1),al*dt,ga,3)
    CALL rk_nonlin_add(c1,vel(:,:,:,3),dvel(:,:,:,3),al*dt,ga,1)

  !!!  Term 23 u  
    prod = ur2*ur3
    IF (MAGEQ) prod=prod-br2*br3
    CALL fft3ds(prod,c1)
    CALL rk_nonlin_add(c1,vel(:,:,:,2),dvel(:,:,:,2),al*dt,ga,3)
    CALL rk_nonlin_add(c1,vel(:,:,:,3),dvel(:,:,:,3),al*dt,ga,2)

  !!!  Term 33 - 22 u
    prod = ur3*ur3-ur2*ur2
    IF (MAGEQ) prod=prod-br3*br3+br2*br2
    CALL fft3ds(prod,c1)
    CALL rk_nonlin_add(c1,vel(:,:,:,3),dvel(:,:,:,3),al*dt,ga,3)
    IF ((.not.SCAEQ).and.(.not.MAGEQ)) DEALLOCATE(ur1,ur2,ur3)
  END IF


!!!  Compute rk_nonlin_add products for the magnetic field
  IF (MAGEQ) THEN
  !!!  Term 12 b
    prod = ur1*br2-br1*ur2
    CALL fft3ds(prod,c1)
    CALL rk_nonlin_add(-c1,mag(:,:,:,1),dmag(:,:,:,1),al*dt,ga,2)
    CALL rk_nonlin_add( c1,mag(:,:,:,2),dmag(:,:,:,2),al*dt,ga,1)
  !!!  Term 13 b
    prod = ur1*br3-br1*ur3
    CALL fft3ds(prod,c1)
    CALL rk_nonlin_add(-c1,mag(:,:,:,1),dmag(:,:,:,1),al*dt,ga,3)
    CALL rk_nonlin_add( c1,mag(:,:,:,3),dmag(:,:,:,3),al*dt,ga,1)
  !!!  Term 23 b
    prod = ur2*br3-br2*ur3
    CALL fft3ds(prod,c1)
    CALL rk_nonlin_add(-c1,mag(:,:,:,2),dmag(:,:,:,2),al*dt,ga,3)
    CALL rk_nonlin_add( c1,mag(:,:,:,3),dmag(:,:,:,3),al*dt,ga,2)
    DEALLOCATE(br1,br2,br3)
    IF (.not.SCAEQ) DEALLOCATE(ur1,ur2,ur3)
  END IF


!!!  Compute rk_nonlin_add products for the scalar fields
  IF (SCAEQ.and.VELEQ) THEN
    ALLOCATE (sr(0:rx,0:ry,0:rz))
    DO iscal=1,NSCAL
      c1 = scal(:,:,:,iscal)
      CALL ifft3ds(c1,sr) 
    !!!  Term us
      prod = sr*ur1
      CALL fft3ds(prod,c1)
      CALL rk_nonlin_add( c1,scal(:,:,:,iscal),dscal(:,:,:,iscal),al*dt,ga,1)
    !!!  Term vs
      prod = sr*ur2
      CALL fft3ds(prod,c1)
      CALL rk_nonlin_add( c1,scal(:,:,:,iscal),dscal(:,:,:,iscal),al*dt,ga,2)
    !!!  Term ws
      prod = sr*ur3
      CALL fft3ds(prod,c1)
      CALL rk_nonlin_add( c1,scal(:,:,:,iscal),dscal(:,:,:,iscal),al*dt,ga,3)
    END DO
    DEALLOCATE(sr,ur1,ur2,ur3)
  END IF
  
  DEALLOCATE(c1,prod)

END SUBROUTINE rk_nonlin



!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!                                                          !!
!!                     rk_nonlin_add                        !!
!!                                                          !!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
SUBROUTINE rk_nonlin_add(c1,var,dvar,fac1,fac2,igrad)
  USE parameters
  IMPLICIT NONE
  
  COMPLEX, DIMENSION(0:cx,0:cz,0:cy), INTENT(IN)       :: c1
  COMPLEX, DIMENSION(0:cx,0:cz,0:cy), INTENT(INOUT)    :: var,dvar
  REAL, INTENT(IN)                                     :: fac1,fac2
  INTEGER, INTENT(IN)                                  :: igrad
  COMPLEX, ALLOCATABLE, DIMENSION(:)                   :: d1, d2
  
  INTEGER :: qx, qy, qz
    
  SELECT CASE (igrad)
  CASE(1)
    ALLOCATE(d1(0:cx),d2(0:cx))
    d1=-fac1*ikx
    d2=-fac2*ikx
    DO qy=0, cy
      DO qz=0, cz
        DO qx=0, cx
          var(qx,qz,qy)  =  var(qx,qz,qy)+d1(qx)*c1(qx,qz,qy)
          dvar(qx,qz,qy) = dvar(qx,qz,qy)+d2(qx)*c1(qx,qz,qy)
        END DO
      END DO
    END DO
  CASE(2)
    ALLOCATE(d1(0:cy),d2(0:cy))
    d1=-fac1*iky
    d2=-fac2*iky
    DO qy=0, cy
      DO qz=0, cz
        DO qx=0, cx
          var(qx,qz,qy)  =  var(qx,qz,qy)+d1(qy)*c1(qx,qz,qy)
          dvar(qx,qz,qy) = dvar(qx,qz,qy)+d2(qy)*c1(qx,qz,qy)
        END DO
      END DO
    END DO
  CASE(3)
    ALLOCATE(d1(0:cz),d2(0:cz))
    d1=-fac1*ikz
    d2=-fac2*ikz
    DO qy=0, cy
      DO qz=0, cz
        DO qx=0, cx
          var(qx,qz,qy)  =  var(qx,qz,qy)+d1(qz)*c1(qx,qz,qy)
          dvar(qx,qz,qy) = dvar(qx,qz,qy)+d2(qz)*c1(qx,qz,qy)
        END DO
      END DO
    END DO
  CASE default
    WRITE(*,*) 'Wrong value of igrad (igrad=', igrad,')'
    STOP
  END SELECT

  DEALLOCATE(d1,d2)
  
END SUBROUTINE rk_nonlin_add



!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!                                                          !!
!!                       rk_expon                           !!
!!                                                          !!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
SUBROUTINE rk_expon(fac)
  USE parameters
  USE variables
  IMPLICIT NONE
  
  REAL, INTENT(IN)                     :: fac
  INTEGER                              :: iscal, icomp, qx, qy, qz, ii
  REAL                                 :: fac1, fac2, cos_omega, sin_omega
  REAL, ALLOCATABLE, DIMENSION(:,:,:)  :: expon
  REAL, ALLOCATABLE, DIMENSION(:,:)    :: rot_opera 
  COMPLEX, ALLOCATABLE, DIMENSION(:)   :: temp, dtemp

  ALLOCATE(expon(0:cx,0:cz,0:cy))

  IF (VELEQ) THEN
    fac1=-fac*nu
    IF (EQTYPE==qs_mhd) THEN 
      fac2=-fac/eta
      DO qy=0, cy
        DO qz=0, cz
          DO qx=0, cx
            expon(qx,qz,qy)=EXP(fac1*ksquare(qx,qz,qy)+fac2*(b0_x*kx(qx)+b0_y*ky(qy)+b0_z*kz(qz))**2*km2(qx,qz,qy))
          END DO
        END DO
      END DO
    ELSE 
      expon=EXP(fac1*ksquare)
    END IF
    DO icomp=1,3
      vel(:,:,:,icomp)  =  vel(:,:,:,icomp)*expon
      dvel(:,:,:,icomp) = dvel(:,:,:,icomp)*expon
    END DO
  END IF
  
  IF (MAGEQ) THEN
    fac1=-fac*eta
    expon=EXP(fac1*ksquare)
    DO icomp=1,3
      mag(:,:,:,icomp)  =  mag(:,:,:,icomp)*expon
      dmag(:,:,:,icomp) = dmag(:,:,:,icomp)*expon
    END DO
  END IF
    
  IF (SCAEQ) THEN
    DO iscal=1,NSCAL
      fac2=-fac*kappa(iscal) 
      expon=EXP(fac2*ksquare)
      scal(:,:,:,iscal)  =  scal(:,:,:,iscal)*expon
      dscal(:,:,:,iscal) = dscal(:,:,:,iscal)*expon
    END DO
  END IF

  DEALLOCATE(expon)

  IF (ROTEQ) THEN
    ALLOCATE(rot_opera(3,3))
    ALLOCATE(temp(3),dtemp(3))
    rot_opera=0.0
    DO qy=0, cy
      DO qz=0, cz
        DO qx=0, cx
          cos_omega=COS(rot_omega(qx,qz,qy)*fac)
          sin_omega=SIN(rot_omega(qx,qz,qy)*fac)
          rot_opera(1,1)=cos_omega
          rot_opera(2,2)=cos_omega
          rot_opera(3,3)=cos_omega
          rot_opera(1,2)=-sin_omega*kz(qz)/MAX(eps,knorm(qx,qz,qy))
          rot_opera(1,3)= sin_omega*ky(qy)/MAX(eps,knorm(qx,qz,qy))
          rot_opera(2,3)=-sin_omega*kx(qx)/MAX(eps,knorm(qx,qz,qy))
          rot_opera(2,1)=-rot_opera(1,2)
          rot_opera(3,1)=-rot_opera(1,3)
          rot_opera(3,2)=-rot_opera(2,3)
          DO ii=1,3
              temp(ii)  =  vel(qx,qz,qy,1)*rot_opera(ii,1)+ vel(qx,qz,qy,2)*rot_opera(ii,2)+ vel(qx,qz,qy,3)*rot_opera(ii,3)
            dtemp(ii) = dvel(qx,qz,qy,1)*rot_opera(ii,1)+dvel(qx,qz,qy,2)*rot_opera(ii,2)+dvel(qx,qz,qy,3)*rot_opera(ii,3)
          END DO
          vel(qx,qz,qy,1) = temp(1)
          vel(qx,qz,qy,2) = temp(2)
          vel(qx,qz,qy,3) = temp(3)
          dvel(qx,qz,qy,1)= dtemp(1)
          dvel(qx,qz,qy,2)= dtemp(2)
          dvel(qx,qz,qy,3)= dtemp(3)
        END DO
      END DO
    END DO
    DEALLOCATE(rot_opera,temp,dtemp)
  END IF
  
END SUBROUTINE rk_expon



!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!                                                          !!
!!                        rk_endstep                        !!
!!                                                          !!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
SUBROUTINE rk_endstep(be)
  USE parameters
  USE variables
  USE mpi_var

  IMPLICIT NONE

  REAL, INTENT(IN)    :: be
  INTEGER :: iscal, qqy, pid
  COMPLEX, ALLOCATABLE, DIMENSION (:,:,:,:)    :: force
 

!!!!! Add the beta terms in the Runge Kutta algorithm
  IF (VELEQ) vel =vel  + dt*be*dvel  
  IF (MAGEQ) mag =mag  + dt*be*dmag  
  IF (SCAEQ) scal=scal + dt*be*dscal


!!!!! Impose zero modes
  pid = whereis_ky(0,1)
  qqy = whereis_ky(0,2)

  IF (my_fpart.eq.pid) THEN
    IF ((VELEQ).AND.(.NOT.ROTEQ)) THEN
      vel(0,0,qqy,1)=u0_x*Ns
      vel(0,0,qqy,2)=u0_y*Ns
      vel(0,0,qqy,3)=u0_z*Ns
    END IF

    IF (MAGEQ) THEN
      mag(0,0,qqy,1)=b0_x*Ns
      mag(0,0,qqy,2)=b0_y*Ns
      mag(0,0,qqy,3)=b0_z*Ns
    END IF
  END IF


!!!!! Impose mode n/2 to zero and the reality condition
  CALL zero_nhalf()
  CALL symmetrize_all()

END SUBROUTINE rk_endstep



!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!                                                          !!
!!                       compute_dt                         !!
!!                                                          !!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
SUBROUTINE compute_dt()
  USE parameters
  USE variables
  USE mpi_var
  IMPLICIT NONE

  REAL, ALLOCATABLE, DIMENSION(:,:,:)        :: vecx, vecy, vecz
  COMPLEX, ALLOCATABLE, DIMENSION(:,:,:)     :: vecc
  REAL                                       :: maxvec_node, maxvec, dt_node
  
  SELECT CASE (fixed_dt)
  CASE(1) 
    dt=dt_dat
  CASE(0)
  
    IF (VELEQ) THEN
      ALLOCATE(vecx(0:rx,0:ry,0:rz),vecy(0:rx,0:ry,0:rz),vecz(0:rx,0:ry,0:rz),vecc(0:cx,0:cz,0:cy))
      vecc=vel(:,:,:,1)
      CALL ifft3ds(vecc,vecx)
      vecc=vel(:,:,:,2)
      CALL ifft3ds(vecc,vecy)
      vecc=vel(:,:,:,3)
      CALL ifft3ds(vecc,vecz)
      maxvec_node=MAX(EPS,pi*MAXVAL(ABS(vecx)/dx+ABS(vecy)/dy+ABS(vecz)/dz))
      DEALLOCATE(vecx,vecy,vecz,vecc)
      SELECT CASE (dt_global)
        CASE(1)
          CALL mpi_allreduce(maxvec_node,maxvec,1,MPI_MYREAL,MPI_MAX,MPI_COMM_WORLD,ierr)
        CASE(0)
          CALL mpi_allreduce(maxvec_node,maxvec,1,MPI_MYREAL,MPI_MAX,MPI_COMM_INSTANCE,ierr)
      END SELECT
      dt=cfl/maxvec
    ELSE
      dt=dt_dat
    ENDIF
    
    IF (MAGEQ) THEN
      ALLOCATE(vecx(0:rx,0:ry,0:rz),vecy(0:rx,0:ry,0:rz),vecz(0:rx,0:ry,0:rz),vecc(0:cx,0:cz,0:cy))
      vecc=mag(:,:,:,1)
      CALL ifft3ds(vecc,vecx)
      vecc=mag(:,:,:,2)
      CALL ifft3ds(vecc,vecy)
      vecc=mag(:,:,:,3)
      CALL ifft3ds(vecc,vecz)
      maxvec_node=MAX(EPS,pi*MAXVAL(ABS(vecx)/dx+ABS(vecy)/dy+ABS(vecz)/dz))
      DEALLOCATE(vecx,vecy,vecz,vecc)
      SELECT CASE (dt_global)
        CASE(1)
          CALL mpi_allreduce(maxvec_node,maxvec,1,MPI_MYREAL,MPI_MAX,MPI_COMM_WORLD,ierr)
        CASE(0)
          CALL mpi_allreduce(maxvec_node,maxvec,1,MPI_MYREAL,MPI_MAX,MPI_COMM_INSTANCE,ierr)
      END SELECT
      dt=MIN(dt,cfl/maxvec)
    END IF

    SELECT CASE (FORCEPARA)
    CASE(1,7) 
      dt=MIN(dt, dt_f)
    END SELECT

    IF(field_exchange.EQ.1) THEN
      dt=MIN(dt,dt_ext)
      dt_node=dt
      CALL mpi_allreduce(dt_node,dt,1,MPI_MYREAL,MPI_MIN,MPI_COMM_WORLD,ierr)
    END IF    
       
  CASE default
    WRITE(6,*) 'Wrong value of the parameter fixed_dt (fixed_dt=',fixed_dt,')'
    STOP
  END SELECT

END SUBROUTINE compute_dt



