!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!  12 August 2011                                          !!
!!                     tt_exter.f90                         !!
!!                                                          !!
!!  external_forcing                                        !!
!!  external_modif_vel                                      !!
!!  external_init_vel                                       !!
!!  external_init_mag                                       !!
!!  external_init_scal                                      !!
!!  external_energy_spectrum                                !!
!!  external_stats_init                                     !!
!!  external_stats_end                                      !!
!!  external_stats_write                                    !!
!!  external_stats_compute                                  !!
!!  test_ensemble_average                                   !!
!!                                             version 2.05 !!
!!                                                          !!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!



!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!                                                          !!
!!                   external_forcing                       !!
!!                                                          !!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
SUBROUTINE external_forcing(force)
  USE parameters
  IMPLICIT NONE

  COMPLEX, DIMENSION(0:cx,0:cz,0:cy,3), INTENT(OUT)    :: force
  
  force=0.0
  
END SUBROUTINE external_forcing



!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!                                                          !!
!!                   external_modif_vel                     !!
!!                                                          !!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
SUBROUTINE external_modif_vel(u)
  USE parameters
  IMPLICIT NONE

  COMPLEX, DIMENSION(0:cx,0:cz,0:cy,3), INTENT(INOUT)    :: u
    
END SUBROUTINE external_modif_vel



!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!                                                          !!
!!                   external_init_vel                      !!
!!                                                          !!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
SUBROUTINE external_init_vel(u)
  USE parameters
  IMPLICIT NONE

  COMPLEX, DIMENSION(0:cx,0:cz,0:cy,3), INTENT(OUT)    :: u

  u=0.0
  
END SUBROUTINE external_init_vel



!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!                                                          !!
!!                   external_init_mag                      !!
!!                                                          !!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
SUBROUTINE external_init_mag(b)
  USE parameters
  IMPLICIT NONE

  COMPLEX, DIMENSION(0:cx,0:cz,0:cy,3), INTENT(OUT)    :: b

  b=0.0
  
END SUBROUTINE external_init_mag



!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!                                                          !!
!!                   external_init_scal                     !!
!!                                                          !!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
SUBROUTINE external_init_scal(s,iscal)
  USE parameters
  IMPLICIT NONE

  COMPLEX, DIMENSION(0:cx,0:cz,0:cy), INTENT(INOUT)    :: s
  INTEGER, INTENT(IN) :: iscal
  
  s=0.01 * iscal
  
END SUBROUTINE external_init_scal
  


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!                                                          !!
!!                external_energy_spectrum                  !!
!!                                                          !!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
FUNCTION external_energy_spectrum(k,pa,pb,pc,pd)
  USE parameters
  IMPLICIT NONE

  REAL        ::     external_energy_spectrum, k, pa , pb, pc, pd
  
  external_energy_spectrum = pa*k**4*exp(-2.0*(k/pb)**2)*(Ck*Ns)**2
   
END FUNCTION external_energy_spectrum



!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!                                                          !!
!!                  external_stats_init                     !!
!!                                                          !!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
SUBROUTINE external_stats_init()
  USE parameters
  USE variables
  USE statmod
  IMPLICIT NONE
   
END SUBROUTINE external_stats_init



!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!                                                          !!
!!                  external_stats_end                      !!
!!                                                          !!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
SUBROUTINE external_stats_end()
  USE parameters
  USE variables
  USE statmod
  IMPLICIT NONE
   
END SUBROUTINE external_stats_end



!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!                                                          !!
!!                  external_stats_write                    !!
!!                                                          !!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
SUBROUTINE external_stats_write()
  USE parameters
  USE variables
  USE statmod
  IMPLICIT NONE
  
END SUBROUTINE external_stats_write



!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!                                                          !!
!!                 external_stats_compute                   !!
!!                                                          !!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
SUBROUTINE external_stats_compute(iter)
  USE parameters
  USE variables
  USE mpi_var
  USE statmod
  USE spec_rest_mod  
  IMPLICIT NONE

  INTEGER, INTENT(IN)  :: iter
  REAL :: eneru, enerb, dissu, dissb, kinhel, dkinhel, crosshel, dcrosshel, maghel,dmaghel, inj_e, inj_h, Lintu, Lintb, Rlambda, kmaxeta
  REAL :: dkh
  
  COMPLEX, ALLOCATABLE, DIMENSION(:,:,:,:) ::  du, kdu, u_dns
  REAL,    ALLOCATABLE, DIMENSION(:,:,:,:) ::  u_real, f_real
  REAL,    ALLOCATABLE, DIMENSION(:,:,:)   ::  f_par, f_perp  
  COMPLEX, DIMENSION(0:cx,0:cz,0:cy,3)     :: rest_a
  REAL, DIMENSION(0:rest_nshell+1,6)       :: spec_A, spec_P, spec_I
  REAL, DIMENSION(0:rest_nshell+1)         :: spec_tot
  
  REAL :: deltaE, deltaD, Aalpha, Agamma, Palpha, Pgamma, Cgamma, Aalpha_par, Aalpha_perp, Agamma_par, Agamma_perp, IDalpha, IDgamma
  REAL :: Edns, Ddns, f_square, u_square
  INTEGER :: qx, qy, qz, ii


  eneru=0.0
  enerb=0.0
  dissu=0.0
  dissb=0.0
  kinhel=0.0
  dkinhel=0.0 
  crosshel=0.0
  dcrosshel=0.0
  maghel=0.0
  dmaghel=0.0
  inj_e=0.0
  inj_h=0.0
  Lintu=0.0
  Lintb=0.0
  Rlambda=0.0
  kmaxeta=0.0
  
  IF (ABS(inj_energy)>eps) THEN
    inj_e=inj_energy/dt
    inj_h=inj_helicity/dt
  END IF 
    
  IF (VELEQ) THEN
    CALL correl_vector(vel,vel,0,0.5,eneru)
    CALL correl_vector(vel,vel,1,nu,dissu)
    CALL correl_helicity(vel,vel,0,0.5,kinhel)
    CALL correl_helicity(vel,vel,1,2.0*nu,dkinhel)    
	CALL correl_vector(vel,vel,-1,0.75*pi/eneru,Lintu)
	CALL compute_Rlambda(eneru,dissu,Rlambda)
	CALL compute_kmaxeta(dissu,kmaxeta)
  END IF

  IF (MAGEQ) THEN
    CALL correl_vector(mag,mag,0,0.5,enerb)
	CALL correl_vector(mag,mag,1,eta,dissb)
    CALL correl_vector(vel,mag,0,1.0,crosshel)
    CALL correl_vector(vel,mag,1,(nu+eta),dcrosshel)    
    CALL correl_helicity(mag,mag,-1,1.0,maghel)
    CALL correl_helicity(mag,mag,0,2.0*eta,dmaghel)
    CALL correl_helicity(vel,mag,1,2.0*nu,dkh)
    dkinhel=dkinhel+dkh
    CALL correl_helicity(mag,vel,1,2.0*eta,dkh)
    dkinhel=dkinhel+dkh
    CALL correl_helicity(mag,mag,1,2.0*eta,dkh)
    dkinhel=dkinhel+dkh
   	CALL correl_vector(mag,mag,-1,0.75*pi/enerb,Lintb)
    CALL compute_kmaxeta(dissu+dissb,kmaxeta)
  END IF
    
  IF (field_exchange.EQ.1) THEN
    ALLOCATE(u_dns(0:cx,0:cz,0:cy,3))
    u_dns=vel
    CALL reduce_send_and_receive(u_dns(:,:,:,1))
    CALL reduce_send_and_receive(u_dns(:,:,:,2))
    CALL reduce_send_and_receive(u_dns(:,:,:,3))
    u_dns(:,:,:,1)=u_dns(:,:,:,1)*zeros_twothird
    u_dns(:,:,:,2)=u_dns(:,:,:,2)*zeros_twothird
    u_dns(:,:,:,3)=u_dns(:,:,:,3)*zeros_twothird
    
    IF (LES_ident.EQ.1) THEN
      CALL correl_vector(u_dns,u_dns,0,0.5,Edns)
      CALL correl_vector(u_dns,u_dns,1,nu,Ddns)
      deltaE=(Edns-eneru)/Edns
      deltaD=(Ddns-dissu)/Ddns
    
      ALLOCATE(du(0:cx,0:cz,0:cy,3))
      ALLOCATE(kdu(0:cx,0:cz,0:cy,3))
      du=u_dns-vel
      kdu(:,:,:,1)=ksquare*(u_dns(:,:,:,1)-vel(:,:,:,1))
      kdu(:,:,:,2)=ksquare*(u_dns(:,:,:,2)-vel(:,:,:,2))
      kdu(:,:,:,3)=ksquare*(u_dns(:,:,:,3)-vel(:,:,:,3))
      CALL correl_vector( du,  du,  0, 1.0, Aalpha )
      CALL correl_vector( kdu, kdu, 0, 1.0, Agamma )
      CALL correl_vector( du,  vel, 0, 1.0, Palpha )
      CALL correl_vector( kdu, vel, 0, 1.0, Pgamma )
      CALL correl_vector( du,  kdu, 0, 1.0, Cgamma )
      CALL correl_vector( du,  vel, 1, nu, IDalpha )
      CALL correl_vector( kdu, vel, 1, nu, IDgamma )      
 
      
      MM(nx/2)=1
      rest_a=du*du
      CALL spec_rest_compute_3D(rest_a, spec_A)
      rest_a=du*vel
      CALL spec_rest_compute_3D(rest_a, spec_P)
      rest_a=nu*du*vel
      CALL spec_rest_compute_3D(rest_a, spec_I)
      MM(nx/2)=0
      
      ALLOCATE(f_real(0:rx,0:ry,0:rz,3))
      ALLOCATE(u_real(0:rx,0:ry,0:rz,3))
      ALLOCATE(f_par(0:rx,0:ry,0:rz))
      ALLOCATE(f_perp(0:rx,0:ry,0:rz))
      f_par=0.0
      f_perp=0.0
      f_real=0.0
      u_real=0.0
      CALL ifft3ds(du(:,:,:,1),f_real(:,:,:,1))
      CALL ifft3ds(vel(:,:,:,1),u_real(:,:,:,1))
      CALL ifft3ds(du(:,:,:,2),f_real(:,:,:,2))
      CALL ifft3ds(vel(:,:,:,2),u_real(:,:,:,2))
      CALL ifft3ds(du(:,:,:,3),f_real(:,:,:,3))
      CALL ifft3ds(vel(:,:,:,3),u_real(:,:,:,3))
      DO qz=0,rz
        DO qy=0,ry
          DO qx=0,rx-2
            u_square=0.0
            f_square=0.0
            DO ii=1,3
              f_par(qx,qy,qz)=f_par(qx,qy,qz)+(f_real(qx,qy,qz,ii)*u_real(qx,qy,qz,ii))
              u_square=u_square+(u_real(qx,qy,qz,ii)*u_real(qx,qy,qz,ii))
              f_square=f_square+(f_real(qx,qy,qz,ii)*f_real(qx,qy,qz,ii))
            END DO
            f_par(qx,qy,qz)=f_par(qx,qy,qz)*f_par(qx,qy,qz)/SQRT(u_square)
            f_perp(qx,qy,qz)=f_square-f_par(qx,qy,qz)
          END DO
        END DO
      END DO      
      Aalpha_par  = SUM(f_par) /REAL(Ns)
      Aalpha_perp = SUM(f_perp)/REAL(Ns)
            
      f_par=0.0
      f_perp=0.0
      f_real=0.0
      CALL ifft3ds(kdu(:,:,:,1),f_real(:,:,:,1))
      CALL ifft3ds(kdu(:,:,:,2),f_real(:,:,:,2))
      CALL ifft3ds(kdu(:,:,:,3),f_real(:,:,:,3))
      DO qz=0,rz
        DO qy=0,ry
          DO qx=0,rx-2
            u_square=0.0
            f_square=0.0
            DO ii=1,3
              f_par(qx,qy,qz)=f_par(qx,qy,qz)+(f_real(qx,qy,qz,ii)*u_real(qx,qy,qz,ii))
              u_square=u_square+(u_real(qx,qy,qz,ii)*u_real(qx,qy,qz,ii))
              f_square=f_square+(f_real(qx,qy,qz,ii)*f_real(qx,qy,qz,ii))
            END DO
            f_par(qx,qy,qz)=f_par(qx,qy,qz)*f_par(qx,qy,qz)/SQRT(u_square)
            f_perp(qx,qy,qz)=f_square-f_par(qx,qy,qz)
          END DO
        END DO
      END DO      
      Agamma_par  = SUM(f_par) /REAL(Ns)
      Agamma_perp = SUM(f_perp)/REAL(Ns)
      DEALLOCATE(f_real)
      DEALLOCATE(f_par)
      DEALLOCATE(u_real)
      DEALLOCATE(f_perp)      
            
      DEALLOCATE(du)
      DEALLOCATE(kdu)
    END IF ! LES_ident
    
    DEALLOCATE(u_dns)
  END IF ! field_exchange
  

  IF (my_fpart.eq.io_fpart) THEN   
    IF (field_exchange.EQ.0) THEN
      WRITE(out_unit,'(21E15.7,I9)') t, eneru , enerb, eneru+enerb, dissu, dissb, dissu+dissb,   &
	                        kinhel, dkinhel, maghel, dmaghel, crosshel/(eneru+enerb), dcrosshel, &
	                        inj_energy, inj_helicity, inj_maghel, inj_cross, Lintu, Lintb, Rlambda, kmaxeta, iter
	ELSE
	  IF (LES_ident.EQ.0) THEN
	    WRITE(out_unit,'(9E15.7,I9)')  t, eneru , dissu, kinhel, inj_e, inj_h, Lintu, Rlambda, kmaxeta, iter
	  ELSE
	    WRITE(out_unit,'(24E15.7,I9)')  t, eneru, dissu, kinhel, inj_e, inj_h, Edns, deltaE, Ddns, deltaD, Aalpha, Aalpha_par, Aalpha_perp, Agamma, Agamma_par, Agamma_perp, Palpha, Pgamma, Cgamma, IDalpha, IDgamma, Lintu, Rlambda, kmaxeta, iter
	    spec_tot= spec_A(:,1)+spec_A(:,2)+spec_A(:,3)
	    WRITE(out_spa,'(I9, 60E15.7)') iter, t, spec_tot(1:rest_nshell)
	    spec_tot= spec_P(:,1)+spec_P(:,2)+spec_P(:,3)
	    WRITE(out_spp,'(I9, 60E15.7)') iter, t, spec_tot(1:rest_nshell)
	    spec_tot= spec_I(:,1)+spec_I(:,2)+spec_I(:,3)
	    WRITE(out_spi,'(I9, 60E15.7)') iter, t, spec_tot(1:rest_nshell)
	  END IF
	END IF
  END IF
    
END SUBROUTINE external_stats_compute



