#!/bin/sh

## -----------------------------------------
## Impose to read the script with bash 
## -----------------------------------------

#$ -S /bin/bash

## -----------------------------------------
## Defines some important parameters :
##   - the name of the simulation (in the present
##     implementation of turbo, it is supposed to
##     be 6 characters)
##   - the number of CPU
##   - the time limit
##   - the run directory
## -----------------------------------------

NAME_RUN=test00
#$ -pe mpich 2
#@$ -lt 23:50:00
MYDIR=`pwd`
echo "$(echo $NAME_RUN)" > name_run.file

## -----------------------------------------
# # The name of the job that appears on ANIC and part of related files
## -----------------------------------------

#$ -N getspec

## -----------------------------------------
## Clean the successfully terminated flag
## -----------------------------------------

rm freeres.bash

## -----------------------------------------
## export the variable MPIR_HOME which needs to be 
## defined in the .bash_profile or .tcshrc.
## -----------------------------------------

MPIR_HOME=/usr/local/mpich-intel-fortran
#$ -v MPIR_HOME

## -----------------------------------------
## Change directory to this job's working directory
## -----------------------------------------

#$ -cwd

## -----------------------------------------
## Make a copy of the list of machines on which 
## the job is run into a local file, e.g. hostlist
## -----------------------------------------

cp $TMPDIR/machines $NAME_RUN.hostlist

## -----------------------------------------
## the number of processors should be the number
## of nodes in hostlist
## -----------------------------------------

NPROC=$(cat $NAME_RUN.hostlist | wc -l)

## -----------------------------------------
## Echo some data regarding the job
## -----------------------------------------

echo Start time is `date`
echo Directory is `pwd`
echo This job runs on the following processors:
cat $NAME_RUN.hostlist
echo This job has allocated $NPROC processors
echo

## -----------------------------------------
## Create a script to kill nicely all the 
## processes on the nodes after the job is
## finished. The script is freeres.bash
## -----------------------------------------

(
  sleep 20
  echo "#!/bin/bash" > freeres.bash
  echo "#$ -S /bin/bash" >> freeres.bash
  echo "#$ -N Free-$(echo $JOB_ID)" >> freeres.bash
  echo "#$ -cwd" >> freeres.bash
  echo "#$ -hold_jid $(echo $JOB_ID)" >> freeres.bash
  for m in $(sort -u $NAME_RUN.hostlist); do
    WORK_STR=$(/usr/kerberos/bin/rsh $m ps -u $USER -o pid,args | awk -v pat="$SGE_O_WORKDIR" '$2 ~ pat {print $1}')
    echo rsh $m kill $WORK_STR >> freeres.bash
  done
  chmod u+x $SGE_O_WORKDIR/freeres.bash
  /usr/kerberos/bin/rsh master ". /etc/profile.d/sge_settings.sh; /usr/local/sge/bin/lx24-x86/qsub $SGE_O_WORKDIR/freeres.bash" 
) &

## -----------------------------------------
## Run the turbo code
## -----------------------------------------

$MPIR_HOME/bin/mpirun -np $NPROC -machinefile $NAME_RUN.hostlist $MYDIR/getspectrum < name_run.file

## -----------------------------------------
## Echo time
## -----------------------------------------

echo End time is `date`

